/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2008 Funambol, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 *
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */


#ifndef SYNCENGINE_H_
#define SYNCENGINE_H_

#include <e32base.h>
#include <e32std.h>
#include "base/fscapi.h"
#include "base/util/symbianUtils.h"
#include "utils.h"
#include "Funambol.hrh"
#include "SyncActiveObject.h"
#include "SyncScheduler.h"

// FORWARD DECLARATIONS
class CFunambolAppUi;
class CSyncScheduler;


/**
 * Synchronization status of current sync.
 */
enum TCurrentSyncStatus
{
    ESyncStatusConnecting,
    ESyncStatusConnected,
    ESyncStatusDisconnected,
    ESyncStatusLoggingOn,
    ESyncStatusLoggedOn,
    ESyncStatusSendingData,
    ESyncStatusReceivingData,
    ESyncStatusMappings,
    ESyncStatusCompleted,
    ESyncStatusNotCompleted,
    ESyncStatusSendingItemNumber,
    ESyncStatusReceivingItemNumber,
    ESyncStatusError,
    ESyncStatusCanceled
};

/**
 *  Possible status of last sync.
 */
enum TLastSyncStatus
{
    ELastSyncSuccessful,
    ELastSyncFailed,
    ELastSyncNotFound
};

/**
 *  The sync can be triggered only from these runners.
 */
enum TSyncRunners
{
    ESyncTriggeredManual,           // From UI buttons
    ESyncTriggeredPush,
    ESyncTriggeredClientPush,
    ESyncTriggeredAsyncAction,      // From Active object's RunL()
    ESyncTriggeredScheduled
};


typedef TInt TSmlJobId;

/**
 * Struct used to pair a jobID with a sourceID, used to understand
 * which source is synced given the current job ID.
 * Also used to flag a job as 'to be aborted' once it starts.
 */
typedef struct JobEntry {
    TSmlJobId   jobID;        /**< The job ID */
    TInt        sourceID;     /**< The corresponding source ID */
    TBool       toAbort;      /**< Set to 'ETrue' if you want to abort the job once it starts */
} JobEntry;



/**
 * This class is a fake CSyncEngine class, used only by Symbian emulator
 * instead of the real CSyncEngine defined in "SyncEngine.h".
 * The SyncML Client API are not released for emulator WINSCW, so we use
 * this fake class to be able to compile the client anyway.
 */
class CSyncEngine : public CActive
{

private:
    CSyncEngine() : CActive(CActive::EPriorityStandard)
    {
    }

    void ConstructL(CFunambolAppUi* aAppUI);
    void DoCancel() {}
    void RunL();
    TInt RunError(TInt aError);

public:

    static CSyncEngine* NewL(CFunambolAppUi* aAppUI);
    static CSyncEngine* NewLC(CFunambolAppUi* aAppUI);

    ~CSyncEngine() {}
    
    void StartSync(const RArray<TInt>& aSourceIDList, const TSyncRunners from);
    void StartSync(const TInt aSourceID);

    TInt StopSync()             { return 0; }
    TBool CheckSyncInProgress() { return EFalse; }
    TInt GetCurrentSourceID()   { return 0; }
    TInt getActiveProfileID() { return 0; }

    void SetCurrentSource(const TInt aSourceID) { iCurrentSource = aSourceID; }

    void syncConfig();

    RArray<TInt> CSyncEngine::GetActiveSourcesList();
    TLastSyncStatus GetLastSyncStatus(const TInt aSourceID, TTime* aLastSyncTime, TInt* aLastSyncResult) { 
        return ELastSyncFailed; 
    }

    TSyncMode getSourceSyncMode(const TInt aSourceID) { return ESyncModeNone; }
    
    /// Also saves the config.
    void setSourceSyncMode(const TInt aSourceID, const TSyncMode aSyncMode) {}
    
    /// Also saves the config.
    void setSourceTimestamp(const TInt aSourceID, const TUint timestamp) {}

    void EmailSyncEnded(TCurrentSyncStatus status) {}
    void enableScheduler(const bool aIsRepetitive, const unsigned int aInterval) {}
    
    /**
     * Disables the sync scheduler (proxy method).
     * If a scheduled sync is already set, it will not complete because
     * it will be discarded on RunL().
     */
    void disableScheduler() {}
    
    /// Checks the sync scheduler (proxy method).
    bool isSchedulerEnabled() { return false; }
    
    /// Removes a sourceID from the scheduler list (proxy method).
    bool removeFromScheduler(const TInt aSourceID) { return false; }
    void syncConfigEnded(TInt status);

    private:
        /// Pointer to the object that created this instance, the appUI.
    /// Used to callback methods on sync events notifications.
    CFunambolAppUi*             iAppUI;
     
    /**
     * Used for scheduled (and delayed) synchronizations.
     */
    CSyncScheduler*              iScheduler;

    /// The current source that is under sync.
    TInt                        iCurrentSource;

    /// Counter of total items received (Server -> Client) for this job.
    /// It's resetted when the job starts (see ManageNewJobStarted())
    TInt                        iTotalClientItems;
    
    /// Counter of total items sent (Client -> Server) for this job.
    /// It's resetted when the job starts (see ManageNewJobStarted())
    TInt                        iTotalServerItems;

    /**
     * List of queued sync jobs.
     * It's necessary to retrieve the sourceID once the job starts, given
     * the job ID: otherwise we cannot know which source is under sync.
     * A new entry is appended every time a job is created, and it's
     * removed once the job really starts (it's no more useful).
     * It's also used to flag a queued job that need to be stopped, so
     * it will be aborted when the job really starts.
     */
    RArray<JobEntry>            iQueuedJobs;

    /// Used to avoid manipulating the iQueuedJobs 
    /// member between different threads.
    RCriticalSection            iCriticalQueuedJobs;

    /**
     * Enumeration of asynchronous SyncML actions.
     * These are actions that can be executed asynchronously, using
     * the active object RunL().
     */
    enum TAsyncAction
    {
        EIdle,
        EStartSync,
        EStopSync,
        EEnableScheduler
    };
    
    /**
     * The requested asynchronous SyncML action.
     * The active object will check this attribute to know
     * which action need to be executed, one of TAsyncAction.
     */
    TAsyncAction               iAsyncAction;
    
    /**
     * A list of source IDs, requested to sync
     * through the active object.
     */
    RArray<TInt>                iRequestedSourceList;

    /**
     * The ID of the thread that owns the RSyncMLSession.
     * @note  It's mandatory to originate and complete an asynchronous 
     *        request inside the same thread, so here is the ID of the sole
     *        thread used to originate and complete requests.
     */
    TInt                        iMainThreadID;

    /** Number of sources that are under synchronization and have not
     * terminated. This value is initialized at the beginning of a sync
     * and updated as each source terminates.
     */
    TInt iNumSources;

};

#endif  // SYNCENGINE_H_
