/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2008 Funambol, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 *
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */


#ifndef __FUNAMBOLAPPUI_h__
#define __FUNAMBOLAPPUI_h__


#include <aknappui.h>
#include <aknwaitdialog.h> // wait note
#include "base/util/symbianUtils.h"
#include "base/util/StringBuffer.h"
#include "SyncSettingsDialog.h"
#include "ProgressDialog.h"

#ifdef __GCCE__
#include "SyncEngine.h"
#else
#include "FakeSyncEngine.h"
#endif

using namespace Funambol;

// FORWARD DECLARATIONS
class CFunambolAppView;
class CPushEngine;
class CModel;
class CCleanupContactsManager;
class CCleanupCalendarManager;
class CCleanupMailManager;
class CCleanupPhotosManager;
class CAccountSettingsDialog;
class CAbout;
class CXmlRpcEngine;

/**
 ***************************************************************************
 * CFunambolAppUi application UI class.
 * Interacts with the user through the UI and request message processing
 * from the handler class.
 * This class it's the main UI controller, it contains the following objects:
 * - CFunambolAppView: the main view class, to manage UI objects
 * - CSyncEngine: used to handle the synchronization process
 * - CPushManager: used to handle the push service (CTP + CP)
 ***************************************************************************
 */
class CFunambolAppUi : public CAknAppUi, public CActive
{
private:

    /**
     * The application view
     * Owned by CFunambolAppUi
     */
    CFunambolAppView*      iAppView;

    /**
     * For autostart management
     */
    CModel*                 iModel;

    /**
     * For waiting on install dialog
     */
    CAknWaitDialog*         iWaitDialog;

    /**
     * For delaying installer start
     */
    CPeriodic*              iTimer;

    /// The sync engine. Manages the synchronizations.
    CSyncEngine*            iSyncEngine;

    /// The push engine. Manages the push activation / notifications.
    CPushEngine*            iPushEngine;


    /**
     * Result of last sync (0 on success)
     * Set by the listener of sync status (updateSourceStatus())
     * Used by ConfigSyncEnded() and by PhotoSyncEnded to take
     * proper actions after the config sync.
     */
    TInt                    iLastSyncResult;

    /**
     * Cleanup manager for contacts (NULL allowed)
     */
    CCleanupContactsManager* contactsManager;

    /**
     * Cleanup manager for calendar (NULL allowed)
     */
    CCleanupCalendarManager* calendarManager;

    /**
     * Cleanup manager for mail (NULL allowed)
     */
    CCleanupMailManager* iMailManager;

    /**
     * Cleanup manager for photos (NULL allowed)
     */
    CCleanupPhotosManager* iPhotosManager;

    /**
     * Wait note used when cleanup is in progress (NULL allowed)
     */
    ProgressDialog*          cleanupWaitDlg;

    RArray<TInt>             iSortedSourceIDs;

    /**
     * This field is meaningful only during the cleanup process. It is used to
     * keep trak of the last source to clean, so the wait dlg can be deleted
     */
    TInt                     iLastCleanupSource;

    /// If true, the application is currently in foreground.
    TBool                    iForeground;


    // ---------- For Active object usage -------------
    /**
     * Enumeration of asynchronous appUI actions.
     * These are actions that can be executed asynchronously, using
     * the active object RunL().
     */
    enum TAsyncAction
    {
        EIdle,
        EDisplayMessage     // will display a message given its ID
    };

    /**
     * The requested asynchronous appUI action.
     * The active object will check this attribute to know
     * which action need to be executed, one of TAsyncAction.
     */
    TAsyncAction            iAsyncAction;

    /**
     * The ID of the main thread.
     * @note  It's mandatory to originate and complete an asynchronous
     *        request inside the same thread, so here is the ID of the sole
     *        thread used to originate and complete requests.
     */
    TInt                    iMainThreadID;

    /// The message ID for a message to be displayed, passing through the active object
    TInt                    iRequestedMsgID;

    /// The warning level for a message to be displayed, passing through the active object
    TWarningLevel           iRequestedWarningLevel;

    // about view (rich text editor control container)
    CAbout*                 iAboutView;

    /// The account settings dialog
    CAccountSettingsDialog* iAccountSettingsDialog;

    // xml-rpc socket server and message parser
    CXmlRpcEngine*          iXmlRpcEngine;


    // ----------- Methods from base classes ---------------
    /**
     * From CEikAppUi, HandleCommandL.
     * Takes care of command handling.
     * @param aCommand Command to be handled.
     */
    void HandleCommandL(TInt aCommand);

    /**
     *  HandleStatusPaneSizeChange.
     *  Called by the framework when the application status pane
     *  size is changed.
     */
    void HandleStatusPaneSizeChange();

    void HandleWsEventL(const TWsEvent &aEvent, CCoeControl *aDestination);

    /**
     * Callback method invoked when the application goes to, or returns
     * from background (used to refresh sources timestamps)
     */
    void HandleForegroundEventL(TBool aForeground);

    TBool ProcessCommandParametersL(CApaCommandLine &aCommandLine);
    void DynInitMenuPaneL(TInt aResourceId, CEikMenuPane* aMenuPane);

    /**
     *  From CCoeAppUi, HelpContextL.
     *  Provides help context for the application.
     *  size is changed.
     */
    CArrayFix<TCoeHelpContext>* HelpContextL() const;


    // ----------- Active Object's methods ------------
    /**
     * Used to request an asynchronous action to be executed.
     * Sets the 'iAsyncAction' member and signals that the asynchronous
     * request associated with 'iStatus' is complete, so that the
     * active object's RunL() will be called by active scheduler.
     * @note   we must signal the status in the same thread that originated the request.
     *
     * @param aAction  the action to be executed by the active object, one of TAsyncAction
     */
    void RequestAsyncAction(const TAsyncAction aAction);

    /**
     * Handles active object's request completion events.
     * It's called by the active scheduler when a request completion event occurs.
     * The asynchronous action to be executed is defined by 'iAsyncAction' member,
     * which is previously set when the request is raised (in RequestAsyncAction()).
     *
     * The active object is used to execute asynchronous actions like DisplayMessage()
     * from the main thread. Other threads cannot use the CEikonEnv::Static() objects.
     */
    void RunL(void);

    /**
     * Implements cancellation of  outstanding request.
     * This function is called as part of the active object's Cancel().
     */
    void DoCancel(void);

    /// Called by the active scheduler if the active object's RunL() function leaves.
    TInt RunError(TInt aError);


    // ----------- Private methods ---------------

    /**
     * Initializes the client.
     * Creates the log and the config, then checks if we have to show
     * the AccountForm. Finally, starts the CTP and CP services.
     */
    void InitializeClient();

    /**
     * Formats a status string, for a generic source, given the source status
     * and the last sync timestamp. This method accesses the string resources and
     * it's used to display the status of each source on the main UI.
     *
     * @param aSourceStatus        the source status, one of TLastSyncStatus enumeration
     * @param aSourceLastSyncTime  the last sync timestamp, as a TTime object
     * @return                     a formatted string with the status of the source.
     *                             (e.g. "25/03/2008 12:22:02" if the status was successful)
     */
    TBuf<32> FormatStatusText(const TLastSyncStatus aSourceStatus, const TTime aSourceLastSyncTime);


    TInt GetSourcesIDFromServerURI(const char* aServerURI);


    /// Returns true if "Sync All" or "Cancel Sync" are currently visible.
    bool IsMainButtonsVisible();

    /**
     * Gets the phone IMEI
     */
    StringBuffer GetImeiL();

    /// Loads the main application icon, from the mbm bitmaps resource.
    void loadAppIcon();

    /// Populates internal array of source names, sorted for the SyncALL.
    void CreateSortedSourcesArray();

    // Goto command handler: switch to Symbian system application view (contatcs, calendar, etc)
    void OpenSourceViewL(TInt aSourceId);

    // Install the Java Email client (in single bundle mode)
    void installJavaEmailClient();

    /// Returns a new allocated WCHAR*: the label of desired source (read from resources)
    /// Returns NULL if input source ID is not one of the valid ones.
    const WCHAR* getSourceLabelFromID(const TInt aSourceID);

public:

    /**
     * 2nd phase constructor.
     * The model is initialized here, as we need it to check the autostart
     * setting in ProcessCommandParametersL().
     * All initialization stuff is moved to IntializeClientL().
     */
    void ConstructL();

    /**
     * CFunambolAppUi.
     * C++ default constructor. This needs to be public due to
     * the way the framework constructs the AppUi
     */
    CFunambolAppUi();

    /**
     * ~CFunambolAppUi.
     * Virtual Destructor.
     */
    virtual ~CFunambolAppUi();

    /**
     * Will exit the application, ensuring that running services are
     * correctly terminated. Then calls CAknAppUi::Exit().
     * If a service is running (push/schedule) a query dialog is displayed before exiting.
     */
    void Exit();

    /// Proxy method to CSyncEngine. Returns ETrue if a sync is running.
    TBool CheckSyncInProgress();

    /// Sends application to background.
    void GoToBackground();

    /// Sends application to foreground.
    void GoToForeground();

    /// Returns true if the application is in foreground.
    TBool isInForeground();

    /// Enables/disables the autostart.
    void EnableAutoStart(const TBool aEnable);

    /// Returns ETrue if autostart is enabled.
    TBool IsAutoStartEnabled();

    void ShowAccountSettingsDialog();
    void ShowSyncSettingsDialog();
    void ShowCleanupDialog();

    /// Shows the Autostart query dialog.
    /// @return  ETrue if the user clicked OK, EFalse if clicked Cancel.
    TBool ShowAutostartDialog();

    void ShowAboutView();
    void RemoveAboutView();

    /**
     * Called when fire button is pressed, by appView.
     */
    void OnFireButton();

    /**
     * Refreshes the main UI screen (calls iAppView::RefreshScreen).
     */
    void RefreshScreen();

    /**
     * Cycles through all sources, and refresh their status.
     * This method is called at the beginning and when main UI should be refreshed.
     */
    void RefreshAllSourceTimestamps();

    /**
     * Refresh the status text for the desired source.
     * Gets its status/timestamp from iSyncEngine and
     * creates a status string that is passed to the iAppView.
     */
    void RefreshSourceTimestamp(const TInt aSourceID);

    /**
     * Cycles through all sources, and refresh their status with the last one.
     * Last status is buffered inside AppView. This is called from HandleForegroundEventL().
     */
    void RefreshAllSourceLastStatus();

    /**
     * Refresh the status text for the desired source, with the last one.
     * Calls iAppView->RefreshSourceBufferedStatus() method.
     */
    void RefreshSourceLastStatus(const TInt aSourceID);


    /**
     * Sets the main menu buttons to "Sync All" or "Cancel Sync", checking if
     * a sync process is currently running.
     * If we're not in main screen, nothing is done.
     */
    void refreshMainButtons();


    /**
     * Starts the sync of all active sources, one by one.
     * The first will start immediately, the others will be queued.
     * @param from   defines who triggered the sync action, one of TSyncRunners
     */
    void StartSync(const TSyncRunners from);

    /**
     * Starts the sync of one source, given the source ID.
     * @param aSourceID  the source ID for the item we want to sync
     * @param from       defines who triggered the sync action, one of TSyncRunners
     */
    void StartSync(const TInt aSourceID, const TSyncRunners from);

    /**
     * Starts the sync of more sources, given a list of sourceIDs.
     * The first will start immediately, the others will be queued.
     * Will forward the call to iSyncEngine->StartSyncL().
     *
     * @param aSourceIDList  list of sourceID for the items we want to sync
     * @param from           defines who triggered the sync action, one of TSyncRunners
     */
    void StartSync(RArray<TInt> &aSourceIDList, const TSyncRunners from);


    /**
     * Checks user's credentials.
     * First stops any active service (push or scheduler), then starts
     * a sync of the configuration that is used also to authenticate the user.
     */
    void checkCredentials();

    /**
     * Stops the push or scheduler service if running.
     * @return  true if effectively stopped the service.
     * @note    the config is not changed to reflect the changed behaviour.
     */
    bool StopServices();

    /**
     * Starts the push or scheduler service, if not running.
     * Checks the config value, to know if a service was running.
     * The service is restarted only if now it's stopped.
     * @return  true if effectively restarted the service.
     */
    bool StartServices();

    /// Proxy method, calls iSyncEngine->disableScheduler().
    void disableSyncScheduler();

    /**
     * Sets the new folder location for C2S push, for the source 'sourceID'.
     * Actually used only for pictures, to set the new pictures folder location
     * to watch (Client push), given the root drive.
     * Note: changing the folder, the filesystem listener will be restarted
     *       automatically if necessary
     *
     * @param sourceID  ID of the source involved
     * @param storageType the drive type to get the folder location (mass memory, memory card or main memory)
     */
    void setClientPushFolderLocation(const int sourceID, const TInt aStorageType);

    /**
     * Updates the UI status string for the desired source.
     * This method is called by iSyncEngine when the status of a source
     * changes during the synchronization process, to refresh the UI.
     * We call iAppView::RefreshSourceStatus() to update the UI.
     * This method accesses the string resources.
     *
     * @param aSourceID     ID of the source to update
     * @param aSyncStatus   the sync status, one of TCurrentSyncStatus (see SyncEngine.h)
     * @param aSyncData     additional data, optional
     *                      (e.g. the number of items if status = ESyncStatusSendingItemNumber)
     */
    void UpdateSourceStatus(const TInt aSourceID, const TCurrentSyncStatus aSyncStatus, const TInt aSyncData = 0);

    /**
     * Displays a warning/error message.
     * Retrieves the correct resource string given a msg ID.
     * @note   the message should not be longer than 100 chars, otherwise it could not fit.
     */
    void DisplayMessage(const TInt aMsgID, const TWarningLevel aWarningLevel);

    /**
     * Displays a YES/NO query popup to the user, with the message corresponding to the msg ID passed.
     * @param aMsgID  ID of resource for the text message to display
     * @return        true if the user selected YES, false if the user selected NO
     */
    TBool DisplayQuery(const TInt aMsgID);

    /**
     * Displays the query for the very first sync, asking the user to continue or not.
     * The dialog has its own button resources: "Sync Now" / "Try Later".
     * @return        true if the user selected "Sync Now"(YES), false if the user selected "Try Later" (NO)
     */
    TBool ShowPicturesFirstSyncWarning();

    /**
     * Displays a custom query dialog where the user can choose between 2 options.
     * The caller must define the dialog title, text and the softkey buttons to be shown.
     * The dialog is not full screen, and will have a vertical scrollbar if the text specified
     * does not fit in one page.
     *
     * @param titleID     the resource ID of the string to be shown as the title
     * @param textID      the resource ID of the text to be shown inside the dialog
     * @param softkeysID  the resource ID of the softkeys buttons
     * @return            the ID of the button selected by the user.
     *                    The returned ID depends on the softkeys defined, for instance if
     *                    the softkeys used are R_AVKON_SOFTKEYS_OK_CANCEL, it will be one of
     *                    EAknSoftkeyOk / EAknSoftkeyCancel.
     */
    TBool ShowCustomQueryDialog(const TInt titleID, const TInt textID, const TInt softkeysID);

    /**
     * Returns a list of active source IDs.
     * With 'active' we mean sources that have been selected by the user
     * to be synchronized.
     */
    RArray<TInt> GetActiveSourcesList();

    /**
     * Change the application title to the value of the given resource
     */
    void changeAppTitle(TInt resource);

    /**
     * Reset the application title to the default title
     */
    void resetAppTitle();

    /**
     * Called when a config sync has ended.
     * Will execute proper actions, based on the result code of the sync.
     * This method is also responsible of updating the account setting dialog
     * if it is currently open
     */
    void ConfigSyncEnded();

    // called on photos sync end to display error pop-ups
    void PhotosSyncEnded();

    /**
     * Called when a mail sync has ended.
     * Will execute proper actions, based on the result code of the sync.
     */
    void  EmailSyncEnded(TCurrentSyncStatus ESyncStatusNotCompleted);

    /**
     * Perform the PIM cleanup
     */
    void cleanup(TBool email, TBool contacts, TBool calendar, TBool notes, TBool photos);

    /// Updates the sofware downloading the latest version (TODO).
    void updateSoftware();

    /// Callback method invoked at the end of contact cleanup
    void contactsCleanupCompleted();

    /// Callback method invoked at the end of calendar cleanup
    void calendarCleanupCompleted();

    /// Callback method invoked at the end of mail cleanup
    void mailCleanupCompleted();

    /// Callback method invoked at the end of notes cleanup
    void notesCleanupCompleted();

    /// Callback method invoked at the end of images folder cleanup
    void photosCleanupCompleted();

    /// Use this method to get a string resource, from its ID.
    StringBuffer readStringResource(const TInt resourceID);

    /**
     * Changes the main menu buttons, given the MENU resource ID.
     * This method is public but it should only be used within the controller
     * classes.
     */
    void changeMainMenuButtons(const TInt menuResourceID);

    /**
     * Changes the main menu buttons, given the MENU resource ID. This
     * method changes the buttons only if the main screen is currently
     * visible and no other dialogs/views are currently displayed.
     * This method is public but it should only be used within the controller
     * classes.
     */
    void changeMainScreenMenuButtons(const TInt menuResourceID);

    /**
     * Utility method, to resize the font used in main view and settings listboxes.
     * If the screen is portrait (width > heigth) and the resolution is low
     * (width < 300), then the font size is reduced.
     * This is necessary for a better result on small screen devices, otherwise
     * the text is not entirely visible.
     */
    bool SetListboxFontSize(CFormattedCellListBoxItemDrawer* drawer);
};

#endif // __FUNAMBOLAPPUI_h__

// End of File
