/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2008 Funambol, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 *
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */


#ifndef PUSH_ENGINE
#define PUSH_ENGINE

#include "base/fscapi.h"
#include "base/util/ArrayList.h"
#include "base/util/StringBuffer.h"
#include "push/PushListener.h"
#include "FunambolAppUi.h"
#include "ClientPush.h"

// Push type: bitmask value [bit 1 = CTP, bit 2 = SMS, bit 3 = CP]
#define PUSH_TYPE_NONE  0
#define PUSH_TYPE_CTP   1
#define PUSH_TYPE_SMS   2
#define PUSH_TYPE_CP    4
#define PUSH_TYPE_ALL   7       // 1 + 2 + 4

class CSMSPushEngine;

using namespace Funambol;


/**
 * 
 */
class CPushEngine : public CBase, public PushListener
{

private: 
    
    /// Pointer to the object that created this instance, the appUI.
    /// Used to callback methods on sync events notifications.
    CFunambolAppUi*            iAppUI;

    /// The client push. Manages the notification of client side changes
    /// for contacts and calendar. (notes: TODO)
    CClientPush*            iClientPush;
 
    // The SMS Push engine for subscribing to push server events
    CSMSPushEngine*         iSMSPushEngine;

    /// Constructor
    CPushEngine();

    /**
     * Contstructs the CPushEngine object.
     * - Initializes the network connection with FConnection
     * - Register this class to receive push notifications
     * - Starts the CTP service
     */
    void ConstructL(CFunambolAppUi* aAppUI);
    
    /**
     * Creates the default configuration for CTPConfig.
     */
    void createDefaultCTPConfig();
 
public:

    static CPushEngine* NewL (CFunambolAppUi* aAppUI);
    static CPushEngine* NewLC(CFunambolAppUi* aAppUI);

    /// Destructor
    ~CPushEngine();

    /**
     * Starts the Push service.
     * Will start the CTPService and the CP listeners.
     */
    void start();
    
    /**
     * Stops the Push service.
     * Will stop the CTPService and the CP listeners.
     */
    void stop();
    
    /// Restarts the Push service (stop+start).
    void restart();
    
    
    /// Returns the current push state
    TInt getPushState();
    
    /// Returns true if push is running (both CTP and CP up).
    /// Returns false if at least one active service is down.
    bool isPushRunning();
    
    /**
     * Sets the path where pictures are stored, to watch for file changes.
     * This is a proxy method calling CClientPush::setPicturesPath().
     * 
     * @param aPicturesPath  the absolute path (like "C:\Data\Images\")
     */
    void setPicturesPath(const StringBuffer& aPicturesPath) {
        if (iClientPush) {
            iClientPush->setPicturesPath(aPicturesPath);
        }
    }
    

    // --------- Listener methods -----------
    /**
     * This method is called when a push notification has arrived.
     * Converts the array of serverURI into corrispective source IDs,
     * and then calls the StartSyncL() method of appUI for the 
     * desired sources.
     * 
     * @see PushListener::onNotificationReceived
     */
    void onNotificationReceived(const ArrayList& serverURIList);
 

    /**
     * This method is called when a SMS push notification has arrived.
     * Calls AppUI StartSync with sources list
     */
    void onNotificationReceived(RArray<TInt>& aSourceList);
    
    /**
     * Method called when a CTP error occurs.
     * Displays an popup error message in case of auth failed or connection failed.
     * @see PushListener::onCTPError
     */
    void onCTPError(const int errorCode, const int additionalInfo = 0);

};

#endif  // PUSH_ENGINE
