/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2008 Funambol, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 *
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */


#ifndef SYNCENGINE_H_
#define SYNCENGINE_H_

#include <e32base.h>
#include <e32std.h>
#include <syncmlclient.h>
#include <syncmlclientds.h>
#include <syncmldef.h>
#include <syncmlobservers.h>
#include "base/fscapi.h"
#include "base/util/stringUtils.h"
#include "base/util/symbianUtils.h"
#include "base/util/ArrayList.h"
#include "Funambol.hrh"


#ifdef FAKE_IMAP_ACCOUNT
#include "FakeImapEmailEngine.h"
#else
#include "ImapEmailEngine.h"
#endif


// FORWARD DECLARATIONS
class CFunambolAppUi;
class CSyncScheduler;

using namespace Funambol;

class ListenerPool
{
public:
    ListenerPool(CSyncEngine* aSyncEngine, TInt aSourceId) :
        iSyncEngine(aSyncEngine), iSourceId(aSourceId) {}
    ~ListenerPool();

    void setListeners();
    void releaseAllListeners();

private:
    CSyncEngine* iSyncEngine;
    TInt iSourceId;
};

/**
 * Synchronization status of current sync.
 */
enum TCurrentSyncStatus
{
    ESyncStatusConnecting,
    ESyncStatusConnected,
    ESyncStatusDisconnected,
    ESyncStatusLoggingOn,
    ESyncStatusLoggedOn,
    ESyncStatusSendingData,
    ESyncStatusReceivingData,
    ESyncStatusMappings,
    ESyncStatusCompleted,
    ESyncStatusNotCompleted,
    ESyncStatusSendingItemNumber,
    ESyncStatusReceivingItemNumber,
    ESyncStatusError,
    ESyncStatusInterrupted,
    ESyncStatusCanceled,
    ESyncStatusCanceling,
    ESyncStatusQuotaExceeded
};

/**
 *  Possible status of last sync.
 */
enum TLastSyncStatus
{
    ELastSyncSuccessful,
    ELastSyncFailed,
    ELastSyncNotFound,
    ELastSyncInterrupted,
    ELastSyncCanceled,
    ELastSyncQuotaExceeded
};


/**
 *  The sync can be triggered only from these runners.
 */
enum TSyncRunners
{
    ESyncTriggeredManual,           // From UI buttons
    ESyncTriggeredPush,
    ESyncTriggeredClientPush,
    ESyncTriggeredScheduled
};



/**
 * The SyncEngine class is the core of the whole synchronization process.
 * The main method is StartSync(), which starts a new sync process for one or a list of sources.
 * The sync can be started for only one source at the time, so the sources are inserted
 * into a sync queue (iSyncQueue), which is a FIFO list of source ID.
 * Then, reading from the sync queue, the sources are loaded and the sync is effectively started.
 *
 * To cancel a synchronization process the method StopSync() is defined, which will also
 * cleanup the sync queue. To query the SyncEngine about the sync status,
 * use the method CheckSyncInProgress().
 */
class CSyncEngine : public CActive,
                    public MSyncMLProgressObserver, public MSyncMLEventObserver
{

private:

    /// Pointer to the object that created this instance, the appUI.
    /// Used to callback methods on sync events notifications.
    CFunambolAppUi*             iAppUI;

    RSyncMLSession              iSyncMLSession;
    TSmlProfileId               iActiveProfileID;

    /**
     * Used for scheduled (and delayed) synchronizations.
     */
    CSyncScheduler*              iScheduler;

    /**
     * The current source ID that is under sync.
     * DON'T change this value, as it's always trusted to update the right UI pane!
     */
    TInt                        iCurrentSource;

    ///The current syncML status: one of TCurrentSyncStatus.
    TCurrentSyncStatus          iCurrentSyncStatus;


    /// Counter of total items received (Server -> Client) for this job.
    /// It's resetted when the job starts (see ManageNewJobStarted())
    TInt                        iTotalClientItems;

    /// Counter of total items sent (Client -> Server) for this job.
    /// It's resetted when the job starts (see ManageNewJobStarted())
    TInt                        iTotalServerItems;

    /**
     * List of queued sync sources to sync.
     * We actively manage the queue of sources to sync, starting them one by one.
     * When startSync is called, the corresponding sourceID is added to
     * this list (if not already there). When sync() is triggered, the first
     * source in this list will be synchronized.
     * Canceling sync will clear this list too.
     */
    RArray<TInt>                iSyncQueue;

    /// Used to avoid manipulating the iSyncQueue
    /// member between different threads.
    RCriticalSection            iCriticalSyncQueue;

    ListenerPool*               iListenerPool;

    // It's set to true at startup, to know if it's the first sync done.
    // If so, the config sync will be added to the sync queue silently.
    bool firstSyncAfterStartup;

    // If true, the config sync running is hidden to the user
    // (the next source pane is updated with "connecting...")
    bool hiddenConfigSync;

    /**
     * It's set when a pictures sync ends, to remind the last error code.
     * It's reset if the app is restarted (it's not stored to config!).
     */
    int lastPicturesSyncError;

    /**
     * Enumeration of asynchronous SyncML actions.
     * These are actions that can be executed asynchronously, using
     * the active object RunL().
     */
    enum TAsyncAction
    {
        EIdle,
        ESyncFromQueue,
        EStopSync,
        EEnableScheduler
    };

    /**
     * The requested asynchronous SyncML action.
     * The active object will check this attribute to know
     * which action need to be executed, one of TAsyncAction.
     */
    TAsyncAction               iAsyncAction;

    /**
     * The ID of the thread that owns the RSyncMLSession.
     * @note  It's mandatory to originate and complete an asynchronous
     *        request inside the same thread, so here is the ID of the sole
     *        thread used to originate and complete requests.
     */
    TInt                        iMainThreadID;

    /// Imap email engine (for email synchronization)
    CImapEmailEngine*           emailEngine;


    /// Counter used when the SyncML profile is not found (see getActiveProfileID())
    /// Note: internal use, do not modify its value!
    TInt                        iCreateProfileRetry;

    /// If true, the sync engine is in the connecting phase just before a sync.
    /// Internal use, this phase is considered part of the sync (see checkSyncInProgress).
    bool                        iConnectingBeforeSync;

private:

    /// Default constructor
    CSyncEngine();

    /**
     * Initialize members, adds this object to the Active Scheduler.
     * Calls InitializeSession() to init SyncML session.
     * @param aAppUI  the parent CFunambolAppUi, used for callbacks
     */
    void ConstructL(CFunambolAppUi* aAppUI);

    /**
     * Initializes the SyncML session, opening RSyncMLSession object.
     * Register this class to receive SyncML events notifications and
     * loads the default profile.
     * @note  may leave with ERR_PROFILE_NOT_FOUND or other Symbian error codes
     */
    void InitializeSessionL();

    bool ReInitializeSessionL();

    /// Sets the iCurrentSource member
    void SetCurrentSource(const TInt aSourceID) { iCurrentSource = aSourceID; }

    /**
     * Returns the datasource name for a desired source.
     * It just scans a static struct.
     * @param aSourceID  the source ID       (e.g. SOURCE_CONTACTS)
     * @return           the datasource name (e.g. "C:Contacts.cdb")
     */
    TBuf<32>   GetDataSourceFromSourceID(const TInt aSourceID);

    /**
     * Returns the sourceID from the datasource name.
     * It just scans a static struct.
     * @param aDataSource  the datasource name (e.g. "C:Contacts.cdb")
     * @return             the source ID       (e.g. SOURCE_CONTACTS)
     */
    TInt   GetSourceIDFromDataSource(const TDesC& aDataSource);

    /**
     * Returns the taskID corresponding to a source ID.
     * The taskID can change on different devices, so we retrieve it
     * scanning all tasks from the active SyncML profile.
     * @param aSourceID  the source ID  (e.g. SOURCE_CONTACTS)
     * @return           the task ID, KErrNotFound if task not found
     */
    TSmlTaskId GetTaskIDFromSourceID(const TInt aSourceID);

    /**
     * Returns the sourceID corresponding to a taskID.
     * The taskID can change on different devices, so we retrieve it
     * scanning all tasks from the active SyncML profile.
     * @param aTaskID  the task ID
     * @return         the source ID (SOURCE_CONTACTS if not found)
     */
    TInt GetSourceIDFromTaskID(const TSmlTaskId aTaskID);

    /**
     * Checks if a sync job is currently running, and returns its ID.
     * @retun  the job ID of the running sync
     *         '-1' if no sync job is running ('0' is a valid job ID!)
     */
    TSmlJobId GetCurrentJobID();

    /**
     * Called when a new sync job (PIM) effectively starts, it handles
     * the beginning of a new source synced.
     * @param aJobID  ID of the job jist started
     */
    void ManageNewJobStarted(const TSmlJobId aJobID);

    /**
     * Stops the running job (PIM), by its ID.
     * @return  0 : no errors
     *         -1 : error opening job
     *         -2 : error stopping job
     */
    TInt StopJob(const TSmlJobId aJobID);


    /**
     * Appends a new SourceID into the sync queue array (iSyncQueue).
     * @param aSourceID  the source ID
     * @return           the new number of queued sources
     */
    TInt AddToSyncQueue(const TInt aSourceID);

    /**
     * Inserts the config syncsource in front of the syncQueue (index 0).
     * The insertion is NOT done if the source config or pictures is already
     * found inside the sync queue (it would be unnecessary).
     * @return  true if the source is inserted, false if not
     */
    bool InsertConfigInSyncQueue();

    /**
     * Inserts the syncsource in the syncQueue, index = pos.
     * The insertion is NOT done if the source is already
     * found inside the sync queue.
     * @return  true if the source is inserted, false if not
     */
    bool InsertInSyncQueue(const TInt aSourceID, const TInt pos = 0);

    /// Clears the sync queue list.
    void resetSyncQueue();


    /// Returns true if the passed source ID is found in the iSyncQueue list.
    TBool findSourceIDInSyncQueue(const TInt aSourceID);


    /**
     * Returns true if that source ID is currently under sync and in 'receiving' state.
     * It's a workaround for ClientPush, as we need to ignore the CP
     * notifications that happen during a sync, when items are added from
     * the Server. Otherwise, a new sync will be started every time when
     * the current sync is ended.
     */
    TBool isSourceUnderSyncAndReceiving(const TInt aSourceID);


    // ----------- Active Object's methods ------------
    /**
     * Used to request an asynchronous action to be executed.
     * Sets the 'iAsyncAction' member and signals that the asynchronous
     * request associated with 'iStatus' is complete, so that the
     * active object's RunL() will be called by active scheduler.
     * @note   we must signal the status in the same thread that originated the request.
     *
     * @param aAction  the action to be executed by the active object, one of TAsyncAction
     */
    void RequestAsyncAction(const TAsyncAction aAction);

    /**
     * Handles active object's request completion events.
     * It's called by the active scheduler when a request completion event occurs.
     * The asynchronous action to be executed is defined by 'iAsyncAction' member,
     * which is previously set when the request is raised (in RequestAsyncAction()).
     *
     * The active object is used to execute asynchronous actions like StartSync()
     * inside the same thread that created the SyncML session - it's used to start
     * synchronizations from push or client push.
     */
    void RunL(void);

    /**
     * Implements cancellation of  outstanding request.
     * This function is called as part of the active object's Cancel().
     */
    void DoCancel(void);

    /// Called by the active scheduler if the active object's RunL() function leaves.
    TInt RunError(TInt aError);
    // ---------------------------------------

    /**
     * Triggers a sync for the first source in the FIFO sync queue (iSyncQueue[0]
     * is the source triggered). If the queue is empty, nothing is done.
     * This method will set the iCurrentSource member and reset listener counters.
     * Note: if a sync is already in progress, nothing is done.
     * Note: this method is able to detect if current thread is the main thread,
     *       and in case not will recall itself from the main thread.
     */
    void syncFromQueue();

    /**
     * Triggers a sync for a PIM source, using native Nokia SyncML Client.
     * Gets the corresponding syncml task from the active profile, and starts
     * the new sync job.
     * @param aSourceID  the PIM source ID to start (one of SOURCE_CONTACTS, SOURCE_CALENDAR, SOURCE_NOTES)
     */
    void syncPIMSource(const TInt aSourceID);

    /**
     * Triggers a sync of emails (IMAP).
     * Uses the ImapEmailEngine object.
     */
    TInt syncEmail();

    /**
     *  Triggers a sync of configuration.
     * This is performed using Funambol APIs.
     */
    void syncConfig();

    /**
     * Triggers a sync of pictures.
     * This is performed using Funambol APIs.
     */
    void syncPhotos();


    TLastSyncStatus GetLastPimSyncStatus(const TInt aSourceID,
                                         TTime* aLastSyncTime,
                                         TInt* aLastSyncResult);

    TLastSyncStatus GetLastEmailSyncStatus(TTime* aLastSyncTime,
                                           TInt*  aLastSyncResult);
    TLastSyncStatus GetLastPhotosSyncStatus(TTime* aLastSyncTime,
                                               TInt*  aLastSyncResult);

    /**
     * Returns the corresponding syncType enumeration from the text decribing it.
     * In config, syncType is stored as text.
     * For example: "one-way-client" -> ESmlOneWayFromClient.
     */
    TSmlSyncType getSyncTypeFromText(const StringBuffer& text);

	/**
	 * Proper actions when a SyncML error is captured.
	 * Filters errors by type, and calls appUI->DisplayMessage() for some of them.
	 */
    void ManageSyncMLErrors(const TInt aError, const TInt aAdditionalData);


    /**
     * This method checks if we are at the end of a sync, and in that case it
     * starts the UpgradeManager to check if there is a new version available.
     * The check is skipped in case of a config-sync (usually login sync).
     */
    void checkForUpdates();

    /**
     * Called at the end of each sync, to eventually enable/hide some sources.
     * Actually enables/disables pictures source (check Server datastores).
     * @return true if something changed in UI
     */
    bool refreshSourcesVisible();

    /**
     * If sync queue not empty, updates the next source pane with the passed sync status.
     * It's used to hide to the user a source syncing (like the config sync on every first sync).
     */
    void updateNextSourceStatus(const TCurrentSyncStatus aSyncStatus);


protected:

    /*
     * SyncML ClientAPI Observers
     */
    void OnSyncMLSyncProgress(TStatus aStatus, TInt aInfo1, TInt aInfo2);
    void OnSyncMLSyncError(TErrorLevel aErrorLevel, TInt aError, TInt aTaskId,
                           TInt aInfo1, TInt aInfo2);
    void OnSyncMLDataSyncModifications(TInt aTaskId,
                                       const TSyncMLDataSyncModifications &aClientMod,
                                       const TSyncMLDataSyncModifications &aServerMod);

    // From MSyncMLEventObserver
    void OnSyncMLSessionEvent(TEvent aEvent, TInt aIdentifier, TInt aError,
                              TInt aAdditionalData);


public:

    static CSyncEngine* NewL(CFunambolAppUi* aAppUI);
    static CSyncEngine* NewLC(CFunambolAppUi* aAppUI);

    /// Destructor
    ~CSyncEngine();


    /**
     * Starts a new sync job, for an array of sources.
     * Will call the StartSyncL(sourceID) for each source: the first
     * First task starts immediately, other tasks are queued (see iSyncQueue).
     *
     * @param aSourceIDList  the array of source IDs to be synced
     * @param from           defines who triggered the sync action, one of TSyncRunners
     */
    void StartSync(RArray<TInt> &aSourceIDList, const TSyncRunners from);

    /**
     * Starts a new sync job, only for 1 desired source.
     * @see StartSync(const RArray<TInt>&, const TSyncRunners);
     * @param aSourceID  the source ID of the desired source (e.g. SOURCE_CONTACTS)
     */
    void StartSync(const TInt aSourceID, const TSyncRunners from);

    /**
     * If a sync is running, tries to abort the current sync.
     * Also clears the sync queue.
     * @return  1 if current sync stopped
     *          0 if no job running
     *         <0 in case of errors stopping the current job
     */
    TInt StopSync();

    /**
     * Checks if a sync is currently running.
     * @retun  Etrue if a sync is running, EFalse if not.
     */
    TBool CheckSyncInProgress();

    /// Returns true if the sync queue is empty.
    bool isSyncQueueEmpty();

    /**
     *  Returns the source ID currently under sync.
     * It's important to know which is the task running, for example
     * to refresh the correct pane corresponding to it.
     */
    TInt GetCurrentSourceID() { return iCurrentSource; }

    /**
     * Called by SyncActiveObject when a sync has ended (using Funambol APIs).
     * Will updade the source status and UI pane.
     * At the end will call syncFromQueue() in order to trigger the next
     * sync, or (if this is the last source in queue) calls checkForUpdate().
     * @param itemsSuccesfullySynced  the number of items synced correctly by this sync session
     */
    void syncEnded(const TInt aSourceId, TInt status, const int itemsSuccesfullySynced);

    /**
     * Called by ImapEmailEngine when a sync of emails has ended.
     * Will updade the source status and UI pane.
     * At the end will call syncFromQueue() in order to trigger the next
     * sync, or (if this is the last source in queue) calls checkForUpdate().
     */
    void EmailSyncEnded(TCurrentSyncStatus status);


    /**
     * Opens the SyncML history and finds the last sync status for the desired source.
     * If the source was never synchronized, returns ELastSyncNotFound
     * (aLastSyncTime and aLastSyncResult are not touched).
     * Otherwise on exit 'aLastSyncTime' param will contain the last timestamp
     * (in local time) and 'aLastSyncResult' the last sync result code.
     * This method scans the RSyncMLHistoryLog for the active profile,
     * starting from the most recent entry and searching for one that
     * contains the desired task (corresponding to the passed sourceID).
     *
     * @param aSourceID         the source ID of the desired sorce to look for
     * @param aLastSyncTime     [OUT] the last sync timestamp, as a TTime object (in local time)
     * @param aLastSyncResult   [OUT] the last sync result code
     * @return                  the last sync status, one of TLastSyncStatus.
     */
    TLastSyncStatus GetLastSyncStatus(const TInt aSourceID, TTime* aLastSyncTime,
                                      TInt* aLastSyncResult);

    /// Proxy method to iAppUI->updateSourceStatus()
    void updateAppUISourceStatus(const TInt aSourceID, const TCurrentSyncStatus aSyncStatus, const TInt aSyncData = 0);

    /// Returns true if the desired source is enalbed.
    const bool isSourceEnabled(const TInt aSourceID);

    /// Sets the 'enable' flag in the corresponding syncsource
    void enableSource(const TInt aSourceID, const bool aEnable = true);

    /// Also saves the config.
    void setSourceTimestamp(const TInt aSourceID, const TUint timestamp);


    /**
     * Enables the sync scheduler.
     * Sets the scheduler to load the list of active sources every time itself,
     * then enables the scheduler with the given interval.
     */
    void enableScheduler(const bool aIsRepetitive, const unsigned int aInterval);

    /**
     * Disables the sync scheduler (proxy method).
     * If a scheduled sync is already set, it will not complete because
     * it will be discarded on RunL().
     */
    void disableScheduler();

    /// Checks the sync scheduler (proxy method).
    bool isSchedulerEnabled();

    /// Removes a sourceID from the scheduler list (proxy method).
    bool removeFromScheduler(const TInt aSourceID);


    // Proxy method, used by the scheduler.
    RArray<TInt> GetActiveSourcesList();

    /**
     * Get the active pim sync profile (-1 if it does not exist)
     */
    TInt getActiveProfileID();

    /**
     * hook for Funambol API listeners
     */
    void onListenerStatusUpdate(const TInt aSourceID, const TCurrentSyncStatus aSyncStatus, const TInt aSyncData=0);

    /// Proxy method, to call FunambolAppUI::DisplayMessage()
    void DisplayMessage(const TInt aMsgID, const TWarningLevel aWarningLevel);
};

#endif  // SYNCENGINE_H_
