/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2008 Funambol, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 *
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */


#ifndef SYNC_SCHEDULER
#define SYNC_SCHEDULER

#include "base/util/StringBuffer.h"
#ifdef __GCCE__
#include "SyncEngine.h"
#else
#include "FakeSyncEngine.h"
#endif


/// Default scheduling time, in seconds
#define SCHEDULER_DEFAULT_INTERVAL      300


using namespace Funambol;


/**
 * This class is responsible of the scheduled synchronizations.
 * ...
 */
class CSyncScheduler : public CActive {

private: 
    
    /// Pointer to the object that owns and created this object, the SyncEngine.
    /// Used to callback methods to start scheduled sync.
    CSyncEngine*        iSyncEngine;
    
    /// If true, the Scheduler is enabled.
    bool                enabled;
    
    /// If true, a new StartSync will be rescheduled forever, 
    /// keeping the sourceList as it is.
    bool                repetitive;
    
    /**
     * If true, we load the source list reading active source list
     * everytime a scheduled sync shoud start. This is used by the scheduled service.
     */
    bool                useActiveSourceList;
    
    /// The next scheduling time
    TTime               iNextSchedulingTime;
    
    /// Timer, used to trigger the active object RunL at the scheduling time.
    RTimer              iTimer;
    
    /// Used to avoid manipulating the iSourceList 
    /// member between different threads.
    RCriticalSection    iCriticalSourceList;
    
    /// The interval used to schedule the next StartSync.
    unsigned int        interval;

    /**
     * A list of source IDs, that will be synced once the
     * scheduling time occurrs.
     */
    RArray<TInt>        iSourceList;
    
    
    /// Constructor
    CSyncScheduler();
    
    void ConstructL(CSyncEngine* aSyncEngine);
    
    /// Reschedule a new sync after 'interval' seconds, using the internal timer.
    void reschedule();
    
    /**
     * In case of CLient-push, updates the 'interval' to follow a retry policy:
     * 1min -> 5min -> 15min and then 15min forever.
     */
    void updateInterval();
    
    /**
     * Called once the scheduling time is reached.
     * Will call the iSyncEngine::StartSyncL(), and then reschedule the
     * next sync if the scheduler is repetitive.
     */
    void RunL();
    
    void DoCancel();
    
    TInt RunError(TInt aError);

    
public:
    
    static CSyncScheduler* NewL(CSyncEngine* aSyncEngine);
    static CSyncScheduler* NewLC(CSyncEngine* aSyncEngine);
    

    
    /// Destructor
    ~CSyncScheduler();
    
    /**
     * Call this method to enable the scheduler.
     * A StartSync action will be scheduled after 'interval' seconds.
     * If another scheduled sync is already set (earlier than this one), 
     * this scheduling will be discarded as it's not necessary.
     */
    void enable();
    void enable(const bool aIsRepetitive);
    void enable(const bool aIsRepetitive, const unsigned int aInterval);
    
    /**
     * Call this method to disable the scheduler.
     * If a scheduled sync is already set, it will not complete because
     * it will be discarded on RunL().
     */
    void disable();
    
    /**
     * Sets the array of source IDs, that will be used for the scheduled sync.
     * @param aSourceList  the array of source IDs
     * @param merge        if true, the source list will be merged with the existing (internal)
     *                     source list, so we don't loose previous source IDs. Default = true.
     */
    void setSourceList(const RArray<TInt>& aSourceList, const bool merge = true);
    
    /**
     * Removes the passed sourceID from the internal array iSourceList.
     * @param aSourceID the sourceID to be searched and removed
     * @return          true if found and removed
     */
    bool removeFromSourceList(const TInt aSourceID);
    
    void clearSourceList();
    
    
    void setRepetitive(const bool val)          { repetitive = val; }
    void setInterval(const unsigned int val)    { interval = val;   }
    void setUseActiveSourceList(const bool val) { useActiveSourceList = val; }
    
    
    RArray<TInt> getSourceList()    { return iSourceList;         }
    TTime GetNextSchedulingTime()   { return iNextSchedulingTime; } 
    bool isRepetitive()             { return repetitive;          }
    bool isEnabled()                { return enabled;             }
    unsigned int getInterval()      { return interval;            }
    
    
    StringBuffer formatNextSchedulingTime();
    unsigned int getSecondsToNextSchedulingTime();

};

#endif  // SYNC_SCHEDULER
