/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2008 Funambol, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 *
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */


#ifndef INCL_SYMBIAN_CLIENT_UTILS
#define INCL_SYMBIAN_CLIENT_UTILS

#include <gulicon.h>
#include <akniconutils.h>
#include <pathinfo.h>
#include "base/fscapi.h"
#include "base/util/StringBuffer.h"
#include "SyncEngine.h"

using namespace Funambol;


/**
 * Used to quicky check if a specific source is visible or not.
 * It checks the sourcesVisible array.
 * Defined in ClientSettings.cpp.
 */
bool isSourceVisible(const TInt aSourceID);
bool isSourceVisible(const char* sourceName);

/**
 * Used to check the total number of sources visible.
 * It checks the sourcesVisible array.
 * Defined in ClientSettings.cpp.
 */
int countSourceVisible();


TInt         getSourceIDFromName(const StringBuffer& name);
StringBuffer getSourceNameFromID(const TInt     aSourceID);

// Used for debugging
StringBuffer getSyncRunnerName(const TSyncRunners syncRunner);

// Used for scheduler option list.
TInt getSchedulerIndexFromMinutes     (const TInt minutes);
TInt getSchedulerResourceIDFromMinutes(const TInt minutes);
TInt getMinutesFromSchedulerIndex     (const TInt index);
TInt getSchedulerResourceIDFromIndex  (const TInt index);

/**
 * If a connection is not already active, we start a new one (using FConnection).
 * In case of errors, retry 'maxNumRetries' times, asking the user for the APN.
 * Called before starting a sync, or before starting the push service.
 * @param maxNumRetries  max number of times we can ask the user to select a new APN
 * @return               true if connected, false if could not connect
 */
bool startConnection(const TInt maxNumRetries);

/**
 * Safely closes the active connection, if the current connection is not
 * used by some other client.
 * Called at the end of every sync session, if push is not used.
 * @return  true if connection closed
 */
bool closeConnection();

/**
 * Loads an icon and appends it into an array of icons 'icons'.
 * @param icons  [OUT] the array of icons, to append the new icon
 * @param bmpID  the identifier of the icon (mbm or mif)
 * @param maskID the identifier of the mask, optional
 */
void appendIcon(CArrayPtr<CGulIcon>* icons, const TInt bmpID, TInt maskID = 0);

/**
 * Returns an integer value rapresenting the build number read from
 * the string version passed (e.g. "6.1.12" -> 60112).
 */
int getBuildNumberFromVersion(const char* swv);

/**
 * Merges sourceArray into destArray.
 * Every value ont found in destArray is copied.
 * @return the new number of items in destArray
 */
int mergeRArray(RArray<TInt>& destArray, const RArray<TInt>& sourceArray);


/**
 * Safely removes an item (TInt) from the passed RArray<TInt>, if found.
 * @param itemToRemove  the integer item to remove
 * @param array         the RArray to search into
 * @return              true if the item was found (and removed), false if not
 */
bool removeFromRArray(const TInt itemToRemove, RArray<TInt>& array);

/**
 * Returns the default path where images are stored, under the desired drive.
 * If drive 'E' is requested but it's not available (memory-card not existing) we
 * fallback to phone-memory 'C'.
 * All paths have the trailing backslash included.
 *
 * @param drive       the requested drive letter ("C" or "E")
 * @param imagesPath  [OUT] the path of images, like "C:\Data\Images\").
 * @return            true if the path is under the requested drive letter,
 *                    false in case of fallback to C:
 */
bool getImagesPath(TInt aStorageType, StringBuffer& imagesPath);

StringBuffer getSourceNameFromServerURI(const StringBuffer& uri);


/**
 * Checks if a precess is running by SID
 *
 * @param aProcessUid   the UID of the process to be checked
 * @return              ETrue if the process is found
 *                      EFalse if no process with such uid is found
 */
TBool CheckProcessRunning(const TUid& aProcessUid);

/**
 * Starts application by name (server or console - see
 * http://wiki.forum.nokia.com/index.php/How_to_start_and_stop_exe)
 */
void StartServerProcessL(const TDesC& aServerName);

/**
 * Starts a process via RProcess
 */
void StartProcessL(const TDesC& aExe, const TDesC& aCmdLine);

/**
 * Starts View-based process via StartDocument
 */
void StartViewProcessL(const TUid& aUid, const TDesC& aDocument);


/**
 * Starts an application from its UID via StartDocument() or
 * brings it to forground if already running
 */
void openAppL(const TUid& aAppUid, const TDesC& aDocument);

/**
 * Starts (or bring to foreground if yet running) a Midlet
 *
 * @param aMidletName   the name of the midlet to be started
 * @return              ETrue if the midlet was found
 *                      EFalse if no midlet with that name is found
 */
TBool startMidletL(const TDesC& aMidletName);

/**
 * Converts a string value into a 64bit integer value (long int).
 * Needs special conversions to manipulate 64bit values.
 */
TUint get64bitValue(const StringBuffer& stringValue);

/**
 * Check if mass memory storage drive is available on the device
 */
TBool massStorageMemoryAvailable(const RFs& aFsSession);
TBool massStorageMemoryAvailable();

/**
 * Check if memory card is available on the device
 */
TBool memoryCardAvailable(const RFs& aFsSession);
TBool memoryCardAvailable();

/**
 *  convert a storage specifier of type TPicturesStorageType to
 *  the corresponding drive (C:, E: or F:)
 */
const StringBuffer getDriveName(TInt aStorageType);

#endif /*INCL_SYMBIAN_CLIENT_UTILS*/
