/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2008 Funambol, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 *
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

#include <e32base.h>
#include <e32cmn.h>

#include "FunambolAppUi.h"
#include "CleanupCalendarManager.h"

/**
 * CCleanupCalendarManager::NewL()
 * Two-phased constructor.
 */
CCleanupCalendarManager* CCleanupCalendarManager::NewL(CFunambolAppUi* appUi)
{
    CCleanupCalendarManager* self = CCleanupCalendarManager::NewLC(appUi);
    CleanupStack::Pop(self);
    return self;
}

/**
 * CCleanupCalendarManager::NewLC()
 * Two-phased constructor.
 */
CCleanupCalendarManager* CCleanupCalendarManager::NewLC(CFunambolAppUi* appUi)
{
    CCleanupCalendarManager* self = new(ELeave) CCleanupCalendarManager(appUi);
    self->ConstructL();
    CleanupStack::PushL(self);
    return self;
}

/**
 * CCleanupCalendarManager::CCleanupCalendarManager()
 * C++ default constructor can NOT contain any code, that might leave.
 */
CCleanupCalendarManager::CCleanupCalendarManager(CFunambolAppUi* appUi)
                                : CActive(CActive::EPriorityStandard),
                                  appUi(appUi),
                                  iCanceled(EFalse)
{
}

/**
 * CCleanupCalendarManager::ConstructL()
 * Symbian 2nd phase constructor can leave.
 */
void CCleanupCalendarManager::ConstructL()
{
    // Connect to calendar server
    iCalSession = CCalSession::NewL();
    // Open default calendar file
    iCalSession->OpenL(KNullDesC);
    // adds itself to the scheduler
    CActiveScheduler::Add(this);
}

/**
 * CCleanupCalendarManager::~CCleanupCalendarManager()
 * Class destructor
 */
CCleanupCalendarManager::~CCleanupCalendarManager()
{
    if( IsActive() ) {
        Cancel();
    }

    delete iCalIter;
    iCalIter = NULL;
    delete iCalEntryView;
    iCalEntryView = NULL;
    delete iCalSession;
    iCalSession = NULL;
}

/**
 * CCleanupCalendarManager::DeleteAllEntriesL()
 */
void CCleanupCalendarManager::DeleteAllEntriesL()
{
    // Open a view
    // Opening a view takes some time and if it is not ready,
    // every operation on it fails; when the view is ready,
    // Completed() is called and we can proceed
    iCalEntryView = CCalEntryView::NewL(*iCalSession,*this);
}

/*
 * Progress()
 * From MCalProgressCallBack
 */
void CCleanupCalendarManager::Progress(TInt aPercentageCompleted)
{
    
}

/*
 * Completed()
 * From MCalProgressCallBack
 */
void CCleanupCalendarManager::Completed(TInt aError)
{
    if(aError==KErrNone) {
        // view is open, we can operate on it
        ReallyDeleteAllEntriesL();
    } 
    else {
        // TODO: notify user or quietly end 
    }
}

/*
 * NotifyProgress()
 * From MCalProgressCallBack
 */
TBool CCleanupCalendarManager::NotifyProgress(void)
{
    return ETrue;
}

void CCleanupCalendarManager::ReallyDeleteAllEntriesL()
{
    LOG.info("Cleaning up all Calendar & tasks...");
    
    iCalIter = CCalIter::NewL(*iCalSession);
    iEntryUid.Zero();
    iEntryUid = iCalIter->FirstL();
    // start active scheduler activity
    TRequestStatus* status = &iStatus;
    User::RequestComplete(status, KErrNone);
    SetActive();
}

void CCleanupCalendarManager::Stop(void)
{
    iCanceled = ETrue;
}

void CCleanupCalendarManager::DoCancel(void)
{
    iCanceled = ETrue;
}

void CCleanupCalendarManager::RunL(void)
{
    if(iCanceled)
    {
        //operation has been stopped
        appUi->calendarCleanupCompleted();
    } else if(iEntryUid == KNullDesC8) {
        // last element in contact database (or empty db)
        // nothing to be deleted
        appUi->calendarCleanupCompleted();
    } else {
        // operation was not canceled and the item is valid
        // get entries associated with this UID
        RPointerArray<CCalEntry> calEntryArray;
        iCalEntryView->FetchL(iEntryUid, calEntryArray);
        CCalEntry* calEntry = calEntryArray[0];
        iCalEntryView->DeleteL(*calEntry);
    
        iEntryUid.Zero();
        // usually we should call iCalIter->NextL()... but this has
        // proven not to work very well... some item is left and
        // not canceled; since all we do is to cancel an item, calling again
        // iCalIter->FirstL is equivalent, and, most important, it works! :-)
        iEntryUid = iCalIter->FirstL();
        calEntryArray.ResetAndDestroy();

        TRequestStatus* status = &iStatus;
        User::RequestComplete(status, KErrNone);
        SetActive();
    }
}

