/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2009 Funambol, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 *
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

#include <e32base.h>
#include "base/Log.h"

#include "FunambolAppUi.h"
#include "CleanupPhotosManager.h"
#include "customization.h"
#include "ClientSettings.h"

#include "base/util/StringBuffer.h"
#include "base/util/symbianUtils.h"

#include "PicturesSyncSource.h"
#include "utils.h"

/**
 * CCleanupPhotosManager::NewL()
 * Two-phased constructor.
 */
CCleanupPhotosManager* CCleanupPhotosManager::NewL(CFunambolAppUi* aAppUi)
{
    CCleanupPhotosManager* self = CCleanupPhotosManager::NewLC(aAppUi);
    CleanupStack::Pop(self);
    
    return self;
}

/**
 * CCleanupPhotosManager::NewLC()
 * Two-phased constructor.
 */
CCleanupPhotosManager* CCleanupPhotosManager::NewLC(CFunambolAppUi* aAppUi)
{
    CCleanupPhotosManager* self = new(ELeave) CCleanupPhotosManager(aAppUi);
    CleanupStack::PushL(self);
    self->ConstructL();
    
    return self;
}

/**
 * CCleanupPhotosManager::CCleanupPhotosManager()
 * C++ default constructor can NOT contain any code, that might leave.
 */
CCleanupPhotosManager::CCleanupPhotosManager(CFunambolAppUi* aAppUi)
                                        : CActive(CActive::EPriorityStandard),
                                        iAppUi(aAppUi)
{
    iDone = EFalse;
    iCanceled = EFalse;
}

/**
 * CCleanupPhotosManager::ConstructL()
 * Symbian 2nd phase constructor can leave.
 */
void CCleanupPhotosManager::ConstructL()
{
    CActiveScheduler::Add(this);
}

/**
 * CCleanupPhotosManager::~CCleanupPhotosManager()
 * Class destructor
 */
CCleanupPhotosManager::~CCleanupPhotosManager()
{
    if( IsActive() ) {
        Cancel();
    }
}

/**
 * CCleanupPhotosManager::DeleteAllPhotosL()
 */
void CCleanupPhotosManager::DeleteAllPhotosL(void)
{
    TRequestStatus* status = &iStatus;
    User::RequestComplete(status, KErrNone);
    SetActive();
}

/**
 * CCleanupPhotosManager::CleanupPhotosFolderL()
 */
void CCleanupPhotosManager::CleanupPhotosFolderL(void)
{
    LOG.debug("Cleaning up all Photos...");
    
    // Retrieve configuration
    ClientSettings* config = ClientSettings::getInstance();
    
    // Get the pictures location folder
    // Continue only if the desired drive is available
    StringBuffer imagesDir;
    TInt picturesStorage = config->getPicturesStorage();
    bool driveAvailable = getImagesPath(picturesStorage, imagesDir);
    
    if (!driveAvailable) {
        bool fallback = false;
        TBool massStorageAvailable = massStorageMemoryAvailable();
        StringBuffer pictureDrive = getDriveName(picturesStorage);
        // if drive is become not available must be a memory card
        LOG.info("Pictures drive '%s' not available: popup query", pictureDrive.c_str());

        if (massStorageAvailable) {
            //FIXME: update resource string
            fallback = iAppUi->DisplayQuery(QUERY_DRIVE_NOT_AVAILABLE_CLEANUP);
        } else {
            fallback = iAppUi->DisplayQuery(QUERY_DRIVE_NOT_AVAILABLE_CLEANUP);
        }

        if (fallback) {
            // User selected YES: save the new drive letter to config and continue
            TInt storageType = EStorageTypeUnset;
            if (massStorageAvailable) {
                storageType = EMassMemoryStorage;
            } else {
                storageType = EPhoneMemoryStorage;
            }

            config->setPicturesStorage(storageType);

            pictureDrive = getDriveName(storageType);
            LOG.debug("Fallback to pictures location: %s", pictureDrive.c_str());
            config->save();
        } else {
            // User selected NO: don't cleanup
            LOG.debug("User selected no: aborting cleanup of pictures");
            goto finally;
        }
    }

    {
    // Retrieve sync source
    wchar_t* wSourceName = toWideChar(SOURCE_PHOTOS_NAME);
    PicturesSyncSource source(wSourceName, 
                              config->getSyncSourceConfig(SOURCE_PHOTOS_NAME),
                              imagesDir);
    
    source.removeAllItems();
    delete [] wSourceName;
    }

    
finally:
    iDone = ETrue;
    
    TRequestStatus* status = &iStatus;
    User::RequestComplete(status, KErrNone);
    SetActive();
}

void CCleanupPhotosManager::DoCancel(void)
{
    iCanceled = ETrue;
}


void CCleanupPhotosManager::RunL(void)
{
    if(iCanceled){
        // operation has been stopped
        iAppUi->photosCleanupCompleted();
    } else if(iDone){
        // operation completed
        iAppUi->photosCleanupCompleted();
    } else {
        CleanupPhotosFolderL();
    }
}

