/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2008 Funambol, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 *
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

#include <etel3rdparty.h>

#include "base/util/stringUtils.h"
#include "base/util/symbianUtils.h"
#include "client/DMTClientConfig.h"
#include "spdm/DMTree.h"
#include "push/CTPService.h"

#include "commonDefs.h"
#include "customization.h"
#include "utils.h"
#ifdef __GCCE__
#include "SyncProfileManager.h"
#else
#include "FakeSyncProfileManager.h"
#endif
#include "ClientSettings.h"
#include "DefaultSymbianConfigFactory.h"
#include "UpdateManager.h"

#ifdef FAKE_IMAP_ACCOUNT
#include "FakeImapAccountManager.h"
#else
#include "ImapAccountManager.h"
#endif

using namespace Funambol;

ClientSettings* ClientSettings::instance = NULL;

const StringBuffer ClientSettings::configContext("/Email");
const StringBuffer ClientSettings::upgradeContext("/upgrade");

const StringBuffer ClientSettings::addressNodeName("Address");
const StringBuffer ClientSettings::displayNodeName("DisplayName");
const StringBuffer ClientSettings::nodePropertyName("value");
const StringBuffer ClientSettings::pendingSyncConfigPropertyName("pending-sync-config");
const StringBuffer ClientSettings::syncUsingPropertyName("syncUsing");
const StringBuffer ClientSettings::lastEmailSyncTimestampPropertyName("last-email");
const StringBuffer ClientSettings::lastEmailSyncStatusPropertyName("last-email-status");
const StringBuffer ClientSettings::firstPicturesSyncTimestampPropertyName("first-pictures");
const StringBuffer ClientSettings::installTimestampPropertyName("install-tstamp");
const StringBuffer ClientSettings::includeOlderPicturesPropertyName("include-older-pictures");
const StringBuffer ClientSettings::lastPhotosSyncStatusPropertyName("last-photos-status");
const StringBuffer ClientSettings::schedulerMinutesPropertyName("schedulerMinutes");
const StringBuffer ClientSettings::cpIntervalPropertyName("cpInterval");
const StringBuffer ClientSettings::iapNamePropertyName("iapName");
const StringBuffer ClientSettings::pictureStorageTypePropertyName("picturesStorageType");
/// This property contains the Funambol software version (usually different from customer's swv)
const StringBuffer ClientSettings::clientSwVersionPropertyName("funambol_swv");
// old software version
const StringBuffer ClientSettings::oldSwVersionPropertyName("oldSwv");
const StringBuffer ClientSettings::pushTypePropertyName("push-type");
const StringBuffer ClientSettings::sourceOrderPropertyName("sourceOrder");


bool isSourceVisible(const TInt aSourceID) {
    
    StringBuffer sourceName = getSourceNameFromID(aSourceID);
    return isSourceVisible(sourceName.c_str());
}

bool isSourceVisible(const char* sourceName) {

    const ArrayList& sourcesVisible = ClientSettings::getInstance()->getSourcesVisible();
    
    for (int i=0; i<sourcesVisible.size(); i++) {
        StringBuffer* name = (StringBuffer*)sourcesVisible.get(i);
        if (*name == sourceName) {
            return true;
        }
    }
    return false;
}

int countSourceVisible() {
    const ArrayList& sourcesVisible = ClientSettings::getInstance()->getSourcesVisible();
    return sourcesVisible.size();
}


/**
 * This private class is needed to get the phone IMEI. Anyone would think that grabbing
 * the IMEI is a matter of few lines of code. But this is Symbian! You MUST have
 * an acrive object to get it working. Noone knows why, but this is how it is.
 */
class CIMEI : public CActive
{
public:
    CIMEI() : CActive(EPriorityStandard) {}

    // Construction
    void CIMEI::ConstructL() {
        iPkg = new (ELeave) CTelephony::TPhoneIdV1Pckg(iV1);
        iTelephony = CTelephony::NewL();
        CActiveScheduler::Add(this);
    }

    // Destruction
    ~CIMEI() {
        Cancel(); // if any request outstanding, calls DoCancel() to cleanup
        delete iTelephony;
        delete iPkg;
    }

    static CIMEI* NewL() {
        CIMEI* self =new(ELeave) CIMEI();
        CleanupStack::PushL(self);
        self->ConstructL();
        CleanupStack::Pop(self);
        return self;
    }

    // Issue request: retrieve IMEI
    void GetIMEI(TRequestStatus& aStatus) {
        aStatus=KRequestPending;
        iIMEIStatus = &aStatus;
        iTelephony->GetPhoneId( iStatus, *iPkg );
        SetActive();
        CActiveScheduler::Start();
    }

    void CIMEI::RunL() {
        if ( (iStatus == KErrNone) ) {
            retrievedIMEI = (*iPkg)().iSerialNumber; // Read IMEI from package buffer
        }
        CActiveScheduler::Stop();

        //Signal complete and pass back result of CActive object
        User::RequestComplete(iIMEIStatus,iStatus.Int());
    }

    void CIMEI::DoCancel() {
        Cancel();
    }

private:
    CTelephony* iTelephony; // telephony object we own
    CTelephony::TPhoneIdV1 iV1;
    CTelephony::TPhoneIdV1Pckg* iPkg;
    TRequestStatus *iIMEIStatus;

public:
    TBuf<50> retrievedIMEI;
};



ClientSettings* ClientSettings::getInstance() {
    if (instance == NULL) {
        instance = new ClientSettings();
    }
    return instance;
}

ClientSettings::ClientSettings() : DMTClientConfig(),
                                   emailAddress     (""),
                                   displayName      (""),
                                   pendingSyncConfig(false),
                                   syncUsing        (ESyncUsingPush),
                                   lastEmailSyncTimestamp(0),
                                   lastEmailSyncStatus(ELastSyncNotFound),
                                   firstPicturesSyncTimestamp(0),
                                   installTimestamp(0),
                                   includeOlderPictures(false),
                                   lastPhotosSyncStatus(ELastSyncNotFound),
                                   schedulerMinutes (30),
                                   cpInterval       (300),
                                   iapName          (""),
                                   dirty            (false),
                                   nativeDirty      (false),
                                   copiedAccountFromProfile(false),
                                   pushType         (DEFAULT_PUSH_TYPE),
                                   pictureStorageType(EStorageTypeUnset)          
{
}



StringBuffer ClientSettings::getImeiL() {

    // Get the phone IMEI
    CIMEI* imei = CIMEI::NewL();
    CleanupStack::PushL(imei);

    TRequestStatus status;
    imei->GetIMEI(status);

    // wait for request to complete
    User::WaitForRequest(status);

    // Check if IMEI retrieved correctly
    StringBuffer imeiSb;
    if (status==KErrNone) {
        TBuf<50> imeiStr(imei->retrievedIMEI);
        imeiSb = bufToStringBuffer(imeiStr);
    }
    CleanupStack::PopAndDestroy();

    return imeiSb;
}

StringBuffer ClientSettings::retrieveDeviceId()
{
    StringBuffer newDevId("");
    StringBuffer imei;
    TRAPD(err, imei = getImeiL(); );

    if (err == KErrNone) {
        // We got the IMEI
        newDevId = "IMEI:";
        newDevId += imei;
        LOG.debug("New device ID=%s", newDevId.c_str());
    }
    else {
        LOG.error("Cannot grab phone IMEI, the device Id is left emtpy");
    }
    return newDevId;
}


//
// --------------- Set methods ---------------
//
void ClientSettings::setPendingSyncConfig(const bool aPendingSyncConfig) {
    if (aPendingSyncConfig != pendingSyncConfig) {
        dirty = true;
        pendingSyncConfig = aPendingSyncConfig;
    }
}

void ClientSettings::setSyncUsing(const TSyncUsing aSyncUsing) {
    if (aSyncUsing != syncUsing) {
        dirty     = true;
        syncUsing = aSyncUsing;
    }
}

void ClientSettings::setLastEmailSyncTimestamp(TInt64 aTime) {
    if (aTime != lastEmailSyncTimestamp) {
        dirty = true;
        lastEmailSyncTimestamp = aTime;
    }
}

void ClientSettings::setLastEmailSyncStatus(TLastSyncStatus aStatus) {
    if (aStatus != lastEmailSyncStatus) {
        lastEmailSyncStatus = aStatus;
        dirty               = true;
    }
}

void ClientSettings::setFirstPicturesSyncTimestamp(TInt64 aTime, const bool storeNow) {
    if (aTime != firstPicturesSyncTimestamp) {
        dirty = true;
        firstPicturesSyncTimestamp = aTime;
        if (storeNow) {
            save();
        }
    }
}

void ClientSettings::setInstallTimestamp(TInt64 aTime) {
    if (aTime != installTimestamp) {
        dirty = true;
        installTimestamp = aTime;
    }
}

void ClientSettings::setIncludeOlderPictures(const bool aInclude) {
    if (aInclude != includeOlderPictures) {
        dirty = true;
        includeOlderPictures = aInclude;
    }
}

void ClientSettings::setLastPhotosSyncStatus(TLastSyncStatus aStatus) {
    if (aStatus != lastPhotosSyncStatus) {
        lastPhotosSyncStatus = aStatus;
        dirty               = true;
    }
}

void ClientSettings::setSchedulerMinutes(const TInt aSchedulerMinutes) {
    if (schedulerMinutes != aSchedulerMinutes) {
        schedulerMinutes = aSchedulerMinutes;
        dirty = true;
    }
}

void ClientSettings::setCpInterval(const TInt aCpInterval) {
    if (cpInterval != aCpInterval) {
        cpInterval = aCpInterval;
        dirty      = true;
    }
}

void ClientSettings::setPushType(const TInt aPushType) {
    if (pushType != aPushType) {
        pushType = aPushType;
        dirty = true;
    }
}

void ClientSettings::setIapName(const StringBuffer& aIapName) {
    if (iapName != aIapName) {
        iapName = aIapName;
        dirty   = true;
        nativeDirty = true;
    }
}

void ClientSettings::setDisplayName(const StringBuffer& aDisplayName)
{
    if (displayName != aDisplayName) {
        displayName = aDisplayName;
        dirty   = true;
        nativeDirty = true;
    }
}
void ClientSettings::setEmailAddress(const StringBuffer& aEmailAddress)
{
    if (emailAddress != aEmailAddress) {
        emailAddress = aEmailAddress;
        dirty   = true;
        nativeDirty = true;
    }
}

void ClientSettings::setPassword(const StringBuffer& aPassword) {
    const StringBuffer& password = getAccessConfig().getPassword();
    if (password != aPassword) {
        getAccessConfig().setPassword(aPassword);
        dirty = true;
        nativeDirty = true;
    }
}

void ClientSettings::setSyncURL(const StringBuffer& aSyncURL) {
    const StringBuffer& syncURL = getAccessConfig().getSyncURL();
    if (syncURL != aSyncURL) {
        getAccessConfig().setSyncURL(aSyncURL);

        // Also sw-update will reflect the new syncUrl.
        UpdateManager updateManager(NULL);
        updateManager.setCheckURL(aSyncURL);

        dirty = true;
        nativeDirty = true;
    }
}

void ClientSettings::setUsername(const StringBuffer& aUsername) {
    const StringBuffer& username = getAccessConfig().getUsername();
    if (username != aUsername) {
        getAccessConfig().setUsername(aUsername);
        dirty = true;
        nativeDirty = true;
    }
}

void ClientSettings::setPicturesStorage(const TInt aPicturesStorage) {
    if (pictureStorageType != aPicturesStorage) {
        pictureStorageType = aPicturesStorage;
        dirty = true;
    }
}

void ClientSettings::setClientSwVersion(const StringBuffer& aClientSwVersion)
{
    if (aClientSwVersion != clientSwVersion) {
        clientSwVersion = aClientSwVersion;
        dirty = true;
    }
}

void ClientSettings::setOldSwVersion(const StringBuffer& aSwVersion)
{
    if (aSwVersion != oldSwVersion) {
        oldSwVersion = aSwVersion;
        dirty = true;
    }
}

void ClientSettings::forceDirtyOnMainNode() {
    dirty = true;
    nativeDirty = true;
}


bool ClientSettings::read()
{
    bool ok = DMTClientConfig::read();
    if (!ok) {
        LOG.debug("DMTClientConfig::read failed");
        return false;
    }

    if (strlen(getSwv()) == 0) {
        LOG.debug("Swv empty -> config corrupted");
        return false;
    }
    LOG.debug("Reading Client settings");

    // Now load client specific settings
    StringBuffer context(APPLICATION_URI);

    //
    // Load the main node settings
    //
    DMTree mainTree(context.c_str());
    ManagementNode* mainNode = mainTree.getNode("");
    if (mainNode == NULL) {
        return false;
    }

    char* buf = mainNode->readPropertyValue(pendingSyncConfigPropertyName.c_str());
    pendingSyncConfig = !strcmp(buf, "true");
    delete [] buf;  buf = NULL;

    buf = mainNode->readPropertyValue(syncUsingPropertyName.c_str());
    if (buf && strlen(buf)>0) {
        syncUsing = (TSyncUsing)atoi(buf);
    }
    delete [] buf;  buf = NULL;

    buf = mainNode->readPropertyValue(lastEmailSyncTimestampPropertyName.c_str());
    // need some special conversion, for 64bit values
    lastEmailSyncTimestamp = get64bitValue(buf);
    delete [] buf;  buf = NULL;

    buf = mainNode->readPropertyValue(firstPicturesSyncTimestampPropertyName.c_str());
    if (buf && strlen(buf)>0) {
        firstPicturesSyncTimestamp = strtol(buf, NULL, 10);
    }
    delete [] buf;  buf = NULL;
    
    buf = mainNode->readPropertyValue(installTimestampPropertyName.c_str());
    if (buf && strlen(buf)>0) {
        installTimestamp = strtol(buf, NULL, 10);
    }
    delete [] buf;  buf = NULL;

    buf = mainNode->readPropertyValue(includeOlderPicturesPropertyName.c_str());
    if (buf && strlen(buf)>0) {
        includeOlderPictures = !strcmp(buf, "true");
    }
    delete [] buf;  buf = NULL;
    
    buf = mainNode->readPropertyValue(lastEmailSyncStatusPropertyName.c_str());
    if (buf && strlen(buf)>0) {
        lastEmailSyncStatus = (TLastSyncStatus)atoi(buf);
    }
    delete [] buf;  buf = NULL;

    buf = mainNode->readPropertyValue(lastPhotosSyncStatusPropertyName.c_str());
    if (buf && strlen(buf)>0) {
        lastPhotosSyncStatus = (TLastSyncStatus)atoi(buf);
    }
    delete [] buf;  buf = NULL;

    buf = mainNode->readPropertyValue(schedulerMinutesPropertyName.c_str());
    if (buf && strlen(buf)>0) {
        schedulerMinutes = atoi(buf);
    }
    delete [] buf;  buf = NULL;

    buf = mainNode->readPropertyValue(cpIntervalPropertyName.c_str());
    if (buf && strlen(buf)>0) {
        cpInterval = atoi(buf);
    }
    delete [] buf;  buf = NULL;

    buf = mainNode->readPropertyValue(pushTypePropertyName.c_str());
    if (buf && strlen(buf)>0) {
        pushType = atoi(buf);
    }
    delete [] buf;  buf = NULL;

    buf = mainNode->readPropertyValue(iapNamePropertyName.c_str());
    if (buf && strlen(buf)>0) {
        iapName = buf;
    }
    delete [] buf;  buf = NULL;

    buf = mainNode->readPropertyValue(pictureStorageTypePropertyName.c_str());
    if (buf && strlen(buf)>0) {
        pictureStorageType = atoi(buf);
    }
    delete [] buf;  buf = NULL;

    buf = mainNode->readPropertyValue(clientSwVersionPropertyName.c_str());
    if (buf && strlen(buf)>0) {
        clientSwVersion = buf;
    }
    delete [] buf;  buf = NULL;

    buf = mainNode->readPropertyValue(oldSwVersionPropertyName.c_str());
    if (buf && strlen(buf)>0) {
        oldSwVersion = buf;
    }
    delete [] buf;
    delete mainNode;
    
    
    // Reads the list of sources visible.
    readSourcesVisible();

    ok = readConfigSyncSourceSettings();
    return ok;
}


void ClientSettings::readSourcesVisible() {

    LOG.debug("ClientSettings::readSourcesVisible");
    sourcesVisible.clear();
    
    StringBuffer context(APPLICATION_URI);
    DMTree mainTree(context.c_str());
    ManagementNode* mainNode = mainTree.getNode("");
    if (mainNode == NULL) {
        return;
    }
    
    // Read the (comma separated) source names
    const char* tmp = mainNode->readPropertyValue(sourceOrderPropertyName.c_str());
    StringBuffer sources(tmp);
    delete [] tmp;
    delete mainNode;
    LOG.debug("sourceOrder = %s", sources.c_str());

    // Get the source names, and add them to the sourceVisible array
    if (!sources.empty()) {
        ArrayList tokens;
        sources.split(tokens, ",");

        for (int i=0; i<tokens.size(); i++) {
            StringBuffer* token = (StringBuffer*)tokens.get(i);
            safeAddSourceVisible(*token);
        }
    }

    // Anyway contacts, calendar&tasks MUST be visible!
    // --- TODO: remove when all sources are dynamically visible ---
    safeAddSourceVisible(SOURCE_CONTACTS_NAME);
    safeAddSourceVisible(SOURCE_CALENDAR_NAME);
    
    LOG.debug("sourcesVisible number = %d", sourcesVisible.size());
    //LOG.debug("exit ClientSettings::readSourcesVisible");
}

void ClientSettings::saveSourcesVisible() {
    
    LOG.debug("ClientSettings::saveSourcesVisible");
    LOG.debug("sourcesVisible number = %d", sourcesVisible.size());

    // Joins all the source names in a comma separated string
    StringBuffer sources;
    sources.join(sourcesVisible, ",");
    
    LOG.debug("sourceOrder = %s", sources.c_str());

    // Saves to config file
    StringBuffer context(APPLICATION_URI);
    DMTree mainTree(context.c_str());
    ManagementNode* mainNode = mainTree.getNode("");
    if (mainNode == NULL) {
        return;
    }
    mainNode->setPropertyValue(sourceOrderPropertyName.c_str(), sources.c_str());
    delete mainNode;
    //LOG.debug("exit ClientSettings::saveSourcesVisible");
}


bool ClientSettings::safeAddSourceVisible(const char* sourceName) {

    LOG.debug("adding source '%s' to sourcesVisible array", sourceName);
    
    for (int i=0; i<sourcesVisible.size(); i++) {
        StringBuffer* element = (StringBuffer*)sourcesVisible.get(i);
        if (*element == sourceName) {
            // found: don't add
            LOG.debug("already existing, don't add (tot = %d)", sourcesVisible.size());
            return false;
        }
    }

    // not found: add
    StringBuffer source(sourceName);
    sourcesVisible.add(source);
    LOG.debug("added (tot = %d)", sourcesVisible.size());
    return true;
}


bool ClientSettings::removeSourceVisible(const char* sourceName) {
    
    LOG.debug("removing source '%s' from sourcesVisible array", sourceName);

    for (int i=0; i<sourcesVisible.size(); i++) {
        StringBuffer* element = (StringBuffer*)sourcesVisible.get(i);
        if (*element == sourceName) {
            // found: remove
            sourcesVisible.removeElementAt(i);
            LOG.debug("found, removed (tot = %d)", sourcesVisible.size());
            return true;
        }
    }

    // not found
    LOG.debug("not found (tot = %d)", sourcesVisible.size());
    return false;
}



bool ClientSettings::isCredentialsEmpty()
{
    StringBuffer username(getUsername());
    StringBuffer password(getPassword());
    StringBuffer syncURL(getSyncURL());

    if (username.empty() || password.empty() || syncURL.empty()) {
        LOG.debug("Credentials empty");
        return true;
    }
    return false;
}


bool ClientSettings::readConfigSyncSourceSettings() {
    //
    // Load the ConfigSyncSource settings (meaning: handled by the config sync source)
    //
    StringBuffer context(APPLICATION_URI);
    context += configContext;
    DMTree emailTree(context.c_str());
    ManagementNode* addressNode = emailTree.getNode(addressNodeName.c_str());
    if (addressNode == NULL) {
        return false;
    }
    char* buf = addressNode->readPropertyValue(nodePropertyName.c_str());
    emailAddress = buf;
    delete addressNode;
    delete [] buf;

    ManagementNode* displayNode = emailTree.getNode(displayNodeName.c_str());
    if (displayNode == NULL) {
        return false;
    }
    buf = displayNode->readPropertyValue(nodePropertyName.c_str());
    displayName = buf;
    delete displayNode;
    delete [] buf;
    
    return true;
}



bool ClientSettings::save() {

    bool dmtSaved = DMTClientConfig::save();

    // If this config is not dirty, there is nothing to save
    if (!dirty) {
        return dmtSaved;
    }

    // We must propagate the information to the
    // native profiles
    StringBuffer username(getUsername());
    StringBuffer password(getPassword());
    StringBuffer syncURL(getSyncURL());
    TInt err;

    CSyncProfileManager* profileManager = NULL;
    CImapAccountManager* imapManager = NULL;
    if (nativeDirty)
    {
        profileManager = CSyncProfileManager::createNewInstance();
        if (! profileManager->setCredentials(username, password, syncURL)) {
            LOG.error("Cannot set credentials in sync profile account");
            return false;
        }

        if (ENABLE_SOURCE_MAIL) {
            TRAP(err,
                imapManager = CImapAccountManager::NewL();
                imapManager->setCredentials(username, password);
            );
            if (err != KErrNone) {
                LOG.error("Cannot create imap account manager");
            }
            // Propagate Display Name and Email Address
            imapManager->setAddress(emailAddress);
            imapManager->setDisplayName(displayName);
        }
    }

    // Now save client specific settings
    StringBuffer context(APPLICATION_URI);

    //
    // Save the main node settings
    //
    DMTree mainTree(context.c_str());
    ManagementNode* mainNode = mainTree.getNode("");
    if (mainNode == NULL) {
        return false;
    }
    mainNode->setPropertyValue(pendingSyncConfigPropertyName.c_str(), 
                               pendingSyncConfig ? "true" : "false");
    StringBuffer tmp;
    tmp.sprintf("%d", syncUsing);
    mainNode->setPropertyValue(syncUsingPropertyName.c_str(), tmp.c_str());

    // In Symbian 64 bits values are not printed properly by our APIs. To
    // workaround this problem, we print the subwords (the endianess is not
    // important here as we just store/load values, do not manipulate them).
    TUint* p = (TUint*)&lastEmailSyncTimestamp;
    tmp.sprintf("%08x%08x", *p, *(p+1));
    mainNode->setPropertyValue(lastEmailSyncTimestampPropertyName.c_str(), tmp.c_str());

    tmp.sprintf("%d", lastEmailSyncStatus);
    mainNode->setPropertyValue(lastEmailSyncStatusPropertyName.c_str(), tmp.c_str());
    
    tmp.sprintf("%li", (long)firstPicturesSyncTimestamp);
    mainNode->setPropertyValue(firstPicturesSyncTimestampPropertyName.c_str(), tmp.c_str());

    tmp.sprintf("%li", (long)installTimestamp);
    mainNode->setPropertyValue(installTimestampPropertyName.c_str(), tmp.c_str());

    mainNode->setPropertyValue(includeOlderPicturesPropertyName.c_str(), 
                               includeOlderPictures ? "true" : "false");
    
    tmp.sprintf("%d", lastPhotosSyncStatus);
    mainNode->setPropertyValue(lastPhotosSyncStatusPropertyName.c_str(), tmp.c_str());

    tmp.sprintf("%d", schedulerMinutes);
    mainNode->setPropertyValue(schedulerMinutesPropertyName.c_str(), tmp.c_str());

    tmp.sprintf("%d", cpInterval);
    mainNode->setPropertyValue(cpIntervalPropertyName.c_str(), tmp.c_str());

    tmp.sprintf("%d", pushType);
    mainNode->setPropertyValue(pushTypePropertyName.c_str(), tmp.c_str());

    mainNode->setPropertyValue(iapNamePropertyName.c_str(), iapName.c_str());
    
    tmp.sprintf("%d", pictureStorageType);
    mainNode->setPropertyValue(pictureStorageTypePropertyName.c_str(), tmp.c_str());
    // save client sw version
    mainNode->setPropertyValue(clientSwVersionPropertyName.c_str(), clientSwVersion.c_str());
    // save old sw version
    mainNode->setPropertyValue(oldSwVersionPropertyName.c_str(), oldSwVersion.c_str());

    // The IAP information must be propagated to the native profiles
    if (nativeDirty) {
        profileManager->setIap(iapName);
        if (ENABLE_SOURCE_MAIL) {
            imapManager->setIap(iapName);
        }
        delete profileManager;
        delete imapManager;
        nativeDirty = false;
    }
    delete mainNode;
    
    // Saves the list of sources visible.
    saveSourcesVisible();

    //
    // Save the ConfigSyncSource settings (meaning: handled by the config sync source)
    //
    context += configContext;
    DMTree emailTree(context.c_str());
    ManagementNode* addressNode = emailTree.getNode(addressNodeName.c_str());
    if (!emailAddress.empty()) {
        addressNode->setPropertyValue(nodePropertyName.c_str(), emailAddress.c_str());
    }
    delete addressNode;

    ManagementNode* displayNode = emailTree.getNode(displayNodeName.c_str());
    if (!displayName.empty()) {
        displayNode->setPropertyValue(nodePropertyName.c_str(), displayName.c_str());
    }
    delete displayNode;
    dirty = false;

    return true;
}

void ClientSettings::saveSourceConfig(int i, ManagementNode& n) {

    char nodeName[DIM_MANAGEMENT_PATH];
    ManagementNode* node = NULL;

    if (n.getChild(i) == NULL) {
        // Create node from Source name.
        char* fn = n.createFullName();
        sprintf(nodeName, "%s/%s", fn, sourceConfigs[i].getName());
        delete [] fn;
        node = dmt->readManagementNode(nodeName);
    }
    else {
        node = (ManagementNode*)n.getChild(i)->clone();
    }

    if (node) {
        DMTClientConfig::saveSourceConfig(i, n, *node);
        saveSourceVars(i, n, *node);
        delete node;
    }
}


const ArrayList& ClientSettings::getSourcesVisible() {
    return sourcesVisible;
}


void ClientSettings::createDefaultConfig()
{
    //
    // AccessConfig
    //
    AccessConfig* ac = DefaultSymbianConfigFactory::getAccessConfig();
    setAccessConfig(*ac);
    delete ac;

    //
    // DeviceConfig
    //
    DeviceConfig* dc = DefaultSymbianConfigFactory::getDeviceConfig();
    setDeviceConfig(*dc);
    delete dc;

    //
    // SyncSources configs
    // We must create all configs, even if the source is hidden to the user.
    //
    RArray<TInt> sourceIDs;
    sourceIDs.Append(SOURCE_CALENDAR);
    sourceIDs.Append(SOURCE_CONFIG);
    sourceIDs.Append(SOURCE_CONTACTS);
    sourceIDs.Append(SOURCE_MAIL);
    sourceIDs.Append(SOURCE_NOTES);
    sourceIDs.Append(SOURCE_PHOTOS);

    // Let's remove previously created sourceConfigs (if any) and reinitialize it.
    // It may happen we have empty/corrupted config files for sources, so remove them.
    if (sourceConfigs) {
        delete [] sourceConfigs;
        sourceConfigs = NULL;
    }
    sourceConfigsCount = 0;

    for (int i=0; i<sourceIDs.Count(); i++) {
        StringBuffer sourceName = getSourceNameFromID(sourceIDs[i]);
        SyncSourceConfig* sc = DefaultSymbianConfigFactory::getSyncSourceConfig(sourceName.c_str());
        addSyncSourceConfig(*sc);
        delete sc;
    }

    // Default sources visible are set here
    sourcesVisible.clear();
    safeAddSourceVisible(SOURCE_CONTACTS_NAME);
    safeAddSourceVisible(SOURCE_CALENDAR_NAME);
    if (ENABLE_SOURCE_MAIL) {
        safeAddSourceVisible(SOURCE_MAIL_NAME);
    }
    if (ENABLE_SOURCE_NOTES) {
        safeAddSourceVisible(SOURCE_NOTES_NAME);
    }
    if (ENABLE_SOURCE_PHOTOS) {
        safeAddSourceVisible(SOURCE_PHOTOS_NAME);
    }
    LOG.debug("default sourceVisible array: %d sources", sourcesVisible.size());

    //
    // Root params
    //
    setPendingSyncConfig     (true);
    setSyncUsing             ((TSyncUsing)DEFAULT_SYNC_USING);
    setLastEmailSyncTimestamp(0);
    setLastEmailSyncStatus   (ELastSyncNotFound);
    setFirstPicturesSyncTimestamp(0);
    setInstallTimestamp      (time(NULL));          // now
    setIncludeOlderPictures  (DEFAULT_INCLUDE_OLDER_PICTURES);
    setLastPhotosSyncStatus  (ELastSyncNotFound);
    setSchedulerMinutes      (DEFAULT_SCHEDULER_INTERVAL);
    setCpInterval            (CP_INTERVAL_SECONDS);
    setPushType              (DEFAULT_PUSH_TYPE);
    setIapName               ("");
    // funambol client sw version
    clientSwVersion = getPropertyFromUpgradeDir(clientSwVersionPropertyName);
    // at first install set old sw version equal to clientSwVersion
    setOldSwVersion         (clientSwVersion);

    //
    // Config syncsource params
    //
    emailAddress = "";
    displayName  = "";

    //
    // Set deviceID + software version + user agent.
    //
    StringBuffer devId = retrieveDeviceId();
    getDeviceConfig().setDevID(devId.c_str());

    StringBuffer swv = getPropertyFromUpgradeDir(PROPERTY_SOFTWARE_VERSION);
    getDeviceConfig().setSwv(swv.c_str());

    StringBuffer userAgent = createUserAgent(swv);
    getAccessConfig().setUserAgent(userAgent.c_str());


    // *** since 7.1.12 ***
    // Check if SYNC_PROFILE_NAME profile exists. If so, get account data from there.
    // This avoids slow syncs in case the user has already a SYNC_PROFILE_NAME profile in sync.
    // If profile not found we force a slow-sync for PIM, to avoid weird issues first time.
    // Note: we must do it before saving the config (it creates the profile)
    CSyncProfileManager* profileManager = CSyncProfileManager::createNewInstance();
    bool profileExists = profileManager->profileExists();
    StringBuffer username, password, syncURL;

    if (profileExists) {
        // Set the account data stored in the existing profile.
        setUsername(profileManager->getUsername());
        setPassword(profileManager->getPassword());
        setSyncURL (profileManager->getServerURI());
        copiedAccountFromProfile = true;

        // Check if all tasks (sources) are configured!
        // It may happen that 'notes' is never used nor configured, so will be created here.
        profileManager->CheckAllTasks();
    }
    else {
        // Reset timestamps for PIM sources.
        SyncSourceConfig* ssc = getSyncSourceConfig(SOURCE_CONTACTS_NAME);
        if (ssc) { ssc->setLast(0); }
        ssc = getSyncSourceConfig(SOURCE_CALENDAR_NAME);
        if (ssc) { ssc->setLast(0); }
        ssc = getSyncSourceConfig(SOURCE_NOTES_NAME);
        if (ssc) { ssc->setLast(0); }
        copiedAccountFromProfile = false;
    }
    delete profileManager;


    // To make sure we spread the settings to sync profile / email account.
    forceDirtyOnMainNode();

    sourceIDs.Close();
}


bool ClientSettings::addSyncSourceConfig(SyncSourceConfig& sc) {

    unsigned int i = 0;
    SyncSourceConfig* s = NULL;

    // copy array in a tmp buffer
    if (sourceConfigsCount > 0) {
        s = new SyncSourceConfig[sourceConfigsCount];
        for (i=0; i<sourceConfigsCount; i++) {
            s[i].assign(sourceConfigs[i]);
        }       
    }

    // delete old one, create new (+1 element)
    if (sourceConfigs) {
        delete [] sourceConfigs;
    }

    sourceConfigs = new SyncSourceConfig[sourceConfigsCount + 1];
    // copy back.
    for (i = 0; i < sourceConfigsCount; i++) {
        sourceConfigs[i].assign(s[i]);
    }
    
    sourceConfigsCount++;
    sourceConfigs[i].assign(sc);

    delete [] s;
    return true;
}


StringBuffer ClientSettings::getPropertyFromUpgradeDir(const StringBuffer& propertyName)
{
    StringBuffer value("");
    if (propertyName.empty()) {
        return value;
    }

    StringBuffer context(APPLICATION_URI);
    context += upgradeContext;
    DMTree upgradeTree(context.c_str());
    ManagementNode* upgradeNode = upgradeTree.getNode("");
    if (!upgradeNode) {
        return value;
    }
    char* buf = upgradeNode->readPropertyValue(propertyName.c_str());
    value = buf;

    delete [] buf;
    delete upgradeNode;
    return value;
}


bool ClientSettings::makeUpgradeActions()
{
    StringBuffer currentSwv = getDeviceConfig().getSwv();
    StringBuffer newSwv = getPropertyFromUpgradeDir(PROPERTY_SOFTWARE_VERSION);

    if (newSwv == currentSwv) {
        // No upgrade necessary.
        return false;
    }

    int oldVersion = getBuildNumberFromVersion(currentSwv.c_str());
    int newVersion = getBuildNumberFromVersion(newSwv.c_str());

    if (newVersion == oldVersion) {
        // No upgrade necessary.
        return false;
    }
    else if (newVersion < oldVersion) {
        // Downgrade software! -> avoid? warn user?
        LOG.info("Warning: the version installed is older than the previous one: '%s'", currentSwv.c_str());
    }

    //
    // Upgrade configuration.
    // **********************
    //
    getDeviceConfig().setSwv(newSwv.c_str());
    StringBuffer userAgent = createUserAgent(newSwv);
    getAccessConfig().setUserAgent(userAgent.c_str());
     // oldSwv will be updated in AppUi InitializeClient()
    setOldSwVersion(currentSwv);

    // client software version: if not set use sw version
    if (clientSwVersion.empty()) {
        clientSwVersion = currentSwv;
    }

    StringBuffer newFunambolSwv = getPropertyFromUpgradeDir(clientSwVersionPropertyName.c_str());
    int oldClientVersion = getBuildNumberFromVersion(clientSwVersion.c_str());
    int newClientVersion = getBuildNumberFromVersion(newFunambolSwv.c_str());

    if (oldClientVersion == newClientVersion) {
        // No upgrade necessary.
        return false;
    }

    setClientSwVersion(newFunambolSwv);

    //
    // Add here any code to ugrade the client based on the previous version.
    //
    if (oldClientVersion < 70104)
    {
        // Remove the obsolete log file: new log name is defined in customization.h
        LOG.debug("Removing old log file");
        RFs fsSession;
        fsSession.Connect();
        RBuf name;
        name.Assign(charToNewBuf("C:\\data\\synclog.txt"));
        fsSession.Delete(name);
        fsSession.Close();
        name.Close();
    }

    // SMS Push added in v.7.2.6
    if (oldClientVersion < 70206) {
        setPushType(DEFAULT_PUSH_TYPE);
    }

    if (oldClientVersion < 80200)
    {
        // Added pictures syncsource in v.8.2.0
        SyncSourceConfig* sc = DefaultSymbianConfigFactory::getSyncSourceConfig(SOURCE_PHOTOS_NAME);
        addSyncSourceConfig(*sc);
        delete sc;
        setLastPhotosSyncStatus(ELastSyncNotFound);
        setFirstPicturesSyncTimestamp(0);

        // MaxMsgSize 10k -> 64k (for pictures sync)
        getAccessConfig().setMaxMsgSize(MAX_SYNCML_MSG_SIZE);

        // Added SyncSource boolean 'enabled' to enable/disable a source
        // without losing the sync direction information.
        // If syncmode = none -> disable ssource.
        for (unsigned int i=0; i<sourceConfigsCount; i++) {
            SyncSourceConfig* ssc = getSyncSourceConfig(i);
            if (ssc) {
                StringBuffer syncMode = ssc->getSync();
                if (syncMode == SYNC_NONE_TEXT) { ssc->setIsEnabled(false); }
                else                            { ssc->setIsEnabled(true ); }
            }
        }
        
        // CTP stretched params, to maximize battery life, since 8.2.0
        CTPConfig* ctpConfig = CTPService::getInstance()->getConfig();
        if (ctpConfig) {
            ctpConfig->setCtpReady(CTP_READY_INTERVAL);
            ctpConfig->setCtpRetry(CTP_RETRY_INTERVAL);
            ctpConfig->setMaxCtpRetry(CTP_MAX_RETRY);
            ctpConfig->saveCTPConfig();
        }
    }
    
    if (oldClientVersion < 80500) 
    {
        // Source panels are dynamically loaded since 8.5.0.
        // Contacts and calendar are always visible.
        if (ENABLE_SOURCE_MAIL) {
            safeAddSourceVisible(SOURCE_MAIL_NAME);
        }
        if (ENABLE_SOURCE_NOTES) {
            safeAddSourceVisible(SOURCE_NOTES_NAME);
        }
        if (ENABLE_SOURCE_PHOTOS) {
            safeAddSourceVisible(SOURCE_PHOTOS_NAME);
        }
        
        // this timestamp was introduced in v8.5, so it rapresents
        // the first execution of the v8.5 client (or later)
        setInstallTimestamp(time(NULL));
        
        // filter on older pictures
        setIncludeOlderPictures(DEFAULT_INCLUDE_OLDER_PICTURES);
    }
    
    
    // ALWAYS - to be sure we don'tsend the devinfo using Funambol APIs.
    getDeviceConfig().setSendDevInfo(false);
    
    // ALWAYS - to be sure we ask for Server caps at the next sync.
    // Note: this is actually useless, since every first sync the client 
    // forces to ask for server caps, via a config sync.
    setServerLastSyncURL("");
    

    save(); // always save config to update clientSwVersion

    LOG.info("Configuration succesfully upgraded to new version: '%s'", newSwv.c_str());
    return true;
}


StringBuffer ClientSettings::createUserAgent(const StringBuffer& swv)
{
    StringBuffer userAgent("");
    userAgent.sprintf("%s v. %s", APPLICATION_NAME, swv.c_str());
    return userAgent;
}

bool ClientSettings::createSyncMLProfile()
{
    StringBuffer username  = getUsername();
    StringBuffer password  = getPassword();
    StringBuffer serverURI = getSyncURL();

    CSyncProfileManager* profileManager = CSyncProfileManager::createNewInstance();
    bool ret = profileManager->createProfile(username, password, serverURI, iapName);

    delete profileManager;
    return ret;
}

bool ClientSettings::checkCopiedAccountFromProfile()
{
    return copiedAccountFromProfile;
}

bool ClientSettings::createIMAPAccount()
{
    StringBuffer username  = getUsername();
    StringBuffer password  = getPassword();

    CImapAccountManager* imapManager = NULL;
    TRAPD(err, imapManager = CImapAccountManager::NewL());
    if (err != KErrNone) {
        LOG.error("Cannot open IMAP account manager (err =  %d)", err);
        return false;
    }

    /*if (imapManager->AccountExists(SYNC_PROFILE_NAME)) {
        // No need to create the IMAP account
        delete imapManager;
        return false;
    }*/

    // This call creates the IMAP account
    bool ret = imapManager->setCredentials(username, password);

    // Propagate all parameters
    imapManager->setAddress(emailAddress);
    imapManager->setDisplayName(displayName);
    imapManager->setIap(iapName);

    delete imapManager;
    return ret;
}


/* --- Not used: dirs are created automatically by DMNode if fopen fails ---
TInt ClientSettings::createAllDirs(const StringBuffer& rootDir, RArray<TInt>& sourceIDs)
{
    TInt err = KErrNone;

    // Connect to the file server
    RFs fileSession;
    err = fileSession.Connect();
    if (err) {
        LOG.error("Cannot connect to the file server (code %d)", err);
        return err;
    }

    CleanupClosePushL(fileSession);

    StringBuffer mainPath(rootDir);
    mainPath += APPLICATION_URI;

    StringBuffer syncmlPath(mainPath);
    syncmlPath += CONTEXT_SPDS_SYNCML;

    StringBuffer sourcesPath(mainPath);
    sourcesPath += CONTEXT_SPDS_SOURCES;

    //
    // "\spds\syncml" dirs
    //
    err = makeDir(syncmlPath, CONTEXT_AUTH,       &fileSession);
    if (err) { goto finally; }
    err = makeDir(syncmlPath, CONTEXT_CONN,       &fileSession);
    if (err) { goto finally; }
    err = makeDir(syncmlPath, CONTEXT_DEV_DETAIL, &fileSession);
    if (err) { goto finally; }
    err = makeDir(syncmlPath, CONTEXT_DEV_INFO,   &fileSession);
    if (err) { goto finally; }
    err = makeDir(syncmlPath, CONTEXT_EXT,        &fileSession);
    if (err) { goto finally; }

    //
    // "\spds\sources" dirs
    //
    for (int i=0; i<sourceIDs.Count(); i++) {
        StringBuffer sourceName = getSourceNameFromID(sourceIDs[i]);
        err = makeDir(sourcesPath, sourceName.c_str(), &fileSession);
        if (err) { goto finally; }
    }

finally:
    CleanupStack::PopAndDestroy(&fileSession);
    return err;
}

TInt ClientSettings::makeDir(const StringBuffer& aPath, const char* aContext, RFs* aFileSession)
{
    TInt err = KErrNone;
    RBuf pathDes;

    StringBuffer path(aPath);
    concatDirs(path, aContext);
    path = contextToPath(path.c_str());     // "/" -> "\"

    path += "\\";
    LOG.debug("mkdir: '%s'", path.c_str());
    pathDes.Assign(stringBufferToNewBuf(path));
    err = aFileSession->MkDirAll(pathDes);
    pathDes.Close();

    return err;
}
*/


