/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2003 - 2009 Funambol, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 *
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

#include <e32base.h>
#include <f32file.h>
#include <s32file.h>
#include <apgcli.h>  // RApaLsSession
#include <bautils.h> // jad file

#include "uid3.h"
#include "Funambol.hrh"
#include "customization.h"
#include "base/util/stringUtils.h"
#include "MidletInstaller.h"

USE_NAMESPACE

CMidletInstaller* CMidletInstaller::NewL(const StringBuffer& aJadFile)
{
    CMidletInstaller* self = new (ELeave) CMidletInstaller(aJadFile);
    CleanupStack::PushL(self);
    self->ConstructL();
    CleanupStack::Pop();

    return self;
}

void CMidletInstaller::ConstructL()
{ 
    iLauncher.Connect();
    iWaiter = CAOSync::NewL();
    iJadPath = charToNewBuf(iJadFile.c_str());
}

CMidletInstaller::CMidletInstaller(const StringBuffer& aJadFile) : 
    iJadFile(aJadFile) {}

CMidletInstaller::~CMidletInstaller()
{
    iLauncher.Close();
    delete iWaiter;
    delete iJadPath;
}

TBool CMidletInstaller::InstallL()
{
    LOG.debug("installing midlet...");
    if (IsJadInstalledL()) {
        if (InstallMidlet() == EFalse) {
            return EFalse;
        }
    } else {
        LOG.error("error: can't found jad file");
        return EFalse;
    }
    
    return ETrue;
}

// MIDlet installation/uninstallation reference:
// http://wiki.forum.nokia.com/index.php/CS000817_-_Installing_and_uninstalling_MIDlets_programmatically

TBool CMidletInstaller::IsMidletInstalledL()
{
    // check if the midlet is installed
    TBool result = EFalse;
    TPtrC midletName( (const TText*) MIDLET_NAME );
    TApaAppInfo appInfo;
    RApaLsSession apaSession;
 
    User::LeaveIfError( apaSession.Connect() );
    CleanupClosePushL( apaSession );
    User::LeaveIfError( apaSession.GetAllApps() );

    while ( apaSession.GetNextApp( appInfo ) == KErrNone ) {
        if (appInfo.iFullName.Right(8).Compare(_L(".fakeapp")) == 0) {
            if (appInfo.iCaption.Compare( midletName ) == 0) {
                LOG.info("Java application found");
                result = ETrue;
                break;
            }
        }
    }
    
    CleanupStack::PopAndDestroy(); // also close apaSession            
    
    return result;
}

TBool CMidletInstaller::IsJadInstalledL()
{
    TBool result = EFalse;
    
    // check if jad file is installed
    RFs fsSession;
    User::LeaveIfError(fsSession.Connect());
    CleanupClosePushL( fsSession );
    
    result = BaflUtils::FileExists(fsSession, iJadPath->Des());
    
    CleanupStack::PopAndDestroy(1, &fsSession);

    return result;
}

TBool CMidletInstaller::InstallMidlet()
{
    CAknWaitDialog* waitDialog = NULL;
    TInt ret = KErrNone;

    LOG.debug("creating wait dialog");

    TRAPD(err, 
        {
            waitDialog = new (ELeave) CAknWaitDialog( reinterpret_cast<CEikDialog**>(&waitDialog), ETrue);
            waitDialog->PrepareLC(R_JAM_INSTALLER_POPUP);
            waitDialog->RunLD();
        } 
    );

    if (err) {
        LOG.error("error creating wait dialog for installer");
        waitDialog = NULL;
    }

    // set installation options
    iInstallOpts.iDrive        = 'C';   
    iInstallOpts.iUntrusted    = SwiUI::EPolicyAllowed; 
    iInstallOpts.iCapabilities = SwiUI::EPolicyAllowed; 
    iInstallOpts.iUpgrade      = SwiUI::EPolicyAllowed;
    iInstallOpts.iOCSP         = SwiUI::EPolicyAllowed;         
    iInstallOpts.iKillApp      = SwiUI::EPolicyAllowed;
    iInstallOpts.iOverwrite    = SwiUI::EPolicyAllowed;
    iInstallOpts.iUpgradeData  = SwiUI::EPolicyAllowed;

    iInstallOptsPckg = iInstallOpts;   
    LOG.debug("connecting to installer");

    if ((ret = iLauncher.Connect()) != KErrNone) {
        LOG.error("error connecting to installation server [%d]", ret);

        return EFalse;
    }

    LOG.debug("launching installation in silent mode...");
    iLauncher.SilentInstall(iWaiter->iStatus, iJadPath->Des(), iInstallOptsPckg);

    iWaiter->Execute();

    LOG.debug("installation done");

    if (waitDialog)
        waitDialog->ProcessFinishedL(); 

    return ETrue;
}

#ifdef ENABLE_MIDLET_UNINSTALL

void CMidletInstaller::UninstallL()
{
    //Check midlet installation
    if (IsMidletInstalledL()) {
        LOG.debug("uninstalling midlet...");

        UninstallMidletL();

        LOG.debug("midlet removed");
    } else {
        LOG.debug("midlet not installed");
    }
}

// This works only for 60 3rd Edition MR and FP1 devices.
// Apparently there's no way to remove programmatically 
// installed Midlets from FP2 devices, see:
// http://discussion.forum.nokia.com/forum/showthread.php?t=158285

TInt CMidletInstaller::UninstallMidletL()
{
    RFs fileSession;
    CDir* dirList;
    TFileName fullPath;
    TUint contextUid;
    TFileName fileName;
    TInt resp = KErrNone;
 
    _LIT(KRegistryDir, "c:\\system\\install\\registry\\");
    _LIT(KRegistryFilePattern,"c:\\system\\install\\registry\\*.*");
    _LIT(KJarName, MIDLET_NAME);

    User::LeaveIfError( fileSession.Connect() );
    CleanupClosePushL( fileSession );
 
    // Get the file list, sorted by name
    User::LeaveIfError( fileSession.GetDir( KRegistryFilePattern, 
                                         KEntryAttMaskSupported, 
                                         ESortByName, dirList ) );
    CleanupStack::PushL(dirList);

    for (TInt i=0; i < dirList->Count(); i++ ) {
        fileName = (*dirList)[i].iName;
        if  (fileName.Compare(_L("temp")) == KErrNone ) 
            continue;  
 
        fullPath = KRegistryDir;
        fullPath.Append(fileName);
        RFile lFile;
        TInt err = lFile.Open( fileSession, fullPath, EFileShareAny | EFileRead );
   
        if (err == KErrNone) {
            TInt lSize;
            lFile.Size(lSize);
            RBuf8 lBuf;
            lBuf.Create(lSize);
            lBuf.CleanupClosePushL();
            lFile.Read(lBuf);//read contents into buffer
            RBuf buff;
            buff.Create(lSize);
            buff.CleanupClosePushL();
            buff.Copy(lBuf);//Copy into 16 bit buffer
            lFile.Close();
    
            if (err == KErrNone) {
                // Check for MIDlet name in the file
                TPtrC midletName( (const TText*) MIDLET_NAME );

                err = buff.Find(midletName);

                if (err != KErrNotFound) {
                    TLex lex( fileName.Mid(0,8) );
                    lex.Val( contextUid, EHex );
                    SwiUI::RSWInstSilentLauncher inst; 
                    
                    if ((resp = inst.Connect()) != KErrNone) {
                        LOG.error("error connecting to installation server: %d", resp);
                    } else {
                        SwiUI::TUninstallOptions options; 
                        SwiUI::TUninstallOptionsPckg optionsPckg; 
                        options.iKillApp = SwiUI::EPolicyAllowed; 
                        options.iBreakDependency = SwiUI::EPolicyAllowed; 
                        optionsPckg = options; 
                        LOG.debug("midlet found (UID %d): uninstalling...", TUid::Uid(contextUid));
                        resp = inst.SilentUninstall( TUid::Uid(contextUid), 
                                           optionsPckg, 
                                           SwiUI::KJavaMIMEType ); 

                        if (resp != KErrNone) {
                            LOG.error("error removing midlet: %d", resp);
                        }

                        inst.Close(); 
                    }

                    CleanupStack::PopAndDestroy(2); // buff, lBuf      
                    break;
                }
            }
        
            CleanupStack::PopAndDestroy(2); // buff, lBuf      
        }
    }
 
    CleanupStack::PopAndDestroy(2); // dirList, fileSession

    return resp;
}
#endif
