/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2008 Funambol, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 *
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

#include <eikedwin.h>
#include <eikseced.h>
#include <avkon.rsg>
#include <eikmenup.h>
#include <akntitle.h> 
#include <barsread.h>
#include "uid3.h"

#include "base/SymbianLog.h"
#include "base/util/StringBuffer.h"
#include "base/util/stringUtils.h"
#include "base/util/symbianUtils.h"
#include "customization.h"

#include "commonDefs.h"
#include "Funambol.hrh"
#include "PhotoOptionsDialog.h"
#include "FunambolAppUi.h"
#include "ClientSettings.h"
#include "utils.h"

using namespace Funambol;


CPhotoOptionsDialog* CPhotoOptionsDialog::NewL(TInt aMenuTitleResourceId, CFunambolAppUi* aAppUI)
{
    CPhotoOptionsDialog* self = new (ELeave)CPhotoOptionsDialog();
    CleanupStack::PushL(self);
    self->ConstructL(aMenuTitleResourceId, aAppUI);
    CleanupStack::Pop(self);
    return self;
}

void CPhotoOptionsDialog::ConstructL(TInt aMenuTitleResourceId, CFunambolAppUi* aAppUI)
{
    iAppUI = aAppUI;
    iSettingList = NULL;
    CAknDialog::ConstructL(aMenuTitleResourceId);
}

CPhotoOptionsDialog::CPhotoOptionsDialog() : iSyncEnabled(ETrue), 
                                             iIncludeOlderPictures(0)
{
    iMassStorageAvailable = massStorageMemoryAvailable();
    
    if (iMassStorageAvailable) {
        iMemoryInUse = EMassMemoryStorage;
    } else {
        iMemoryInUse = EPhoneMemoryStorage;
    }
}

CPhotoOptionsDialog::~CPhotoOptionsDialog()
{
    if (iSettingList) {
        delete iSettingList;
        iSettingList = NULL;
    }
}

void CPhotoOptionsDialog::PreLayoutDynInitL()
{
    if (!iSettingList) // if we haven't already created iSettingList
    {
        // construct control and set parent
        iSettingList = new (ELeave) CCommonOptionsSettingList();
        iSettingList->SetMopParent(this);

        // Loads data into the setting list.
        load();

        // construct control based on resource definition
        TResourceReader reader;

        // CreateResourceReaderLC will allocate a buffer to be used by
        // the TResourceReader. This buffer is pushed onto the cleanup
        // stack - not the TResourceReader itself
        if (iMassStorageAvailable) {
            // if mass storage is available we need a setting list with
            // multiple choices for pictures storage options
            iCoeEnv->CreateResourceReaderLC(reader, R_PHOTOS_SETTING_LIST_ENUMERATED);
        } else {
            // photo setting list with binary option choice
            iCoeEnv->CreateResourceReaderLC(reader, R_PHOTOS_SETTING_LIST);
        }
    
        iSettingList->ConstructFromResourceL(reader);

        // Clean up the buffer allocated above, NOT the reader itself.
        // Cannot use expected item overload of PopAndDestroy() as buffer 
        // is inaccessible. 
        CleanupStack::PopAndDestroy();

        // Removes the vertical scrollbar
        iSettingList->initScrollbar();

        // Activate the settings list
        iSettingList->ActivateL();
    }

    CAknDialog::PreLayoutDynInitL();

    iAppUI->changeAppTitle(R_STRING_PHOTOS_OPTIONS_TITLE);
}

TBool CPhotoOptionsDialog::OkToExitL(TInt aButtonId)
{
    // Returning ETrue causes the form to exit
    switch (aButtonId)
    {
    case EAknSoftkeyBack:
    {
        // Save data into the config.
        save();
        return ETrue;
    }
    case ESubOptionsChange:
    {
        EditCurrentItem();
        break;
    }

    default:
    {
        // Other commands forwarded to CAknDialog.
        return CAknDialog::OkToExitL(aButtonId);
    }
    }
    return EFalse;
}



TKeyResponse CPhotoOptionsDialog::OfferKeyEventL(const TKeyEvent &aKeyEvent, TEventCode aType)
{
    if (aType != EEventKey) {
        return EKeyWasNotConsumed;
    }

    if (aKeyEvent.iScanCode == EStdKeyDevice3) // ???????? why this code ????????
    {
        EditCurrentItem();
        return EKeyWasConsumed;
    }

    switch ( aKeyEvent.iCode ){
        // Up & Down arrow key's event transfer to setting list
        case EKeyDownArrow:
            return iSettingList->OfferKeyEventL( aKeyEvent, aType );
            break;
        case EKeyUpArrow:
            return iSettingList->OfferKeyEventL( aKeyEvent, aType );
            break;
        default:
            break;
    }
    return EKeyWasNotConsumed;
}


void CPhotoOptionsDialog::EditCurrentItem() 
{
    if (iSettingList->getCurrentItemID() == ESettingsListSyncMode) {
        // Override opening: syncmode is fixed to "one-way-from-client" (warn a popup?)
    }
    else {
        iSettingList->EditCurrentItemL();
    }
}



void CPhotoOptionsDialog::load()
{
    LOG.info("Reading Photo sync Options data...");
    ClientSettings* config = ClientSettings::getInstance();
    config->open();

    //
    // Read values from config
    //
    SyncSourceConfig* ssConfig = config->getSyncSourceConfig(SOURCE_PHOTOS_NAME);
    if (!ssConfig) {
        LOG.error("Config not found for source '%s'", SOURCE_PHOTOS_NAME);
        iSyncEnabled = EFalse;
        return;
    }
    
    iSyncEnabled = ssConfig->isEnabled();
    
    // Show the sync direction even if source is disabled
    // Pictures direction: is always smart-one-way-from-client
    iSyncMode = ESyncModeToServerOnly;
    

    //
    // Memory in use (C:, E: of F: drive)
    // No popup, no fallback in case the memory card is not available 
    // (will be done if the user confirms it exiting the dialog)
    //
    TInt pictureStorage = config->getPicturesStorage();

    if (pictureStorage == EStorageTypeUnset) {
        // SHOULD NOT HAPPEN - get the drive available (memory card is the preferred drive)
        LOG.debug("First time (storage type unset)");
        StringBuffer imagesDir;
        bool driveAvailable = false;

        driveAvailable = getImagesPath(EMemoryCardStorage, imagesDir);

        if (driveAvailable) { 
            iMemoryInUse = EMemoryCardStorage;
        } else { 
            if (iMassStorageAvailable) {
                // if present, fall back on mass storge if memory card isn't available
                iMemoryInUse = EMassMemoryStorage;
            } else {
                iMemoryInUse = EPhoneMemoryStorage;
            }
        }
    } else {
        iMemoryInUse = pictureStorage;
    }

    StringBuffer picturesStorage = getDriveName(iMemoryInUse);
    LOG.debug("Pictures drive used is %s:", picturesStorage.c_str());

    // Include older pictures flag
    iIncludeOlderPictures = config->getIncludeOlderPictures();
    
    // Set values into the settinglist
    iSettingList->setSyncEnabled         (iSyncEnabled);
    iSettingList->setSyncMode            (iSyncMode);
    iSettingList->setMemoryInUse         (iMemoryInUse);
    iSettingList->setIncludeOlderPictures(iIncludeOlderPictures);

    config->close();
}

TBool CPhotoOptionsDialog::save()
{
    TBool newSyncEnabled          = iSettingList->getSyncEnabled();
    TInt newSyncMode              = iSettingList->getSyncMode();
    TInt newMemoryInUse           = iSettingList->getMemoryInUse();
    TBool newIncludeOlderPictures = iSettingList->getIncludeOlderPictures();
    
    //
    // Memory in use (C:, E: or F:)
    // If memorycard not available, fallback to C: (or E: if mass storage device is present) 
    // but warn the user with a popup.
    //
    if (newMemoryInUse == EMemoryCardStorage) {
        StringBuffer imagesDir;
        bool driveAvailable = getImagesPath(EMemoryCardStorage, imagesDir);

        if (!driveAvailable) {
            if (iMassStorageAvailable) {
                // Popup warning
                LOG.info("User selected memory card drive for pitures but it's not available -> fallback to mass storage drive (E:)");
                iAppUI->DisplayMessage(INFO_FALLBACK_MASS_STORAGE, EWarningLevel);
                newMemoryInUse = EMassMemoryStorage;
            } else {
                // Popup warning
                LOG.info("User selected memory card drive for pitures but it's not available -> fallback to phone main memory (C:)");
                iAppUI->DisplayMessage(INFO_FALLBACK_MAIN_MEMORY, EWarningLevel);
                newMemoryInUse = EPhoneMemoryStorage;
            }
        }
    } 

    // Check if values changed
    // Must be done AFTER the check on drive letter (the drive could be changed due to the fallback).
    if ( (newSyncEnabled          == iSyncEnabled) && 
         (newSyncMode             == iSyncMode)    &&
         (newMemoryInUse          == iMemoryInUse) &&
         (newIncludeOlderPictures == iIncludeOlderPictures) )
    {
        LOG.debug("Data unchanged -> no need to save config.");
        return EFalse;
    }

    
    // 
    // Source enabled / sync direction.
    //
    ClientSettings* config = ClientSettings::getInstance();
    config->open();
    SyncSourceConfig* ssConfig = config->getSyncSourceConfig(SOURCE_PHOTOS_NAME);
    if (!ssConfig) {
        LOG.error("Config not found for source '%s'", SOURCE_PHOTOS_NAME);
        return EFalse;
    }
    
    ssConfig->setIsEnabled(newSyncEnabled);

    // Pictures direction is ONLY smart-one-way-from-client.
    if (newSyncMode != ESyncModeToServerOnly) {
        LOG.info("Warning! Sync direction for pictures will be forced to '%s'", DEFAULT_PICTURES_SYNC_MODE);   
    }
    ssConfig->setSync(DEFAULT_PICTURES_SYNC_MODE);

    if (newSyncEnabled) {  
        // The scheduler could be disabled if no sources are active.
        if (config->getSyncUsing() == ESyncUsingScheduled)
        {
            iAppUI->StartServices();
        }
    }
   
    StringBuffer picturesDrive =  getDriveName(newMemoryInUse);
    LOG.debug("setting picture storage location to config (%s:)", picturesDrive.c_str());

    config->setPicturesStorage(newMemoryInUse);
    
    // Include older pictures flag
    LOG.debug("setting the include older pictures flag (%s)", newIncludeOlderPictures? "yes":"no");
    config->setIncludeOlderPictures(newIncludeOlderPictures);
    
    //
    // Store values to config
    //
    config->save();
    config->close();
    
    //
    // Restart the C2S push for pictures, in case the folder location has changed.
    // Setting the new pictures location, the filesystem listener will automatically restart if necessary
    // if it's changed
    //
    if (newMemoryInUse != iMemoryInUse) {
        LOG.info("Pictures location changed -> restarting the C2S push for pictures");
        iAppUI->setClientPushFolderLocation(SOURCE_PHOTOS, newMemoryInUse);
    }
    
    return ETrue;
}


