/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2008 Funambol, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 *
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

#include "base/SymbianLog.h"
#include "base/symbianadapter.h"
#include "base/util/symbianUtils.h"
#include "base/util/stringUtils.h"
#include "base/FConnection.h"
#include "push/CTPService.h"

#include "utils.h"
#include "commonDefs.h"
#include "customization.h"
#include "PushEngine.h"
#include "SMSPushEngine.h"
#include "ClientSettings.h"

using namespace Funambol;


CPushEngine* CPushEngine::NewL(CFunambolAppUi* aAppUI)
{
    CPushEngine* self = CPushEngine::NewLC(aAppUI);
    CleanupStack::Pop(self);
    return self;
}

CPushEngine* CPushEngine::NewLC(CFunambolAppUi* aAppUI)
{
    CPushEngine* self = new(ELeave) CPushEngine();
    CleanupStack::PushL(self);
    self->ConstructL(aAppUI);
    return self;
}


CPushEngine::CPushEngine() 
{
    iAppUI      = NULL;
    iClientPush = NULL;
    iSMSPushEngine    = NULL;
}


void CPushEngine::ConstructL(CFunambolAppUi* aAppUI)
{
    iAppUI = aAppUI;
    CTPService* ctp = CTPService::getInstance();
    
    if (ctp->getConfig()->getCtpPort() == 0) {
        // Config corrupted (or missing) -> generate default and save it.
        createDefaultCTPConfig();
        ctp->getConfig()->saveCTPConfig();
    }
    
    // Now receive notifications for push messages.
    ctp->registerPushListener(*this);
    
    // Create the ClientPush. Don't start the listeners now.
    TInt err = KErrNone;
    TRAP(err, iClientPush = CClientPush::NewL(iAppUI);)
    if (err) {
        LOG.error("Error initializing ClientPush: %d", err);
        iAppUI->DisplayMessage(err, EErrorLevel);
    }

    // Create the SMS Push Engine. Don't start the listeners now.
    TRAP(err, iSMSPushEngine = CSMSPushEngine::NewL(this);)
    if (err) {
        LOG.error("Error initializing SMS push engine: %d", err);
        iAppUI->DisplayMessage(err, EErrorLevel);
    }
}

CPushEngine::~CPushEngine() 
{
    CTPService* ctp = CTPService::getInstance();
    ctp->stopCTP();
    
    delete iClientPush;
    delete iSMSPushEngine;
}


void CPushEngine::start()
{
    // Read the push type from config
    ClientSettings* config = ClientSettings::getInstance();
    const int pushType = config->getPushType();
    

    if (pushType & PUSH_TYPE_CTP) {
        // Try to setup an internet connection. in case, retry 3 times asking the user.
        if (!startConnection(MAX_CONNECTION_RETRY_ASKING_USER)) {
            return;
        }
        
        LOG.info("Starting CTP service...");
        CTPService* ctp = CTPService::getInstance();
        ctp->startCTP();
    }
    
    if (pushType & PUSH_TYPE_CP) {   
        //
        // Get the pictures location folder
        //
        StringBuffer picturesPath;
        TInt picturesStorage = ClientSettings::getInstance()->getPicturesStorage();
        bool driveAvailable = getImagesPath(picturesStorage, picturesPath);

        // Set the pictures path only if the selected drive is available.
        // If the drive is empty it's ok, because it's the first time (default = empty)
        if (driveAvailable || (picturesStorage == EStorageTypeUnset)) {
            iClientPush->setPicturesPath(picturesPath);
        } else {
            StringBuffer pictureDrive = getDriveName(picturesStorage);
            LOG.info("Pictures drive '%s' not available: local pictures changes won't be detected", pictureDrive.c_str());
        }

        //
        // Starts listening for CP notifications.
        //
        iClientPush->start();
    }

    if (pushType & PUSH_TYPE_SMS) {
        iSMSPushEngine->start();
    }
  
}

void CPushEngine::stop() 
{
    LOG.info("Stopping CTP service...");
    CTPService* ctp = CTPService::getInstance();
    ctp->stopCTP();
    
    // Stops listening for CP notifications.
    iClientPush->stop();

    // Stops listening for SMS push.
    iSMSPushEngine->stop();
}

void CPushEngine::restart() 
{
    stop();
    start();
}


TInt CPushEngine::getPushState()
{
    CTPService* ctp = CTPService::getInstance();
    return ctp->getCtpState();
}

bool CPushEngine::isPushRunning()
{
    // Read the push type from config
    ClientSettings* config = ClientSettings::getInstance();
    const int pushType = config->getPushType();
    
    if (pushType & PUSH_TYPE_CTP) {
        TInt state = getPushState();
        if (state == CTPService::CTP_STATE_DISCONNECTED) {
            return false;
        }
    }
    if (pushType & PUSH_TYPE_CP) {
        if (iClientPush->isRunning() == false) {
            return false;
        }
    }
    if (pushType & PUSH_TYPE_SMS) {
        if (iSMSPushEngine->isRunning() == false) {
            return false;
        }
    }
    
    // All active services up and running.
    return true;

}


void CPushEngine::onNotificationReceived(const ArrayList& serverURIList) 
{
    LOG.info("Push notification received: %d sources", serverURIList.size());
    RArray<TInt> sourceList;
    
    // Convert each ServerURI string into a source ID.
    for (int i=0; i < serverURIList.size(); i++) {
        StringBuffer* uri = (StringBuffer*)serverURIList[i];
        if ( !uri || !(uri->length()) ) {
            continue;
        }
        StringBuffer sourceName = getSourceNameFromServerURI(*uri);
        TInt sourceID = getSourceIDFromName(sourceName);
        if (sourceID != -1) {
            sourceList.Append(sourceID);
        }
    }
    
    if (sourceList.Count() > 0) {
        // Start the push-triggered sync!
        iAppUI->StartSync(sourceList, ESyncTriggeredPush);
    }
    sourceList.Close();
}

void CPushEngine::onNotificationReceived(RArray<TInt>& aSourceList)
{
    if (aSourceList.Count() > 0) {
        // Start the push-triggered sync!
        iAppUI->StartSync(aSourceList, ESyncTriggeredPush);
    }
}


void CPushEngine::onCTPError(const int errorCode, const int additionalInfo)
{
    LOG.debug("PushEngine: CTP error occurred: code %d", errorCode);
    
    switch (errorCode) 
    {
        case CTPService::CTP_ERROR_NOT_AUTHENTICATED:
        case CTPService::CTP_ERROR_UNAUTHORIZED:
        case CTPService::CTP_ERROR_AUTH_FORBIDDEN:
        {
            // Should not happen: auth is checked during config sync...
            iAppUI->DisplayMessage(ERR_AUTH_FAILED, EErrorLevel);
            break;
        }
        case CTPService::CTP_ERROR_CONNECTION_FAILED:
        {
            // It happens only when the max CTPRetry time is reached (15 min)
            iAppUI->DisplayMessage(ERR_CTP_NO_CONNECTION, EErrorLevel);
            break;
        }
        default:
            // Ignore other errors.
            break;
    }
}

void CPushEngine::createDefaultCTPConfig()
{
    LOG.debug("Generating default CTP configuration...");
    
    CTPConfig* ctpConfig = CTPService::getInstance()->getConfig();
    if (!ctpConfig) {
        return;
    }

    ClientSettings* config = ClientSettings::getInstance();
    StringBuffer hostName = ctpConfig->getHostName(config->getSyncURL());
   
    LOG.debug("CTP connection parameters:");
    ctpConfig->setUrlTo         (hostName);
    ctpConfig->setCtpCmdTimeout (CTP_CMD_TIMEOUT);
    LOG.debug("\tcmd timeout: %d", CTP_CMD_TIMEOUT);
    ctpConfig->setCtpConnTimeout(CTP_CONN_TIMEOUT);
    LOG.debug("\tconnection timeout: %d", CTP_CONN_TIMEOUT);
    ctpConfig->setCtpPort       (CTP_PORT);
    LOG.debug("\tport: %d", CTP_PORT);
    ctpConfig->setCtpReady      (CTP_READY_INTERVAL);
    LOG.debug("\tready interval: %d", CTP_READY_INTERVAL);
    ctpConfig->setCtpRetry      (CTP_RETRY_INTERVAL);
    LOG.debug("\tretry interval: %d", CTP_RETRY_INTERVAL);
    ctpConfig->setMaxCtpRetry   (CTP_MAX_RETRY);
    LOG.debug("\tmax retry: %d", CTP_MAX_RETRY);
    ctpConfig->setNotifyTimeout (CTP_NOTIFY_TIMEOUT);
    LOG.debug("\tnotify timeout: %d", CTP_NOTIFY_TIMEOUT);
    
    // -- not used --
    //ctpConfig->setQueuePush   (true);
    //ctpConfig->setPush        (1);
    //ctpConfig->setPolling     (0);
}



