/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2008 Funambol, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 *
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

#include <eikenv.h>

#include "base/util/StringBuffer.h"
#include "base/util/symbianUtils.h"
#include "base/util/timeUtils.h"
#include "base/Log.h"
#include "base/util/PropertyFile.h"
#include "base/FConnection.h"
#include "spds/SyncManager.h"
#include "client/SyncClient.h"

#include "commonDefs.h"
#include "customization.h"
#include "ConfigSyncSource.h"
#include "PicturesSyncSource.h"
#include "PicturesSyncSourceSyncMLData.h"

#include "ClientSettings.h"
#include "utils.h"
#include "SyncActiveObject.h"

SyncActiveObject*       SyncActiveObject::syncAO                 = NULL;
TimeoutTimer*           SyncActiveObject::tt                     = NULL;
CSyncEngine*            SyncActiveObject::iSyncEngine            = NULL;
SyncActiveObject::State SyncActiveObject::iState                 = ENone;
TInt                    SyncActiveObject::iSourceId              = -1;
int                     SyncActiveObject::itemsSuccesfullySynced = 0;

class TimeoutTimer : public CActive
{
public:

    static TimeoutTimer* TimeoutTimer::NewL()
    {
        TimeoutTimer* pSelf = new(ELeave) TimeoutTimer();
        CleanupStack::PushL(pSelf);
        pSelf->ConstructL();
        CleanupStack::Pop();

        return pSelf;
    }

    void TimeoutTimer::Start(const TInt seconds = 0)
    {
        Cancel();

        if (seconds) {
            iTimeout = seconds;
        }
        
        iTimer.After(iStatus, iTimeout * 1000000);
        SetActive();
    }

    TimeoutTimer::~TimeoutTimer()
    {
        Cancel();
        iTimer.Close();
    }

private:

    TimeoutTimer::TimeoutTimer() : CActive(EPriorityStandard),
                                   iCancelled(EFalse),
                                   iTimeout(SYNC_PROCESS_TIMEOUT)
    {
    }

    void TimeoutTimer::ConstructL()
    {
        iTimer.CreateLocal();
        CActiveScheduler::Add(this);
    }

    void TimeoutTimer::RunL()
    {
        if (!iCancelled) {
            LOG.debug("TimeoutTimer triggered, ending sync");
            SyncActiveObject::CancelSync();
        }
        // Self destroy
        delete this;
    }

    void TimeoutTimer::DoCancel()
    {
        iTimer.Cancel();
        iCancelled = ETrue;
    }

private:
    RTimer            iTimer;
    TBool             iCancelled;
    TInt              iTimeout;
};



SyncActiveObject* SyncActiveObject::NewL(const TInt aSourceId,  const StringBuffer& aCacheFile)
{
    SyncActiveObject* self = SyncActiveObject::NewLC(aSourceId, aCacheFile);
    CleanupStack::Pop(self);
    return self;
}

SyncActiveObject* SyncActiveObject::NewLC(const TInt aSourceId,  const StringBuffer& aCacheFile)
{
    SyncActiveObject* self = new(ELeave) SyncActiveObject(aSourceId, aCacheFile);
    self->ConstructL();
    CleanupStack::PushL(self);
    return self;
}

void SyncActiveObject::syncLD(CSyncEngine* aSyncEngine,  const TInt aSourceId, const StringBuffer&  aCacheFile)
{
    iSyncEngine = aSyncEngine;
    tt = NULL;
    
    // We don't want more than 1 sync instance
    if (CheckSyncInProgress()) {
        LOG.info("A sync is already in progress");
        return;
    }

    // Timeout on the entire sync process (0 = no timeout)
    TInt timeout = SYNC_PROCESS_TIMEOUT;
    
    if (aSourceId == SOURCE_CONFIG) { 
        timeout = CREDENTIALS_CHECK_TIMEOUT;    // 2 min
    }
    
    LOG.info("Starting sync (timeout = %d seconds)", timeout);
    syncAO = SyncActiveObject::NewL(aSourceId, aCacheFile);
    syncAO->sync();

    if (timeout > 0) {
        LOG.debug("Starting timeout timer");
        tt = TimeoutTimer::NewL();
        tt->Start(timeout);
    }
}


void SyncActiveObject::syncEnded(const bool aborted, const TInt status)
{
    LOG.debug("Sync ended");
    if (!aborted) {
        LOG.debug("Sync terminated normally. code = %d", status);
        if (tt) {
            tt->Cancel();
        }
        syncAO = NULL;
    } else {
        // Timeout or Cancel-Sync
        LOG.info("Sync has been aborted");
        if (syncAO) {
            syncAO->Stop();
        }
    }
    
    // If we get a timeout, but the sync has already succesfully finished, then
    // we simply ignore it
    if (!(aborted && syncAO == NULL)) {
        iSyncEngine->syncEnded(iSourceId, status, itemsSuccesfullySynced);
    }
}


bool SyncActiveObject::CheckSyncInProgress()
{
    if (iState == ESync) {
        return true;
    }
    return false;
}


SyncActiveObject::SyncActiveObject(const TInt aSourceId, const StringBuffer& aCacheFile)
                                       : CActive(CActive::EPriorityStandard),
                                       iCacheFile(aCacheFile),
                                       status(500)
{
    iSourceId = aSourceId; // initialize static data member
}

void SyncActiveObject::ConstructL()
{
    CActiveScheduler::Add(this);
    iState = ENone;
}

SyncActiveObject::~SyncActiveObject()
{
    Cancel();
}

void SyncActiveObject::Stop() {
    iState = EStopped;
    Cancel();
}

// This is static public method to cancel the sync from outside
void SyncActiveObject::CancelSync() 
{
    if (syncAO && syncAO->CheckSyncInProgress()) 
    {
        if (getLastErrorCode() != KErrCancel) {
            LOG.info("Aborting current sync");
            setError(KErrCancel, "Sync aborted by the user");
            
            // This is to interrupt the http transaction, in case we were stuck there.
            FConnection::getInstance()->closeConnection();
            
            // Update the UI with "Canceling Sync..."
            iSyncEngine->updateAppUISourceStatus(iSourceId, ESyncStatusCanceling);
        }
    }
}



void SyncActiveObject::sync()
{
    if (iState != EStopped) {
        LOG.debug("SyncActiveObject about to start");
        TRequestStatus* status = &iStatus;
        User::RequestComplete(status, KErrNone);
        iState = ESync;
        SetActive();
    } else {
        LOG.debug("The sync cannot start as it has been stopped already");
    }
}


void SyncActiveObject::syncL(void)
{
    ClientSettings* config = ClientSettings::getInstance();
    SyncClient client;
	
	StringBuffer sourceName = getSourceNameFromID(iSourceId);
	if (sourceName.empty()) {
	    LOG.error("unrecognized source id %d", iSourceId);
	    return;
	}
	wchar_t* wSourceName = toWideChar(sourceName);
    
    LOG.debug("starting synchronization of source:  %s", sourceName.c_str());
    itemsSuccesfullySynced = 0;
    
    if (iSourceId == SOURCE_CONFIG) {  
        StringBuffer ssPath;
        PropertyFile* cache = new PropertyFile(iCacheFile.c_str());
        
        ssPath.sprintf("%s/", APPLICATION_URI);
    	ConfigSyncSource source(wSourceName, ssPath, config->getSyncSourceConfig(sourceName), cache);
    	SyncSource* ssArray[] = { &source, NULL };
    	// We remove all the old values, as they are going to be replaced. This is
    	// already done by doing a refresh from server, but in order to be sure
    	// we do it also here
    	source.removeAllItems();
    	source.setPreferredSyncMode(SYNC_REFRESH_FROM_SERVER);

    	// ***************
		// Start the sync!
    	// ***************
		status = client.sync(*config, ssArray);

		if (status) {
			LOG.error("error in  source '%s' synchronization", sourceName.c_str());
		} else {
			// Save config params only if sync successful!
			// This is a bit tricky. The sync changed the configuration on disk,
			// but the config object has not been updated. We need to reread the
			// config in order to refresh the config (which acts like a cache)
			config->readConfigSyncSourceSettings();
			// This is needed because we need to push the parameters into the
			// native configuration (the ConfigSyncSource updates only the DM)
			config->save();
		}
    } 
    
    else if (iSourceId == SOURCE_PHOTOS) {
        
        // Read the path where images are stored
        StringBuffer imagesDir;
        TInt pictureStorage = config->getPicturesStorage();
        bool driveAvailable = getImagesPath(pictureStorage, imagesDir);
        
        // Continue only if the desired drive is available (should not happen: there's a check in SyncEngine::syncPictures)
        // If the drive is empty it's ok, because it's the first time (default = empty)
        if (driveAvailable || (pictureStorage == EStorageTypeUnset)) {
            
            LOG.info("Reading images from dir: %s", imagesDir.c_str());
            
            SyncSource* pss;
            SyncSourceConfig* pssc = config->getSyncSourceConfig(sourceName);
            
            // Here we switch between a PictureSS with upload based on HTTP OR syncML
            // checking the server capabilities.
            if (config->getServerMediaHttpUpload()) {
                LOG.debug("Using HTTP UPLOAD mode (data via HTTP)");
                pss = new PicturesSyncSource(wSourceName, pssc, imagesDir);
            } else {
                LOG.debug("Using backward compatibility mode (data via syncML)");
                pss = new PicturesSyncSourceSyncMLData(wSourceName, pssc, imagesDir);
            }
            SyncSource* ssArray[] = { pss, NULL };
    
            // ***************
            // Start the sync!
            // ***************
            status = client.sync(*config, ssArray);
            
            if (status) {
                LOG.error("error in  source '%s' synchronization", sourceName.c_str());
            } else {
                LOG.info("saving conf for source '%s'", sourceName.c_str());
                config->save();
            }
            delete pss;
        }
        else {
            StringBuffer pictureDrive = getDriveName(pictureStorage);
            LOG.info("Pictures drive '%s' not available: aborting sync!", pictureDrive.c_str());
            client.getSyncReport()->setLastErrorCode(KErrCancel);
            client.getSyncReport()->setLastErrorMsg("Pictures location not available");
        }
    } 
    else {
        LOG.error("unable to handle synchronization of source %s", sourceName.c_str());
    }
    delete [] wSourceName; 
    
    // Store the # of items succesfully synced, from the report. 
    // It's used by the client to know if at least one item was synced (status interrupted / failed).
    SyncSourceReport* ssr = client.getSyncReport()->getSyncSourceReport(sourceName);
    if (ssr) {
        if (config->getServerMediaHttpUpload()) {
            // new upload http: the number of items added/modified is the number of items really uploaded
            itemsSuccesfullySynced  = ssr->getItemReportSuccessfulCount(SERVER, HTTP_UPLOAD);
            itemsSuccesfullySynced += ssr->getItemReportSuccessfulCount(SERVER, COMMAND_DELETE);
        } else {
            itemsSuccesfullySynced = ssr->getTotalSuccessfulCount();
        }
    }
   
    // Print sync results.
    StringBuffer res("");
    client.getSyncReport()->toString(res);
    LOG.info("\n%s", res.c_str());
    
    // If the global error is 0 (none) but the source has failed, 
    // set the return code (the source error is lost with the SyncClient here!)
    // NOTE: this is possible because we sync ONLY 1 source at time.
    if ((status == ERR_NONE) && (ssr->getState() == SOURCE_ERROR)) {
        status = ssr->getLastErrorCode();
    }
}

void SyncActiveObject::DoCancel(void)
{
    LOG.debug("SyncActiveObject::DoCancel");
    iState = ECancelled;
}


void SyncActiveObject::RunL(void)
{
    //LOG.debug("SyncActiveObject::RunL");

    switch(iState) {
        case ESync:
        {
            // Start the sync
            TRAPD(err, syncL();)
            if (err) {
                LOG.error("Error in sync (code %d)", err);
                iState = EStopped;
                syncEnded(true, err);
                delete this;
                break;
            }

            // Move to the done status (if we were not cancelled)
            if (iState == ESync) {
                iState = EDone;
            }
            TRequestStatus* s = &iStatus;
            User::RequestComplete(s, KErrNone);
            SetActive();
            break;
        }
        case ECancelled:
            LOG.debug("SyncActiveObject Cancelled");
            syncEnded(true, KErrCancel);                // Code -3 = sync canceled
            delete this;
            break;
        case EDone:
            syncEnded(false, status);                   // Return the sync result code
            delete this;
            break;
        default:
            break;
    }
}

