/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2008 Funambol, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 *
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

#include <eikenv.h>
#include <aknglobalnote.h>
#include <syncmlerr.h>

#ifdef FAKE_IMAP_ACCOUNT
#include "FakeImapEmailEngine.h"
#else
#include "ImapEmailEngine.h"
#endif

#include "base/util/symbianUtils.h"
#include "base/util/timeUtils.h"
#include "base/SymbianLog.h"
#include "base/util/PropertyFile.h"
#include "base/adapter/PlatformAdapter.h"
#include "base/FConnection.h"
#include "spds/SyncManager.h"
#include "client/SyncClient.h"
#include "client/FileSyncSource.h"
#include "updater/Updater.h"

#include "event/ManageListener.h"
#include "event/SetListener.h"
#include "event/SymbianSyncListener.h"
#include "event/SymbianSyncItemListener.h"
#include "event/SymbianSyncSourceListener.h"
#include "event/SymbianTransportListener.h"

#include "FunambolAppView.h"
#include "commonDefs.h"
#include "customization.h"
#include "ClientSettings.h"
#include "SyncScheduler.h"
#include "ClientPush.h"
#include "UpdateManager.h"
#include "SyncActiveObject.h"
#include "utils.h"

#include "SyncEngine.h"


using namespace Funambol;

typedef struct SyncApp {
    TInt          sourceId;       /**< The corresponding source ID */
    TBuf<32>      dataSource;     /**< The dataSource for this task */
} SyncApps;


#define NUM_PIM_SYNC_APPS                   3

static SyncApps SyncTasks[NUM_PIM_SYNC_APPS] = {
//  sourceId                dataSource
//  -------------------------------------------
    {SOURCE_CONTACTS,    _L("C:Contacts.cdb")},
    {SOURCE_CALENDAR,    _L("C:Calendar")    },
    {SOURCE_NOTES,       _L("c:Notepad.dat") }
};


ListenerPool::~ListenerPool()
{
    releaseAllListeners();
}

void ListenerPool::setListeners()
{
    LOG.debug("setting listeners");

    setSyncListener      (new (ELeave) SymbianSyncListener      (iSyncEngine, iSourceId));
    setSyncItemListener  (new (ELeave) SymbianSyncItemListener  (iSyncEngine));
    setSyncSourceListener(new (ELeave) SymbianSyncSourceListener(iSyncEngine));
    setTransportListener (new (ELeave) SymbianTransportListener (iSyncEngine, iSourceId));
}

void ListenerPool::releaseAllListeners()
{
    LOG.debug("releasing listeners");
    ManageListener::releaseAllListeners();
}

CSyncEngine* CSyncEngine::NewL(CFunambolAppUi* aAppUI)
{
    CSyncEngine* self = CSyncEngine::NewLC(aAppUI);
    CleanupStack::Pop(self);
    return self;
}

CSyncEngine* CSyncEngine::NewLC(CFunambolAppUi* aAppUI)
{
    CSyncEngine* self = new(ELeave) CSyncEngine();
    CleanupStack::PushL(self);
    self->ConstructL(aAppUI);
    return self;
}

/*
 * Class constructor
 */
CSyncEngine::CSyncEngine() : CActive(CActive::EPriorityStandard),
                             iActiveProfileID(-1),
                             lastPicturesSyncError(0),
                             iConnectingBeforeSync(false)
{
}


void CSyncEngine::ConstructL(CFunambolAppUi* aAppUI)
{
    iAppUI = aAppUI;

    iCurrentSource = 0;
    iSyncQueue.Reset();

    firstSyncAfterStartup = true;
    iCreateProfileRetry   = 0;

    // We must originate and complete an asynchronous requests
    // inside the same thread, so we save its ID.
    RThread thisThread;
    iMainThreadID = thisThread.Id();
    //LOG.debug("thread id = %d", iMainThreadID);

    // Adds this object to the Active Scheduler
    CActiveScheduler::Add(this);

    // Opens SyncML session.
    InitializeSessionL();

    // Creates the scheduler.
    iScheduler = CSyncScheduler::NewL(this);

    // Email engine is built on demand
    emailEngine = NULL;
    // photo sync
    iListenerPool = NULL;
}

void CSyncEngine::InitializeSessionL()
{
    //
    // open syncML session
    //
    LOG.debug("Opening SyncML session...");
    iSyncMLSession.OpenL();


  /*// *Note*: THIS IS MANDATORY TO USE THE SAME SESSION FROM DIFFERENT THREADS!
    // After calling this function the RSyncMLSession object
    // may be used by threads other than than the one that created it.
    TInt err = iSyncMLSession.ShareAuto();
    if (err != KErrNone) {
        LOG.error("CSyncEngine error: unable to share RSyncMLSession (code %d)", err);
        User::Leave(ERR_SHARE_SYNCML_SESSION);
    }*/

    // Request event notifications
    iSyncMLSession.RequestProgressL(*this);
    iSyncMLSession.RequestEventL(*this);

    iActiveProfileID = -1;
}

bool CSyncEngine::ReInitializeSessionL()
{
    TInt err = KErrNone;

    Funambol::LOG.debug("%s: closing SyncML session...", __FUNCTION__);
    iSyncMLSession.Close();

    // open syncML session
    Funambol::LOG.debug("Opening SyncML session...");
    TRAP(err, iSyncMLSession.OpenL() );

    if (err) {
        LOG.error("%s: error opening syncML session", __FUNCTION__);

        return false;
    }


    // Request event notifications
    iSyncMLSession.RequestProgressL(*this);
    iSyncMLSession.RequestEventL(*this);

    iActiveProfileID = -1;

    LOG.debug("%s: resetting active profile ID", __FUNCTION__);
    getActiveProfileID();

    return true;
}


TInt CSyncEngine::getActiveProfileID() {

    if (iActiveProfileID != -1) {
        return iActiveProfileID;
    }

    // get the list of available profiles
    RArray<TSmlProfileId> profilesArray;
    iSyncMLSession.ListProfilesL(profilesArray, ESmlDataSync);

    RBuf profileName;
    LOG.debug("Loading Sync profile...");
    profileName.Assign(wcharToNewBuf(WSYNC_PROFILE_NAME));

    // search for desired profile
    for(TInt count=0; count<profilesArray.Count(); ++count)
    {
        RSyncMLDataSyncProfile profile;
        profile.OpenL(iSyncMLSession, profilesArray[count], ESmlOpenReadWrite);
        if (profile.DisplayName().Compare(profileName) == 0)
        {
            iActiveProfileID = profile.Identifier();
            profile.Close();
            LOG.debug("Profile found: id = %d", iActiveProfileID);
            break;
        }
        profile.Close();

        /* Feature not supported... :(
        RSyncMLSettings settings;
        settings.OpenL(iSyncMLSession);
        settings.SetValueL(ESmlSettingDisplayBgResult, ESmlDisplayAllResults);
        settings.Close(); */
    }
    profileName.Close();
    profilesArray.Close();


    if(iActiveProfileID == -1) {
        //
        // No active profile found
        //
        if (iCreateProfileRetry < 1) {
            // Create the profile and try again (recursive call, only 1 time!)
            LOG.info("SyncML profile '%s' not found: creating it...", SYNC_PROFILE_NAME);
            iCreateProfileRetry ++;
            if (ClientSettings::getInstance()->createSyncMLProfile()) {
                getActiveProfileID();
            }
            else {
                LOG.error("Cannot create SyncML profile");
            }
        }
        else {
            LOG.error("SyncML profile '%s' not found", SYNC_PROFILE_NAME);
        }
    }
    return iActiveProfileID;
}


/*
 * Class destructor
 */
CSyncEngine::~CSyncEngine()
{
    //LOG.debug("Closing CSyncEngine");
    CActive::Cancel();
    iSyncMLSession.Close();
    iSyncQueue.Close();
    delete iScheduler;
}


/*
 * Sync an array of sources (exclude config)
 */
void CSyncEngine::StartSync(RArray<TInt> &aSourceIDList, const TSyncRunners from)
{
    LOG.debug("CSyncEngine::StartSync - from %s, %d sources", getSyncRunnerName(from).c_str(), aSourceIDList.Count());

    // Fix for pictures: don't sync if last time quota was exceeded (code 420).
    // BTW manual syncs are always allowed.
    // Note: if app is restarted, the lastPicturesSyncError will be reset to 0.
    if ( (from != ESyncTriggeredManual) && (lastPicturesSyncError == ERR_SERVER_QUOTA_EXCEEDED) ) {
        if (removeFromRArray(SOURCE_PHOTOS, aSourceIDList)) {
            LOG.debug("Source pictures will not run (last error is quota exceed). %d sources left", aSourceIDList.Count());
        }
    }
    if (aSourceIDList.Count() == 0) {
        return;
    }


    //
    // Here we take different actions, based on who triggered the startsync.
    //
    switch(from)
    {
        // ****** MANUAL SYNC ******
        case ESyncTriggeredManual:
        {
            if (CheckSyncInProgress()) {
                // A sync is already running
                iAppUI->DisplayMessage(INFO_SYNC_IN_PROGRESS, EInfoLevel);
                return;
            }
            LOG.debug("StartSync: manual (%d sources)", aSourceIDList.Count());
            break;
        }
        // ****** S2C PUSH (CTP) ******
        case ESyncTriggeredPush:
        {
            LOG.debug("StartSync: S2C push (%d sources)", aSourceIDList.Count());
            break;
        }
        // ****** C2S PUSH (CLIENT PUSH) ******
        case ESyncTriggeredClientPush:
        {
            // We discard the CP sync if the source is under sync and
            // during 'receiving' status...
            if (isSourceUnderSyncAndReceiving((TInt)aSourceIDList[0])) {
                LOG.debug("CP notification discarded: source %d is under sync", aSourceIDList[0]);
                return;
            }
            ClientSettings* config = ClientSettings::getInstance();
            TInt interval = config->getCpInterval();

            // Set the source list for the scheduler, merge the list.
            iScheduler->setSourceList(aSourceIDList, true);

            // Enable ripetitive scheduler for CP: if sync fails, it will be
            // rescheduled again. If sync success, scheduler will be disabled
            // (so outstanding syncs will be discarded).
            iScheduler->enable(true, interval);
            LOG.debug("StartSync: C2S push (%d sources) - scheduled at %d sec", aSourceIDList.Count(), interval);
            return;
        }
        // ****** Scheduler (Active Object's RunL) ******
        case ESyncTriggeredScheduled:
        {
            LOG.debug("StartSync: from sync scheduler (%d sources)", aSourceIDList.Count());
            break;
        }
        default:
            return;
    }

    //
    // Adds all sources to the sync queue.
    //
    for (TInt i=0; i<aSourceIDList.Count(); i++) {
        AddToSyncQueue((TInt)aSourceIDList[i]);
    }

    // Starts sync the first source from the queue.
    syncFromQueue();
}



// Syncs a single data source.
void CSyncEngine::StartSync(const TInt aSourceID, const TSyncRunners from)
{
    RArray<TInt> sourceIDList;
    sourceIDList.Append(aSourceID);
    StartSync(sourceIDList, from);
    sourceIDList.Close();
}


void CSyncEngine::syncFromQueue()
{
    // We MUST start PIM sync from the same thread that opened the SyncML session!
    RThread thisThread;
    TInt thisThreadID = thisThread.Id();
    if (thisThreadID != iMainThreadID) {
        RequestAsyncAction(ESyncFromQueue);
        return;
    }

    if (iSyncQueue.Count() == 0) {
        LOG.debug("No more sources in sync queue");
        return;
    }

    // If a sync is already in progress, there's no need to
    // trigger another one (we want only 1 source syncing at time).
    // Remember the next source is always triggered when the sync ends!
    if (CheckSyncInProgress()) {
        LOG.debug("Sync already in progress (source %d)", iCurrentSource);
        return;
    }

    //
    // Only the first sync after each startup, we force a sync of config.
    // The sync is hidden in the first source "connecting" phase (no user perception).
    // This is necessary to force a GET of Server caps, as they could be lost if syncing
    // using the native sync client.
    //
    hiddenConfigSync = false;
    if (firstSyncAfterStartup) {
        LOG.debug("This is the first sync after Client startup");
        if (InsertConfigInSyncQueue()) {
            LOG.info("Added config source to sync queue (to GET Server caps)");
            hiddenConfigSync = true;
        }
        // Anyway, force the GET of Server caps (only this session!).
        ClientSettings::getInstance()->setForceServerDevInfo(true);
    }
    LOG.info("Sync queue contains %d sources", iSyncQueue.Count());


    // Get the first element and remove it from queue.
    iCriticalSyncQueue.Wait();
    TInt sourceID = (TInt)iSyncQueue[0];
    iSyncQueue.Remove(0);
    iCriticalSyncQueue.Signal();

    // Safe check: the source MUST be visible and enabled.
    // (it may happen, for CP or CTP push)
    if (sourceID != SOURCE_CONFIG) {
        if (!isSourceVisible(sourceID) || !isSourceEnabled(sourceID)) {
            LOG.debug("Aborting sync of '%s' (source is disabled)", getSourceNameFromID(sourceID).c_str());
            syncFromQueue();    // Skip to next source. Warning: recursive call!
            return;
        }
    }

    // ------------------------
    // Ok... kick-off the SYNC!
    // ------------------------
    SetCurrentSource(sourceID);
    LOG.info("*** STARTING SYNC OF '%s' ***", getSourceNameFromID(sourceID).c_str());
    
    // Update the UI: while trying to setup the connection, the UI shows "connecting..."
    if (hiddenConfigSync) {
        updateNextSourceStatus(ESyncStatusConnecting);
    } else {
        iAppUI->UpdateSourceStatus(sourceID, ESyncStatusConnecting);
    }

    // Make sure we have an active connection - if not, retry 3 times asking the user.
    // This also fixes bug #9109: to avoid the native client drop the connection after the sync.
    iConnectingBeforeSync = true;
    if (!startConnection(MAX_CONNECTION_RETRY_ASKING_USER)) {
        iConnectingBeforeSync = false;
        return;
    }
    
    iConnectingBeforeSync = false;
    iTotalServerItems = 0;
    iTotalClientItems = 0;

    if (sourceID == SOURCE_MAIL) {
        syncEmail();
    }
    else if (sourceID == SOURCE_CONFIG) {
        syncConfig();
    }
    else if (sourceID == SOURCE_PHOTOS) {
        syncPhotos();
    }
    else {
        syncPIMSource(sourceID);
    }
}


TBool CSyncEngine::CheckSyncInProgress()
{
    // First of all, check if we're setting up a connection for a sync
    if (iConnectingBeforeSync) {
        return ETrue;
    }
    
    switch (iCurrentSource)
    {
        case SOURCE_CONFIG:
        case SOURCE_PHOTOS:
        {
            if (SyncActiveObject::CheckSyncInProgress()) {
                //LOG.debug("Config sync is in progress");
                return ETrue;
            }
            return EFalse;
        }

        case SOURCE_MAIL:
        {
            if (emailEngine && emailEngine->CheckSyncInProgress()) {
                return ETrue;
            }
            return EFalse;
        }

        //case SOURCE_CONTACTS:
        //case SOURCE_CALENDAR:
        //case SOURCE_NOTES:
        default:
        {
            if (getActiveProfileID() != -1) {
                TSmlJobId jobID = GetCurrentJobID();
                if (jobID != -1) {
                    return ETrue;   // sync is running
                }
            }
            return EFalse;
        }
    }
    return EFalse;
}


TInt CSyncEngine::StopSync()
{
    LOG.debug("CSyncEngine::StopSync (current source = %s)", getSourceNameFromID(iCurrentSource).c_str());

    // Clear the sync queue!
    resetSyncQueue();

    // Connecting phase is now asynchronous: cancel it
    if (iConnectingBeforeSync) {
        FConnection::getInstance()->closeConnection();
        return 1;
    }

    // Stop the current sync
    switch (iCurrentSource)
    {
        case SOURCE_PHOTOS:
        case SOURCE_CONFIG:
        {
            if (SyncActiveObject::CheckSyncInProgress()) {
                SyncActiveObject::CancelSync();
                return 1;
            }
            return 0;
        }
        case SOURCE_MAIL:
        {
            if (emailEngine && emailEngine->CheckSyncInProgress()) {
                emailEngine->CancelOperation();
                return 1;
            }
            return 0;
        }

        //case SOURCE_CONTACTS:
        //case SOURCE_CALENDAR:
        //case SOURCE_NOTES:
        default:
        {
            if (getActiveProfileID() == -1) {
                // No sync profile, available, there is nothing we can stop
                return 0;
            }

            TSmlJobId jobID = GetCurrentJobID();
            if (jobID == -1) {
                return 0;    // Nothing to do
            }
            LOG.debug("stopping a PIM sync job: job ID = %d", jobID);

            // Stop the running job.
            TInt error = StopJob(jobID);
            if (!error) {
                return 1;   // Ok, job stopped.
            }
            return 0;
        }
    }
}

void CSyncEngine::resetSyncQueue()
{
    iCriticalSyncQueue.Wait();
    iSyncQueue.Reset();
    iCriticalSyncQueue.Signal();
}


TSmlJobId CSyncEngine::GetCurrentJobID()
{
    TInt err = KErrNone;
    TSmlJobId jobID;
    TSmlUsageType jobType;

    // Returns '-1' if no job running!  Note: Doc is wrong! :(
    TRAP(err, iSyncMLSession.CurrentJobL(jobID, jobType);)
    if (err) {
        LOG.error("Error getting job running (code %d): trying to reinitialize syncML session", err);
        // try to reinitialize session with Symbian syncML server
        if (ReInitializeSessionL() == false) {
            return KErrNotFound;
        }

        TRAP(err, iSyncMLSession.CurrentJobL(jobID, jobType); )

        if (err) {
            LOG.error("Error getting job running: %d", err);
            return KErrNotFound;
        }
    }

    return jobID;
}



void CSyncEngine::OnSyncMLSyncProgress(TStatus aStatus, TInt aInfo1, TInt aInfo2)
{
    // Save the actual syncml status, used in OnSyncMLDataSyncModifications()
    switch(aStatus)
    {
        // These events are forwarded to AppUI, for proper actions
        // Here is a mapping between 'TStatus' defined by Symbian SyncML and our 'TCurrentSyncStatus'.
        case ESmlConnecting:
            iCurrentSyncStatus = ESyncStatusConnecting;
            LOG.info("Connecting to SyncML Server...");
            break;
        case ESmlConnected:
            iCurrentSyncStatus = ESyncStatusConnected;
            LOG.debug("Connected to SyncML Server");
            break;
        case ESmlDisconnected:
            iCurrentSyncStatus = ESyncStatusDisconnected;
            LOG.debug("Disconnected from SyncML Server");
            break;
        case ESmlLoggingOn:
            iCurrentSyncStatus = ESyncStatusLoggingOn;
            LOG.debug("Logging on...");
            break;
        case ESmlLoggedOn:
            iCurrentSyncStatus = ESyncStatusLoggedOn;
            LOG.debug("Logged on");
            break;
        case ESmlSendingModificationsToServer:
            iCurrentSyncStatus = ESyncStatusSendingData;
            LOG.info("Sending modifications...");
            break;
        case ESmlReceivingModificationsFromServer:
            iCurrentSyncStatus = ESyncStatusReceivingData;
            LOG.info("Receiving modifications...");
            break;
        case ESmlSendingMappingsToServer:
            iCurrentSyncStatus = ESyncStatusMappings;
            LOG.info("Sending mappings...");
            break;
        case ESmlCompleted:
            LOG.debug("Synchronization ended.");
            // NOTE: don't update the UI, we still don't know if there was an error.
            //       A SyncMLSessionEvent will be generated for this.
            return;

        // Other events are ignored!
        default:
            return;
    }

    // Send the notification to the AppUI, to resfresh the View.
    iAppUI->UpdateSourceStatus(iCurrentSource, iCurrentSyncStatus);
}


void CSyncEngine::OnSyncMLDataSyncModifications(TInt aTaskId,
                                                const TSyncMLDataSyncModifications& aClientMod,
                                                const TSyncMLDataSyncModifications& aServerMod)
{

    if (iCurrentSyncStatus == ESyncStatusSendingData) {
        //
        // Client -> Server
        //
        iTotalServerItems += aServerMod.iNumAdded + aServerMod.iNumDeleted +
                             aServerMod.iNumMoved + aServerMod.iNumFailed  + aServerMod.iNumReplaced;

        iAppUI->UpdateSourceStatus(iCurrentSource, ESyncStatusSendingItemNumber, iTotalServerItems);
    }
    else if (iCurrentSyncStatus == ESyncStatusReceivingData) {
        //
        // Server -> Client
        //
        iTotalClientItems += aClientMod.iNumAdded + aClientMod.iNumDeleted +
                             aClientMod.iNumMoved + aClientMod.iNumFailed  + aClientMod.iNumReplaced;

        iAppUI->UpdateSourceStatus(iCurrentSource, ESyncStatusReceivingItemNumber, iTotalClientItems);
    }

}

void CSyncEngine::OnSyncMLSyncError(TErrorLevel aErrorLevel, TInt aError, TInt aTaskId, TInt aInfo1, TInt aInfo2)
{
    LOG.error("SyncML sync error occurred: code %d", aError);
    iCurrentSyncStatus = ESyncStatusError;
    iAppUI->UpdateSourceStatus(iCurrentSource, ESyncStatusError, aError);

    // Not here: works only on N91.
    // This call is moved to OnSyncSessionEvent()
    //ManageSyncMLErrors(aError, aInfo1);
}


void CSyncEngine::ManageSyncMLErrors(const TInt aError, const TInt aAdditionalData)
{
    switch(aError)
    {
        case SyncMLError::KErrTransportRequest:
        case SyncMLError::KErrTransportHeader:
        case SyncMLError::KErrTransportSvrNoUri:
        case SyncMLError::KErrTransportSvrError:
        case SyncMLError::KErrTransportDisconnected:
        case SyncMLError::KErrConnectionError:
        case SyncMLError::KErrCannotCommunicateWithServer:
        case SyncMLError::KErrPhoneOnOfflineMode:
        {
            LOG.error("Connection error (code %d). Stopping the sync.", aError);
            resetSyncQueue();
            iAppUI->DisplayMessage(ERR_CONNECTING, EErrorLevel);
            break;
        }
        case SyncMLError::KErrTransportAuthenticationFailure:
        case SyncMLError::KErrAuthenticationFailure:
        {
            LOG.error("Authentication failed (code %d). Stopping the sync.", aError);
            resetSyncQueue();
            iAppUI->DisplayMessage(ERR_AUTH_FAILED, EErrorLevel);
            break;
        }
        case SyncMLError::KErrServerDatabaseNotFound:
        {
            LOG.error("Source not found on Server (code %d).", aError);
            iAppUI->DisplayMessage(ERR_SOURCE_NOT_FOUND, EErrorLevel);
            break;
        }
        case SyncMLError::KErrDatabaseAdapterIndexError:
        case SyncMLError::KErrSessionIdNotValid:
        case SyncMLError::KErrVerDTDMissing:
        case SyncMLError::KErrProtoMissing:
        case SyncMLError::KErrTargetLocMissing:
        case SyncMLError::KErrMappingInfoMissing:
        case SyncMLError::KErrMappingInfoNotAcked:
        case SyncMLError::KErrVersionDTDNotSupported:
        case SyncMLError::KErrProtoNotSupported:
        case SyncMLError::KErrStatusCommandForCommandNotSent:
        case SyncMLError::KErrStatusCommandNotTheSame:
        case SyncMLError::KErrMissingStatusCommands:
        case SyncMLError::KErrFailedToAddRecord:
        case SyncMLError::KErrFailedToReplaceRecord:
        case SyncMLError::KErrFailedToDeleteRecord:
        case SyncMLError::KErrObjectTooLargeToSendToServer:
        case SyncMLError::KErrUserInteractionFailed:
        case SyncMLError::KErrStatusSizeMismatch:
        case SyncMLError::KErrFailedToCloseDatabase:
        case SyncMLError::KErrGeneral:
        case SyncMLError::KErrProfileAccessDenied:
        case SyncMLError::KErrObjectInUse:
        case SyncMLError::KErrServerIdNotMatching:
        {
            LOG.error("SyncML error (code %d).", aError);
            iAppUI->DisplayMessage(ERR_SYNCML_GENERIC_FAILURE, EErrorLevel);
            break;
        }
        case SyncMLError::KErrSourceLocMissing:
        case SyncMLError::KErrClientDatabaseNotFound:
        {
            LOG.error("Source not found on Client (code %d).", aError);
            iAppUI->DisplayMessage(ERR_SYNCML_GENERIC_FAILURE, EErrorLevel);
            break;
        }
        case SyncMLError::KErrSlowSync:
        {
            // TODO: display a warning popup?
            LOG.info("Server requested a SLOW SYNC for '%s'", getSourceNameFromID(iCurrentSource).c_str());
            break;
        }
        default:
        {
            LOG.error("Unexpected SyncML error (code %d)", aError);
            iAppUI->DisplayMessage(ERR_SYNCML_GENERIC_FAILURE, EErrorLevel);
            break;
        }
    }
}


void CSyncEngine::OnSyncMLSessionEvent(TEvent aEvent, TInt aIdentifier, TInt aError, TInt aAdditionalData)
{
    LOG.debug("SyncML session event: event=%d, id=%d, err=%d, data=%d",
                                     aEvent, aIdentifier, aError, aAdditionalData);

    switch (aEvent)
    {
        case EJobStart:
        {
            if (aError) {
                LOG.debug("SyncML session error occurred on JobStart: code %d", aError);
                iCurrentSyncStatus = ESyncStatusError;
                iAppUI->UpdateSourceStatus(iCurrentSource, ESyncStatusError, aError);
                // Should we display a msg for these errors?
                //ManageSyncMLErrors(aError, aAdditionalData);
            }
            else {
                LOG.debug("Starting sync job of '%s'", getSourceNameFromID(iCurrentSource).c_str());
            }
            return;
        }

        case EJobStop:
        {
            firstSyncAfterStartup = false;
            if (aError) {
                if (aError == KErrCancel) {
                    // User canceled the sync
                    LOG.info("Sync cancelled by the user (%d items sent, %d items received)", iTotalServerItems, iTotalClientItems);
                    iCurrentSyncStatus = ESyncStatusCanceled;
                }
                else {
                    // Generic failure
                    LOG.info("Sync ended with error (%d items sent, %d items received)", iTotalServerItems, iTotalClientItems);

                    int itemsSuccesfullySynced = iTotalServerItems + iTotalClientItems;
                    if (itemsSuccesfullySynced > 0) {
                        // if at least 1 item was synced
                        iCurrentSyncStatus = ESyncStatusInterrupted;
                    } else {
                        // 0 items synced, treat it as a sync failed
                        iCurrentSyncStatus = ESyncStatusNotCompleted;
                    }
                    ManageSyncMLErrors(aError, aAdditionalData);
                }
            }
            else {
                // OK. The sync has ended succesfully.
                iCurrentSyncStatus = ESyncStatusCompleted;
                LOG.info("Sync ended succesfully! (%d items sent, %d items received)", iTotalServerItems, iTotalClientItems);
            }
            iAppUI->UpdateSourceStatus(iCurrentSource, iCurrentSyncStatus, aError);

            if (iSyncQueue.Count() > 0) {
                // Start the next source from the queue (if any)
                syncFromQueue();
            }
            else {
                // Last source: check for sw updates
                checkForUpdates();
                
                // Safley closes active connection if not used anymore.
                ClientSettings* config = ClientSettings::getInstance();
                TSyncUsing syncUsing = config->getSyncUsing();
                if (syncUsing != ESyncUsingPush) {
                    closeConnection();
                }

                // Enable/disable pictures source (check Server datastores)
                // The source name and the preferred data type must match.
                refreshSourcesVisible();
            }
            return;
        }

        default:
        {
            // Other events: just log errors
            if (aError) {
                LOG.error("SyncML session error: event = %d, error = %d", aEvent, aError);
            }
            return;
        }

    }
}





TInt CSyncEngine::StopJob(const TSmlJobId aJobID)
{
    TInt err = KErrNone;

    if (aJobID >= 0) {
        RSyncMLDataSyncJob job;
        TRAP(err, job.OpenL(iSyncMLSession, aJobID);)
        if (err) {
            LOG.error("StopJob: error opening job %d, code %d: trying to reinitialize syncML session", aJobID, err);
            if (ReInitializeSessionL() == false) {
                return -1;
            }

            TRAP(err, job.OpenL(iSyncMLSession, aJobID);)

            if (err) {
                Funambol::LOG.error("StopJob: error opening job %d, code %d", aJobID, err);
                return -1;
            }
        }

        LOG.debug("Stopping job (id = %d)", aJobID);
        TRAP(err, job.StopL();)
        if (err) {
            LOG.error("StopJob: error stopping job, code %d", err);
        }
        job.Close();
    }
    return err;
}




TLastSyncStatus CSyncEngine::GetLastSyncStatus(const TInt aSourceID,
                                               TTime* aLastSyncTime,
                                               TInt* aLastSyncResult)
{
    if (aSourceID == SOURCE_MAIL) {
        return GetLastEmailSyncStatus(aLastSyncTime, aLastSyncResult);
    } else if (aSourceID == SOURCE_PHOTOS) {
        return GetLastPhotosSyncStatus(aLastSyncTime, aLastSyncResult);
    } else {
        return GetLastPimSyncStatus(aSourceID, aLastSyncTime, aLastSyncResult);
    }
}

TLastSyncStatus CSyncEngine::GetLastPhotosSyncStatus(TTime* aLastSyncTime,
                                                    TInt*  aLastSyncResult)
{
    TLastSyncStatus lastSyncStatus;
    ClientSettings* config = ClientSettings::getInstance();
    SyncSourceConfig* photosConf = config->getSyncSourceConfig(SOURCE_PHOTOS_NAME);
    TInt64 secs = photosConf->getLast();

    if (secs == 0)
    {
        // Never synchronized or unable to get the timestamp
        lastSyncStatus = ELastSyncNotFound;
    }
    else
    {
        //
        // Our timstamps (anchors) are stored in 'seconds from 01/01/1970'
        // Symbian TTime needs 'microseconds from year 0', so we need this conversion.
        // (62136892800 is a contant = # of seconds from year 0 to 01/01/1970)
        //
        //TTime constant1970;
        //constant1970.Set(_L("19700101"));
        //TInt64 c = constant1970.Int64();
        TInt64 last = secs + CONSTANT_SECONDS_1970;

        TTime tmp(last * 1000 * 1000);
        TTime nowUTC;
        nowUTC.UniversalTime();

        if (tmp > nowUTC) {
            // Fix: last sync time is in the future -> set to 'not found'
            LOG.debug("last tstamp of photos is in the future: set to 0");
            return ELastSyncNotFound;
        }

        utcToLocalTime(tmp);         // it was a UTC value
        *aLastSyncTime = tmp;
        lastSyncStatus = (TLastSyncStatus)config->getLastPhotosSyncStatus();
        (*aLastSyncResult) = config->getLastPhotosSyncStatus();
    }

    return lastSyncStatus;
}

TLastSyncStatus CSyncEngine::GetLastEmailSyncStatus(TTime* aLastSyncTime,
                                                    TInt*  aLastSyncResult)
{
    TLastSyncStatus lastSyncStatus;
    ClientSettings* config = ClientSettings::getInstance();
    if (config->getLastEmailSyncTimestamp() == 0)
    {
        // Never synchronized or unable to get the timestamp
        lastSyncStatus = ELastSyncNotFound;
    }
    else
    {
        TInt64 microsecs = config->getLastEmailSyncTimestamp();
        TTime tmp(microsecs);

        TTime nowUTC;
        nowUTC.UniversalTime();
        if (tmp > nowUTC) {
            // Fix: last sync time is in the future -> set to 'not found'
            LOG.debug("last tstamp of emails is in the future: set to 0");
            return ELastSyncNotFound;
        }

        utcToLocalTime(tmp);         // it was a UTC value
        *aLastSyncTime = tmp;
        lastSyncStatus = (TLastSyncStatus)config->getLastEmailSyncStatus();
        (*aLastSyncResult) = config->getLastEmailSyncStatus();
    }
    return lastSyncStatus;
}



TLastSyncStatus CSyncEngine::GetLastPimSyncStatus(const TInt aSourceID,
                                                  TTime* aLastSyncTime,
                                                  TInt* aLastSyncResult)
{
    //LOG.debug("CSyncEngine::GetLastPimSyncStatus for %s", getSourceNameFromID(aSourceID).c_str());
    TInt err = KErrNone;
    TLastSyncStatus lastSyncStatus = ELastSyncNotFound;

    TTime nowUTC;
    nowUTC.UniversalTime();

    // Get the corresponding task ID to search for
    TSmlTaskId taskID = GetTaskIDFromSourceID(aSourceID);
    //LOG.debug("task = %d", taskID);

    // Opens the historyLog on this profile.
    RSyncMLHistoryLog history;
    TRAP(err, history.OpenL(iSyncMLSession, getActiveProfileID());)
    if (err) {
        LOG.error("Error opening RSyncMLHistoryLog (code %d): trying to reinitialize syncML session..", err);

        if (ReInitializeSessionL() == false) {
            return ELastSyncNotFound;
        }

        TRAP(err, history.OpenL(iSyncMLSession, getActiveProfileID());)

        if (err) {
            LOG.error("Error opening RSyncMLHistoryLog: code %d", err);
            return ELastSyncNotFound;
        }
    }

    // Sort the list by time, last entry is the most recent sync.
    history.SortEntries(CSyncMLHistoryEntry::ESortByTime);
    TInt numEntries = history.Count();

    TInt entryIndex = -1;
    RSyncMLDataSyncProfile profile;
    TRAP(err, profile.OpenL(iSyncMLSession, getActiveProfileID(), ESmlOpenRead); )

    if (err) {
        LOG.error("Error opening RSyncMLDataSyncProfile (code %d): trying to reinitialize syncML session...", err);
        if (ReInitializeSessionL() == false) {
            goto finally;
        }

        TRAP(err, profile.OpenL(iSyncMLSession, getActiveProfileID(), ESmlOpenRead); )
        if (err) {
            Funambol::LOG.error("Error opening RSyncMLDataSyncProfile: code %d", err);
            goto finally;
        }
    }

    // Cycle through the entries, starting from the most recent one.
    // We want to find the latest sync job that contains this task.
    for (TInt i=numEntries-1; i>=0; i--)
    {
        const CSyncMLHistoryJob* hjob = CSyncMLHistoryJob::DynamicCast(&(history.Entry(i)));
        TInt numTasks = hjob->TaskCount();

        // Check if this task was included in this sync job.
        for (int j=0; j<numTasks; j++) {
            if (hjob->TaskAt(j).iTaskId == taskID) {
                entryIndex = i;
                break;
            }
        }

        if (entryIndex >= 0) {
            //
            // Source was synced in the past
            //
            TTime lastSyncTimeUTC = hjob->FinishTime();     // returns an UTC value
            if (lastSyncTimeUTC > nowUTC) {
                // Fix: last sync time is in the future -> ignore this entry.
                LOG.debug("last tstamp of %s is in the future: set to 0", getSourceNameFromID(aSourceID).c_str());
                entryIndex = -1;
                continue;
            }

            if (hjob->ResultCode() == 0) {
                //
                // Last sync successful: get last timestamp and set the out parameters
                //
                (*aLastSyncTime) = lastSyncTimeUTC;
                utcToLocalTime(*aLastSyncTime);

                lastSyncStatus = ELastSyncSuccessful;
                (*aLastSyncResult) = 0;
            }
            else {
                //
                // Last sync not completed
                //
                lastSyncStatus = ELastSyncFailed;
                (*aLastSyncResult) = hjob->ResultCode();
            }
            goto finally;
        }
    }

    //
    // If here, the source was never synced. Note: out parameters are not modified.
    //
    lastSyncStatus = ELastSyncNotFound;


finally:
    profile.Close();
    history.Close();
    return lastSyncStatus;
}


const bool CSyncEngine::isSourceEnabled(const TInt aSourceID)
{
    // Read the config
    StringBuffer name = getSourceNameFromID(aSourceID);
    ClientSettings* config = ClientSettings::getInstance();
    SyncSourceConfig* ssc = config->getSyncSourceConfig(name.c_str());
    if (!ssc) {
        LOG.error("Config not found for source '%s'", name.c_str());
        return false;
    }

    return ssc->isEnabled();
}

void CSyncEngine::enableSource(const TInt aSourceID, const bool aEnable)
{
    // Read the config
    StringBuffer name = getSourceNameFromID(aSourceID);
    ClientSettings* config = ClientSettings::getInstance();
    SyncSourceConfig* ssc = config->getSyncSourceConfig(name.c_str());
    if (!ssc) {
        LOG.error("Config not found for source '%s'", name.c_str());
        return;
    }

    ssc->setIsEnabled(aEnable);
}


void CSyncEngine::setSourceTimestamp(const TInt aSourceID, const TUint timestamp)
{
    StringBuffer name = getSourceNameFromID(aSourceID);
    ClientSettings* config = ClientSettings::getInstance();
    SyncSourceConfig* ssc = config->getSyncSourceConfig(name.c_str());
    if (!ssc) {
        LOG.error("Config not found for source '%s'", name.c_str());
        return;
    }
    ssc->setLast(timestamp);
    config->save();
}


/** Funambol API listeners hook
 * set sync status of source and updates UI
 */
void CSyncEngine::onListenerStatusUpdate(const TInt aSourceID, const TCurrentSyncStatus aSyncStatus, const TInt aSyncData)
{
    //LOG.debug("onListenerStatusUpdate (source = %s, status %d)", getSourceNameFromID(aSourceID).c_str(), aSyncStatus);

    // only briefcase sources are handled
    if (aSourceID == SOURCE_PHOTOS) {
        iCurrentSyncStatus = aSyncStatus;

        // filter out events to be forwarded to UI
        switch (aSyncStatus) {
            case ESyncStatusConnecting:
            case ESyncStatusConnected:
            case ESyncStatusDisconnected:
            case ESyncStatusLoggingOn:
            case ESyncStatusLoggedOn:
            case ESyncStatusSendingData:
            case ESyncStatusReceivingData:
            case ESyncStatusMappings:
            case ESyncStatusCompleted:
            case ESyncStatusNotCompleted:
            case ESyncStatusSendingItemNumber:
            case ESyncStatusReceivingItemNumber:
            case ESyncStatusError:
            case ESyncStatusCanceled:
                iAppUI->UpdateSourceStatus(aSourceID, aSyncStatus, aSyncData);
                break;
            default:
                break;
        }
    }
    else if ((aSourceID == SOURCE_CONFIG) && hiddenConfigSync) {
        LOG.debug("Hidden config sync running");
        // A hidden config sync just started: update the next source pane with "connecting..."
        updateNextSourceStatus(ESyncStatusConnecting);
    }
}


TBuf<32> CSyncEngine::GetDataSourceFromSourceID(const TInt aSourceID)
{
    for (int i=0; i<NUM_PIM_SYNC_APPS; i++) {
        if (SyncTasks[i].sourceId == aSourceID) {
            return (SyncTasks[i].dataSource);
        }
    }
    return SyncTasks[0].dataSource;
}

TInt CSyncEngine::GetSourceIDFromDataSource(const TDesC& aDataSource)
{
    for (int i=0; i<NUM_PIM_SYNC_APPS; i++) {
        if (SyncTasks[i].dataSource == aDataSource) {
            return (SyncTasks[i].sourceId);
        }
    }
    return SyncTasks[0].sourceId;
}


TSmlTaskId CSyncEngine::GetTaskIDFromSourceID(const TInt aSourceID)
{
    TInt err = KErrNone;
    TSmlTaskId taskID = KErrNotFound;

    // Open the active profile
    RSyncMLDataSyncProfile profile;
    TRAP(err, profile.OpenL(iSyncMLSession, getActiveProfileID(), ESmlOpenRead);)
    if (err) {
        LOG.error("Error opening RSyncMLDataSyncProfile (code %d): trying to reinitialize syncML session", err);

        if (ReInitializeSessionL() == false) {
            return KErrNotFound;
        }


        TRAP(err, profile.OpenL(iSyncMLSession, getActiveProfileID(), ESmlOpenRead);)

        if (err) {
            LOG.error("Error opening RSyncMLDataSyncProfile: code %d", err);
            return KErrNotFound;
        }
    }

    // Get the active list of tasks
    RArray<TSmlTaskId> pTaskArray;
    TBuf<32> taskDataSource;
    TRAP(err, profile.ListTasksL(pTaskArray);)
    if (err) {
        LOG.error("Error getting task list: code %d", err);
        goto finally;
    }

    // Get the dataSource name of the task we want to synchronize
    taskDataSource = GetDataSourceFromSourceID(aSourceID);

    //
    // Search the desired task inside the profile task array
    //
    for (TInt count=0; count<pTaskArray.Count(); ++count)
    {
        RSyncMLTask task;
        TRAP(err, task.OpenL(profile, pTaskArray[count]);)
        if (err) {
            LOG.error("Error opening RSyncMLTask: code %d", err);
            goto finally;
        }

        // search for a particular task based on 'taskDataSource'
        if (task.ClientDataSource().Compare(taskDataSource) == 0) {
            taskID = task.Identifier();
            task.Close();
            break;
        }
        task.Close();
    }

finally:
    pTaskArray.Reset();
    profile.Close();
    return taskID;
}


TInt CSyncEngine::GetSourceIDFromTaskID(const TSmlTaskId aTaskID)
{
    TInt err = KErrNone;
    TInt sourceID = SOURCE_CONTACTS;

    // Open the active profile
    RSyncMLDataSyncProfile profile;
    TRAP(err, profile.OpenL(iSyncMLSession, getActiveProfileID(), ESmlOpenRead);)
    if (err) {
        LOG.error("Error opening RSyncMLDataSyncProfile: code %d", err);
        if (ReInitializeSessionL() == false) {
            return KErrNotFound;
        }

        TRAP(err, profile.OpenL(iSyncMLSession, getActiveProfileID(), ESmlOpenRead);)

        if (err) {
            return SOURCE_CONTACTS; // as a default
        }
    }

    // Get the active list of tasks
    RArray<TSmlTaskId> pTaskArray;
    TRAP(err, profile.ListTasksL(pTaskArray);)
    if (err) {
        LOG.error("Error getting task list: code %d", err);
        goto finally;
    }

    // Find the desired task inside the profile task array
    TInt index = pTaskArray.Find(aTaskID);
    if (index != KErrNotFound)
    {
        // Open the task and get the dataSource name
        RSyncMLTask task;
        TRAP(err, task.OpenL(profile, pTaskArray[index]);)
        if (err) {
            LOG.error("Error opening RSyncMLTask: code %d", err);
            goto finally;
        }

        sourceID = GetSourceIDFromDataSource(task.ClientDataSource());
        task.Close();
    }

finally:
    pTaskArray.Reset();
    profile.Close();
    return sourceID;
}



TInt CSyncEngine::AddToSyncQueue(const TInt aSourceID)
{
    // Check if the current source is already in the queue.
    if (findSourceIDInSyncQueue(aSourceID)) {
        LOG.debug("source %d is already queued, discard.", aSourceID);
    }
    else {
        iCriticalSyncQueue.Wait();
        iSyncQueue.Append(aSourceID);
        iCriticalSyncQueue.Signal();
    }
    return iSyncQueue.Count();
}

bool CSyncEngine::InsertConfigInSyncQueue()
{
    // Check if the config source is already in the queue.
    if (findSourceIDInSyncQueue(SOURCE_CONFIG)) {
        return false;
    }

    // Optimization: no need to insert a config sync if
    // we're going to sync the pictures this session.
    if (findSourceIDInSyncQueue(SOURCE_PHOTOS)) {
        return false;
    }

    iCriticalSyncQueue.Wait();
    iSyncQueue.Insert(SOURCE_CONFIG, 0);   // insert on the front of the queue
    iCriticalSyncQueue.Signal();

    return true;
}

bool CSyncEngine::InsertInSyncQueue(const TInt aSourceID, const TInt pos)
{
    // Check if the source is already in the queue.
    if (findSourceIDInSyncQueue(aSourceID)) {
        return false;
    }
    
    // constrain the index to correct values
    TInt index = Max(pos, 0);
    index = Min(pos, iSyncQueue.Count());

    iCriticalSyncQueue.Wait();
    iSyncQueue.Insert(aSourceID, pos);
    iCriticalSyncQueue.Signal();

    return true;
}


TBool CSyncEngine::findSourceIDInSyncQueue(const TInt aSourceID)
{
    for (TInt i=0; i<iSyncQueue.Count(); i++) {
        if (iSyncQueue[i] == aSourceID) {
            return ETrue;
        }
    }
    return EFalse;
}

bool CSyncEngine::isSyncQueueEmpty()
{
    return (iSyncQueue.Count() == 0)? true : false;
}


TBool CSyncEngine::isSourceUnderSyncAndReceiving(const TInt aSourceID)
{
    // First: check on the Calendar source
    if (aSourceID == SOURCE_CALENDAR)
    {
        // Calendar Change notification can arrive a bit later...
        // So we also check if a sync has just finished (error window = 2 seconds).
        TInt lastSyncResult = 0;
        TTime now, lasSyncTime;
        now.HomeTime();
        TLastSyncStatus result = GetLastPimSyncStatus(aSourceID, &lasSyncTime, &lastSyncResult);
        if (result != ELastSyncNotFound) {
            if (lasSyncTime > (now - TTimeIntervalSeconds(2))) {
                LOG.debug("Calendar sync just finished");
                return ETrue;
            }
        }
    }

    if (!CheckSyncInProgress()) {
        return EFalse;
    }

    if (aSourceID == iCurrentSource) {
        if ( iCurrentSyncStatus == ESyncStatusReceivingData       ||
             iCurrentSyncStatus == ESyncStatusReceivingItemNumber ||
             iCurrentSyncStatus == ESyncStatusMappings ) {      // Just to be sure, maybe a little delay...
            return ETrue;
        }
    }

    return EFalse;
}

TSmlSyncType CSyncEngine::getSyncTypeFromText(const StringBuffer& text)
{
    if (text == SYNC_ONE_WAY_FROM_CLIENT_TEXT) {
        return ESmlOneWayFromClient;
    }
    else if (text == SYNC_ONE_WAY_FROM_SERVER_TEXT) {
        return ESmlOneWayFromServer;
    }
    else if (text == SYNC_SLOW_TEXT) {
        return ESmlSlowSync;
    }
    else {
        return ESmlTwoWay;  // default
    }
}


void CSyncEngine::RunL()
{
    //LOG.debug("RunL - status = %d, state = %d", iStatus.Int(), iAsyncAction);
    if(iStatus == KErrNone)
    {
        switch(iAsyncAction)
        {
            case ESyncFromQueue:
            {
                syncFromQueue();
                break;
            }
            case EStopSync:
            {
                StopSync();
                break;
            }
            case EEnableScheduler:
            {
                // Currently not used.
                // Can be used to trigger a scheduled sync, from a thread
                // different than the main thread.
                iScheduler->enable();
                break;
            }
            default:
                break;
        }
        iAsyncAction = EIdle;
    }
    else {
        LOG.error("CSyncEngine::RunL() error: status = %d", iStatus.Int());
    }
}

void CSyncEngine::DoCancel()
{
    // nothing to do
}

TInt CSyncEngine::RunError(TInt aError) {
    LOG.error("CSyncEngine RunL() error: %d", aError);
    return 0;
}


void CSyncEngine::RequestAsyncAction(const TAsyncAction aAction)
{
    LOG.debug("Requesting async action! state is %d", aAction);

    // Active object will switch to the correct method to call, based on this value.
    iAsyncAction = aAction;

    if(!IsActive()) {
        iStatus = KRequestPending;
        SetActive();

        // Signals that the asynchronous request associated with 'iStatus' is complete
        // Note: we must do it in the same thread that originated the request,
        //       so we signal it in the main thread.
        RThread mainThread;
        mainThread.Open(iMainThreadID);

        TRequestStatus* status = &iStatus;
        mainThread.RequestComplete(status, KErrNone);
    }
    else {
        LOG.debug("CSyncEngine - async request already pending");
    }
}




void CSyncEngine::syncPIMSource(const TInt aSourceID)
{
    // Read the syncType from config.
    ClientSettings* config = ClientSettings::getInstance();
    StringBuffer sourceName = getSourceNameFromID(aSourceID);
    SyncSourceConfig* ssc = config->getSyncSourceConfig(sourceName.c_str());
    if (!ssc) {
        LOG.error("StartSync aborted: config not found for source '%s'", sourceName.c_str());
        return;
    }

    TSmlSyncType syncType;
    if (ssc->getLast() == 0) {
        if (aSourceID == SOURCE_NOTES) {
            // SlowSync is not supported for Notes!! :-0
            LOG.info("Starting a TWO-WAY SYNC for '%s' (can't force a slow sync)", sourceName.c_str());
            syncType = ESmlTwoWay;
        }
        else {
            // Last timestamp is 0: this means it's the first sync, or the source was cleaned up.
            LOG.info("Starting a SLOW SYNC for '%s'", sourceName.c_str());
            syncType = ESmlSlowSync;
        }
    }
    else {
        syncType = getSyncTypeFromText(ssc->getSync());
    }

    // Check if the SyncML profile exists.
    // This must be done before retrieving the task ID.
    if (getActiveProfileID() == KErrNotFound) {
        iAppUI->DisplayMessage(ERR_PROFILE_NOT_FOUND, EErrorLevel);
        return;
    }

    // Get the corresponding task from the list of tasks of active profile
    TSmlTaskId taskID = GetTaskIDFromSourceID(aSourceID);
    LOG.debug("task id = %d", taskID);

    if(taskID == KErrNotFound) {
        LOG.error("Internal error: sync task not found for source '%s'", sourceName.c_str());
        return;
    }
    else {
        // Applications are tasks; every profile has a list of tasks associated
        // We use a specific array of tasks "required", with only 1 taskId.
        RArray<TSmlTaskId> rTaskArray;
        rTaskArray.Append(taskID);

        // Start the new job of synchronization
        RSyncMLDataSyncJob job;
        TRAPD(err, job.CreateL(iSyncMLSession, getActiveProfileID(), syncType, rTaskArray))

        if (err) {
            if (err == KErrServerTerminated) {
                LOG.info("reinitializing syncml session after it has become invalid");

                if (ReInitializeSessionL() == false) {
                    return;
                }

                TRAPD(err, job.CreateL(iSyncMLSession, getActiveProfileID(), syncType, rTaskArray))
                rTaskArray.Close();

                if (err) {
                    LOG.error("Error creating RSyncMLDataSyncJob: code %d", err);
                    return;
                }
            } else {
                LOG.error("Error creating RSyncMLDataSyncJob: code %d", err);
                rTaskArray.Close();

                return;
            }
        } else {
            rTaskArray.Close();
        }

        job.Close();
    }
}


TInt CSyncEngine::syncEmail()
{
    TInt err = KErrNone;
    LOG.debug("Beginning of sync email");
    logMemInfo();

    if (emailEngine == NULL) {
        emailEngine = CImapEmailEngine::NewL(this, iAppUI, SOURCE_MAIL);
    } else {
        // If another sync is already in progress, then we do not
        // start a new one
        if (emailEngine->CheckSyncInProgress()) {
            LOG.info("Email sync already in progress");
            return 1;
        }
    }

    // TODO: check if account exist?

    LOG.debug("Starting sync of Emails");
    TRAP(err, emailEngine->PrepareL());
    if (err) {
        if (err = KErrNotFound) {
            LOG.error("Cannot sync emails: IMAP account not found");
            iAppUI->DisplayMessage(ERR_ACCOUNT_NOT_FOUND, EErrorLevel);
        }
        else {
            LOG.error("Failed to prepare Email sync: code %d", err);
        }
        return err;
    }
    TRAP(err, emailEngine->RemoteFetchL());
    if (err) {
        LOG.error("Failed to start Email sync: code %d", err);
        return err;
    }

    return 0;
}



void CSyncEngine::checkForUpdates()
{
    if (iCurrentSource != SOURCE_CONFIG) {
        UpdateManager updateManager(iAppUI);
        updateManager.checkForUpdates();
    }
}

void CSyncEngine::syncConfig()
{
    // We don't want more than 1 sync instance
    if (SyncActiveObject::CheckSyncInProgress()) {
        LOG.info("Config sync already in progress");
        return;
    }

    // Initialize the Cache File Name
    StringBuffer fullPath(PlatformAdapter::getHomeFolder());
    StringBuffer cacheFile;
    cacheFile.sprintf("%s%s", fullPath.c_str(), "config_sync_source.dat");
    TRAPD(err, SyncActiveObject::syncLD(this, SOURCE_CONFIG, cacheFile));


    if (hiddenConfigSync) {
        LOG.debug("Hidden config sync started");
        // A hidden config sync just started: update the next source pane with "connecting..."
        updateNextSourceStatus(ESyncStatusConnecting);
    }
}

void CSyncEngine::syncPhotos()
{
    // We don't want more than 1 sync instance
    if (SyncActiveObject::CheckSyncInProgress()) {
        LOG.info("Photo sync already in progress");
        return;
    }

    if (iListenerPool == NULL) {
        iListenerPool = new (ELeave)ListenerPool(this, SOURCE_PHOTOS);
    }

    if (iListenerPool) {
        iListenerPool->setListeners();
    }

    ClientSettings* config = ClientSettings::getInstance();

    //
    // Check if the pictures location is available (memory card removed)
    // If not, ask the user if it's ok to sync phone's memory pictures.
    //
    StringBuffer imagesDir;
    TInt picturesStorage = ClientSettings::getInstance()->getPicturesStorage();
    bool driveAvailable = getImagesPath(picturesStorage, imagesDir);
   
    if (!driveAvailable) {
        bool fallback = false;
        TBool massStorageAvailable = massStorageMemoryAvailable();
        StringBuffer picturesDrive = getDriveName(picturesStorage);
        // if drive is become not available must be a memory card 
        LOG.info("Pictures drive '%s' not available: popup query", picturesDrive.c_str());

        if (massStorageAvailable) {
            fallback = iAppUI->DisplayQuery(QUERY_DRIVE_NOT_AVAILABLE_MASS_STORAGE);
        } else {
            fallback = iAppUI->DisplayQuery(QUERY_DRIVE_NOT_AVAILABLE_SYNC);
        }

        if (fallback) {
            // User selected YES: save the new drive letter to config and continue
            TInt storageType = EStorageTypeUnset;
            if (massStorageAvailable) {
                storageType = EMassMemoryStorage;
            } else {
                storageType = EPhoneMemoryStorage;
            }

            config->setPicturesStorage(storageType);
            picturesDrive = getDriveName(storageType);

            LOG.debug("Fallback to pictures location: %s", picturesDrive.c_str());
            config->save();
        }
        else {
            // User selected NO: don't sync
            LOG.debug("User selected no: aborting sync of pictures");
            return;
        }
    }

    //
    // Popup the first-sync-warning: only the very first sync of pictures
    //
    LOG.debug("firstPicturesSyncTimestamp = %lu", config->getFirstPicturesSyncTimestamp());
    if (config->getFirstPicturesSyncTimestamp() == 0) {
        LOG.info("This is the first sync of pictures");
        // persist immediately the new timestamp, so the popup will never be shown again
        config->setFirstPicturesSyncTimestamp(time(NULL), true);

        if (config->getIncludeOlderPictures()) {
            TBool ok = iAppUI->ShowPicturesFirstSyncWarning();
            if (!ok) {
                LOG.debug("User selected try later: aborting sync of pictures");
                return;
            }
        } else {
            LOG.debug("the filtering of old pictures is ON");
        }
    }

    LOG.debug("Starting sync of pictures...");
    TRAPD(err, SyncActiveObject::syncLD(this, SOURCE_PHOTOS));

    // Sync just started: update the UI pane with "connecting..."
    iAppUI->UpdateSourceStatus(SOURCE_PHOTOS, ESyncStatusConnecting);
}


void CSyncEngine::syncEnded(const TInt aSourceId, TInt status, const int itemsSuccesfullySynced)
{
    TLastSyncStatus lastStatus;
    LOG.debug("[%s] %i items succesfully synchronized.", getSourceNameFromID(aSourceId).c_str(), itemsSuccesfullySynced);
    firstSyncAfterStartup = false;

    if (status == KErrCancel) {
        //
        // Sync canceled by the user
        //
        iAppUI->UpdateSourceStatus(aSourceId, ESyncStatusCanceled, status);
        lastStatus = ELastSyncCanceled;
    }
    else if (status == ERR_RESTART_SYNC) {
        // 
        // Sync needs to be restarted. 
        // Actually used only for pictures in case 1st sync after an upgrade.
        //
        InsertInSyncQueue(aSourceId, 0);
        syncFromQueue();
        return;
    }
    else if (status == ERR_SERVER_QUOTA_EXCEEDED) {
        // Quota exceeded on server (used by pictures sync)
        iAppUI->UpdateSourceStatus(aSourceId, ESyncStatusQuotaExceeded, status);
        lastStatus = ELastSyncQuotaExceeded;
    }
    else if (status) {
        //
        // All other errors (status != 0)
        //
        if (itemsSuccesfullySynced > 0) {
            // if at least 1 item was synced
            iAppUI->UpdateSourceStatus(aSourceId, ESyncStatusInterrupted, status);
            lastStatus = ELastSyncInterrupted;
        } else {
            // 0 items synced, treat it as a sync failed
            iAppUI->UpdateSourceStatus(aSourceId, ESyncStatusError, status);
            lastStatus = ELastSyncFailed;
        }
        
        if (status >= ERR_TRANSPORT_BASE && 
            status <= ERR_TRANSPORT_LAST) {
            LOG.debug("Network error: canceling queued syncs");
            resetSyncQueue();
        }
    }
    else {
        //
        // sync completed successfully (status = 0)
        //
        iAppUI->UpdateSourceStatus(aSourceId, ESyncStatusCompleted, status);
        lastStatus = ELastSyncSuccessful;
    }


    if (aSourceId == SOURCE_CONFIG) {
        // To update the account setting dialog
        iAppUI->ConfigSyncEnded();
    }
    else if (aSourceId == SOURCE_PHOTOS) {
        lastPicturesSyncError = status;     // local usage
        ClientSettings* config = ClientSettings::getInstance();
        config->setLastPhotosSyncStatus(lastStatus);
        config->save();

        if (iListenerPool) {
            iListenerPool->releaseAllListeners();
        }

        iAppUI->PhotosSyncEnded();
    }


    // This needs to be done at the very end.
    if (iSyncQueue.Count() > 0) {
        // Start the next source from the queue (if any)
        syncFromQueue();
    }
    else {
        // Last source: check for sw updates
        checkForUpdates();

        // Safley closes active connection if not used anymore.
        ClientSettings* config = ClientSettings::getInstance();
        TSyncUsing syncUsing = config->getSyncUsing();
        if (syncUsing != ESyncUsingPush) {
            closeConnection();
        }

        // Enable/disable pictures source (check Server datastores)
        // The source name and the preferred data type must match.
        refreshSourcesVisible();
    }
}


bool CSyncEngine::refreshSourcesVisible()
{
    ClientSettings* config = ClientSettings::getInstance();
    bool ret = false;

    if (DYNAMICALLY_SHOW_PICTURES) {
        LOG.debug("checking the server caps for pictures datastore");
        DataStore* dataStore = config->getServerDataStore(PICTURES_REMOTE_URI);
        if ( dataStore && !strcmp(dataStore->getRxPref()->getCTType(), OMA_MIME_TYPE) ) {
            LOG.debug("'%s' datastore found in Server capabilities", SOURCE_PHOTOS_NAME);
            ret = config->safeAddSourceVisible(SOURCE_PHOTOS_NAME);
        }
        else {
            LOG.debug("'%s' datastore NOT found in Server capabilities", SOURCE_PHOTOS_NAME);
            ret = config->removeSourceVisible(SOURCE_PHOTOS_NAME);
        }
    }

    // If necessary, refresh the main  UI and store config.
    if (ret) {
        iAppUI->RefreshScreen();
        config->saveSourcesVisible();
    }
    return ret;
}



void CSyncEngine::EmailSyncEnded(TCurrentSyncStatus /* status */)
{
    CImapEmailEngine* tmpEngine = emailEngine;
    emailEngine = NULL;
    delete tmpEngine;

    LOG.debug("End of sync email");
    logMemInfo();

    if (iSyncQueue.Count() > 0) {
        // Start the next source from the queue (if any)
        syncFromQueue();
    }
    else {
        // Last source: check for sw updates
        checkForUpdates();
    }
}


void CSyncEngine::updateAppUISourceStatus(const TInt aSourceID, const TCurrentSyncStatus aSyncStatus, const TInt aSyncData)
{
    iAppUI->UpdateSourceStatus(aSourceID, aSyncStatus, aSyncData);
}


void CSyncEngine::updateNextSourceStatus(const TCurrentSyncStatus aSyncStatus)
{
    if (!isSyncQueueEmpty()) {
        TInt nextSourceID = (TInt)iSyncQueue[0];
        iAppUI->UpdateSourceStatus(nextSourceID, aSyncStatus, 0);
        LOG.debug("updated next source pane (%s)", getSourceNameFromID(nextSourceID).c_str());
    }
}

void CSyncEngine::enableScheduler(const bool aIsRepetitive, const unsigned int aInterval)
{
    if (iScheduler) {
        iScheduler->setUseActiveSourceList(true);
        iScheduler->enable(aIsRepetitive, aInterval);
    }
}

void CSyncEngine::disableScheduler()
{
    if (iScheduler) {
        iScheduler->disable();
    }
}

bool CSyncEngine::isSchedulerEnabled()
{
    if (iScheduler) {
        return iScheduler->isEnabled();
    }
    return false;
}

bool CSyncEngine::removeFromScheduler(const TInt aSourceID)
{
    if (iScheduler) {
        return iScheduler->removeFromSourceList(aSourceID);
    }
    return false;
}

RArray<TInt> CSyncEngine::GetActiveSourcesList() {
    return iAppUI->GetActiveSourcesList();
}


void CSyncEngine::DisplayMessage(const TInt aMsgID, const TWarningLevel aWarningLevel) {
    if (iAppUI) {
        iAppUI->DisplayMessage(aMsgID, aWarningLevel);
    }
}

